
// ColorFactory.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, October 2021

/* Based on ColorFactory.java by J. H. S.
   lobochief@users.sourceforge.net
   April 17, 2005
   http://www.java2s.com/Code/Java/2D-Graphics-GUI/ColorFactory.htm

   Simplified, and modified to support Color3f.
*/


import java.awt.*;
import java.util.*;
import java.util.Map.Entry;

import org.jogamp.vecmath.*;



public class ColorFactory
{

  private static Random rand = new Random();

  private static TreeMap<String,Color3f> colorMap;

  // static initializer
  static {
    initMap();
  }


  private ColorFactory()
  { }


  private static void initMap()
  {
    colorMap = new TreeMap<>();

    //http://www.w3schools.com/css/css_colornames.asp
    colorMap.put("aliceblue", toColor3f(0xf0f8ff));
    colorMap.put("antiquewhite", toColor3f(0xfaebd7));
    colorMap.put("aqua", toColor3f(0x00ffff));
    colorMap.put("aquamarine", toColor3f(0x7fffd4));
    colorMap.put("azure", toColor3f(0xf0ffff));
    colorMap.put("beige", toColor3f(0xf5f5dc));
    colorMap.put("bisque", toColor3f(0xffe4c4));
    colorMap.put("black", toColor3f(0x000000));
    colorMap.put("blackish", new Color3f(0.2f, 0.2f, 0.2f));
    colorMap.put("blanchedalmond", toColor3f(0xffebcd));
    colorMap.put("blue", toColor3f(0x0000ff));
    colorMap.put("blueviolet", toColor3f(0x8a2be2));
    colorMap.put("brown", toColor3f(0xa52a2a));
    colorMap.put("burlywood", toColor3f(0xdeb887));
    colorMap.put("cadetblue", toColor3f(0x5f9ea0));
    colorMap.put("chartreuse", toColor3f(0x7fff00));
    colorMap.put("chocolate", toColor3f(0xd2691e));
    colorMap.put("coral", toColor3f(0xff7f50));
    colorMap.put("cornflowerblue", toColor3f(0x6495ed));
    colorMap.put("cornsilk", toColor3f(0xfff8dc));
    colorMap.put("crimson", toColor3f(0xdc143c));
    colorMap.put("cyan", toColor3f(0x00ffff));
    colorMap.put("darkblue", toColor3f(0x00008b));
    colorMap.put("darkcyan", toColor3f(0x008b8b));
    colorMap.put("darkgoldenrod", toColor3f(0xb8860b));
    colorMap.put("darkgray", toColor3f(0xa9a9a9));
    colorMap.put("darkgrey", toColor3f(0xa9a9a9));
    colorMap.put("darkgreen", toColor3f(0x006400));
    colorMap.put("darkkhaki", toColor3f(0xbdb76b));
    colorMap.put("darkmagenta", toColor3f(0x8b008b));
    colorMap.put("darkolivegreen", toColor3f(0x556b2f));
    colorMap.put("darkorange", toColor3f(0xff8c00));
    colorMap.put("darkorchid", toColor3f(0x9932cc));
    colorMap.put("darkred", toColor3f(0x8b0000));
    colorMap.put("darksalmon", toColor3f(0xe9967a));
    colorMap.put("darkseagreen", toColor3f(0x8fbc8f));
    colorMap.put("darkslateblue", toColor3f(0x483d8b));
    colorMap.put("darkslategray", toColor3f(0x2f4f4f));
    colorMap.put("darkslategrey", toColor3f(0x2f4f4f));
    colorMap.put("darkturquoise", toColor3f(0x00ced1));
    colorMap.put("darkviolet", toColor3f(0x9400d3));
    colorMap.put("deeppink", toColor3f(0xff1493));
    colorMap.put("deepskyblue", toColor3f(0x00bfff));
    colorMap.put("dimgray", toColor3f(0x696969));
    colorMap.put("dimgrey", toColor3f(0x696969));
    colorMap.put("dodgerblue", toColor3f(0x1e90ff));
    colorMap.put("firebrick", toColor3f(0xb22222));
    colorMap.put("floralwhite", toColor3f(0xfffaf0));
    colorMap.put("forestgreen", toColor3f(0x228b22));
    colorMap.put("fuchsia", toColor3f(0xff00ff));
    colorMap.put("gainsboro", toColor3f(0xdcdcdc));
    colorMap.put("ghostwhite", toColor3f(0xf8f8ff));
    colorMap.put("gold", toColor3f(0xffd700));
    colorMap.put("goldenrod", toColor3f(0xdaa520));
    colorMap.put("grass", toColor3f(76, 235, 52));
    colorMap.put("gray", toColor3f(0x808080));
    colorMap.put("grey", toColor3f(0x808080));
    colorMap.put("green", toColor3f(0x008000));
    colorMap.put("greenyellow", toColor3f(0xadff2f));
    colorMap.put("honeydew", toColor3f(0xf0fff0));
    colorMap.put("hotpink", toColor3f(0xff69b4));
    colorMap.put("indianred", toColor3f(0xcd5c5c));
    colorMap.put("indigo", toColor3f(0x4b0082));
    colorMap.put("ivory", toColor3f(0xfffff0));
    colorMap.put("khaki", toColor3f(0xf0e68c));
    colorMap.put("lavender", toColor3f(0xe6e6fa));
    colorMap.put("lavenderblush", toColor3f(0xfff0f5));
    colorMap.put("lawngreen", toColor3f(0x7cfc00));
    colorMap.put("lemonchiffon", toColor3f(0xfffacd));
    colorMap.put("lightblue", toColor3f(0xadd8e6));
    colorMap.put("lightcoral", toColor3f(0xf08080));
    colorMap.put("lightcyan", toColor3f(0xe0ffff));
    colorMap.put("lightgoldenrodyellow", toColor3f(0xfafad2));
    colorMap.put("lightgray", toColor3f(0xd3d3d3));
    colorMap.put("lightgrey", toColor3f(0xd3d3d3));
    colorMap.put("lightgreen", toColor3f(0x90ee90));
    colorMap.put("lightpink", toColor3f(0xffb6c1));
    colorMap.put("lightsalmon", toColor3f(0xffa07a));
    colorMap.put("lightseagreen", toColor3f(0x20b2aa));
    colorMap.put("lightskyblue", toColor3f(0x87cefa));
    colorMap.put("lightslategray", toColor3f(0x778899));
    colorMap.put("lightslategrey", toColor3f(0x778899));
    colorMap.put("lightsteelblue", toColor3f(0xb0c4de));
    colorMap.put("lightyellow", toColor3f(0xffffe0));
    colorMap.put("lime", toColor3f(0x00ff00));
    colorMap.put("limegreen", toColor3f(0x32cd32));
    colorMap.put("linen", toColor3f(0xfaf0e6));
    colorMap.put("magenta", toColor3f(0xff00ff));
    colorMap.put("maroon", toColor3f(0x800000));
    colorMap.put("mediumaquamarine", toColor3f(0x66cdaa));
    colorMap.put("mediumblue", toColor3f(0x0000cd));
    colorMap.put("mediumorchid", toColor3f(0xba55d3));
    colorMap.put("mediumpurple", toColor3f(0x9370d8));
    colorMap.put("mediumseagreen", toColor3f(0x3cb371));
    colorMap.put("mediumslateblue", toColor3f(0x7b68ee));
    colorMap.put("mediumspringgreen", toColor3f(0x00fa9a));
    colorMap.put("mediumturquoise", toColor3f(0x48d1cc));
    colorMap.put("mediumvioletred", toColor3f(0xc71585));
    colorMap.put("midnightblue", toColor3f(0x191970));
    colorMap.put("mintcream", toColor3f(0xf5fffa));
    colorMap.put("mistyrose", toColor3f(0xffe4e1));
    colorMap.put("moccasin", toColor3f(0xffe4b5));
    colorMap.put("navajowhite", toColor3f(0xffdead));
    colorMap.put("navy", toColor3f(0x000080));
    colorMap.put("oldlace", toColor3f(0xfdf5e6));
    colorMap.put("olive", toColor3f(0x808000));
    colorMap.put("olivedrab", toColor3f(0x6b8e23));
    colorMap.put("orange", toColor3f(0xffa500));
    colorMap.put("orangered", toColor3f(0xff4500));
    colorMap.put("orchid", toColor3f(0xda70d6));
    colorMap.put("palegoldenrod", toColor3f(0xeee8aa));
    colorMap.put("palegreen", toColor3f(0x98fb98));
    colorMap.put("paleturquoise", toColor3f(0xafeeee));
    colorMap.put("palevioletred", toColor3f(0xd87093));
    colorMap.put("papayawhip", toColor3f(0xffefd5));
    colorMap.put("peachpuff", toColor3f(0xffdab9));
    colorMap.put("peru", toColor3f(0xcd853f));
    colorMap.put("pink", toColor3f(0xffc0cb));
    colorMap.put("plum", toColor3f(0xdda0dd));
    colorMap.put("powderblue", toColor3f(0xb0e0e6));
    colorMap.put("purple", toColor3f(0x800080));
    colorMap.put("red", toColor3f(0xff0000));
    colorMap.put("rosybrown", toColor3f(0xbc8f8f));
    colorMap.put("royalblue", toColor3f(0x4169e1));
    colorMap.put("saddlebrown", toColor3f(0x8b4513));
    colorMap.put("salmon", toColor3f(0xfa8072));
    colorMap.put("sandybrown", toColor3f(0xf4a460));
    colorMap.put("seagreen", toColor3f(0x2e8b57));
    colorMap.put("seashell", toColor3f(0xfff5ee));
    colorMap.put("sienna", toColor3f(0xa0522d));
    colorMap.put("silver", toColor3f(0xc0c0c0));
    colorMap.put("skyblue", toColor3f(52, 168, 235));
    colorMap.put("slateblue", toColor3f(0x6a5acd));
    colorMap.put("slategray", toColor3f(0x708090));
    colorMap.put("slategrey", toColor3f(0x708090));
    colorMap.put("snow", toColor3f(0xfffafa));
    colorMap.put("springgreen", toColor3f(0x00ff7f));
    colorMap.put("steelblue", toColor3f(0x4682b4));
    colorMap.put("tan", toColor3f(0xd2b48c));
    colorMap.put("teal", toColor3f(0x008080));
    colorMap.put("thistle", toColor3f(0xd8bfd8));
    colorMap.put("tomato", toColor3f(0xff6347));
    colorMap.put("turquoise", toColor3f(0x40e0d0));
    colorMap.put("violet", toColor3f(0xee82ee));
    colorMap.put("wheat", toColor3f(0xf5deb3));
    colorMap.put("white", toColor3f(0xffffff));
    colorMap.put("whitish", new Color3f(0.8f, 0.8f, 0.8f));
    colorMap.put("whitesmoke", toColor3f(0xf5f5f5));
    colorMap.put("yellow", toColor3f(0xffff00));
    colorMap.put("yellowgreen", toColor3f(0x9acd32));
  }  // end of initMap()


  public static Color3f toColor3f(int colorValue) 
  {
    float r = ((colorValue >> 16) & 0xff) / 255.0f;
    float g = ((colorValue >>  8) & 0xff) / 255.0f;
    float b = ((colorValue      ) & 0xff) / 255.0f;
    // float a = ((colorValue >> 24) & 0xff) / 255.0f;
    return new Color3f(r, g, b);
  }


  public static Color3f toColor3f(int x, int y, int z)
  {
    float r = (float)x / 255.0f;
    float g = (float)y / 255.0f;
    float b = (float)z / 255.0f;
    return new Color3f(r, g, b);
  }


  public static Color3f toColor3f(Color c)
  {
    float r = (c.getRed() / 255.0f);
    float g = (c.getGreen() / 255.0f);
    float b = (c.getBlue() / 255.0f);
    return new Color3f(r, g, b);
  }



  public static Color3f getColor(String colorStr)
  {
    colorStr = colorStr.toLowerCase();

    Color3f ColorFactory = colorMap.get(colorStr);
    if (ColorFactory != null)
      return ColorFactory;

    ColorFactory = toColor3f(Color.WHITE);
    try {  // get a color by using its hex or octal value
      ColorFactory = toColor3f(Color.decode(colorStr));
      colorMap.put(colorStr, ColorFactory);
    } 
    catch (NumberFormatException nfe) {
      System.out.println(colorStr + " unknown; using white");
    }
    return ColorFactory;
  } // end of  getColor()


 
  public static int size()
  {  return colorMap.size();  }


  public static Color3f getColor(int idx)
  {
    if ((idx < 0) || (idx > colorMap.size()-1)) {
      System.out.println("Index out of range; using first ColorFactory");
      idx = 0;
    }
    ArrayList<Color3f> list = new ArrayList<Color3f>(colorMap.values());
    return list.get(idx);
  }


  public static Color3f getRandColor()
  {  return getColor( rand.nextInt(colorMap.size()));  }




  public static String getColorName(Color3f col) 
  {
    for (Entry<String,Color3f> entry : colorMap.entrySet()) {
      if (col.equals(entry.getValue()))
        return entry.getKey();
    }
    System.out.println("Name not found for color " + col);
    return null;
  }


  public static String getColorName(int idx)
  {
    if ((idx < 0) || (idx > colorMap.size()-1)) {
      System.out.println("Index out of range; using first color");
      idx = 0;
    }
    ArrayList<String> list = new ArrayList<String>(colorMap.keySet());
    return list.get(idx);
  }


  public static ArrayList<String> listColors()
  {  return new ArrayList<String>(colorMap.keySet());  }


  public static String getColorNames()
  {  
    ArrayList<String> nms = new ArrayList<String>(colorMap.keySet());
    Collections.sort(nms);
    return String.join(", ", nms);
  }


  // ------------------- test rig -------------------------


  public static void main(String[] args)
  {
   // System.out.println("tomato: " + ColorFactory.getColor("tomato"));
   // System.out.println("#ff6688: " + ColorFactory.getColor("#ff6688"));
   // System.out.println("0x40A8CC: " + ColorFactory.getColor("0x40A8CC"));

    System.out.println("Available Colors: \n" + ColorFactory.getColorNames());
  }  // end of main()


}  // end of ColorFactory class
