
// KeysBehavior.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, October 2021

/* Key-based navigation for the camera:

      Up-arrow or W: Forward
      Down-arrow or S: Backward
      Left-arrow or A: Left
      Right-arrow or D: Right
      Page-up or Q: Up
      Page-down or E: Down
      
      Alt Up-arrow: Rotate up
      Alt Down-arrow: Rotate down
      Alt Left-arrow: Rotate left
      Alt Right-arrow: Rotate Right
      Alt Page-up: Tilt Right
      Alt Page-down: Tilt Left

   Used in conjuction with OrbitBehavior for mouse-based
   camera control.

*/

import java.awt.AWTEvent;
import java.awt.event.*;
import java.util.*;

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;
import org.jogamp.java3d.utils.universe.*;
import org.jogamp.java3d.utils.behaviors.vp.*;



public class KeysBehavior extends ViewPlatformBehavior
{
  private static final double ROT_AMT = Math.PI / 36.0;   // 5 degrees
  private static final double MOVE_STEP = 0.2;


  // hardwired movement vectors
  private static final Vector3d FWD = new Vector3d(0,0,-MOVE_STEP);
  private static final Vector3d BACK = new Vector3d(0,0,MOVE_STEP);
  private static final Vector3d LEFT = new Vector3d(-MOVE_STEP,0,0);
  private static final Vector3d RIGHT = new Vector3d(MOVE_STEP,0,0);
  private static final Vector3d UP = new Vector3d(0,MOVE_STEP,0);
  private static final Vector3d DOWN = new Vector3d(0,-MOVE_STEP,0);

  private static enum Direction {X, Y, Z}


  private TransformGroup viewTG = null;   // the ViewPlatform's TG 

  private WakeupCondition wakeupCond = null;

  // for repeated calcs
  private Transform3D t3d = new Transform3D();
  private Transform3D toMove = new Transform3D();
  private Transform3D toRot = new Transform3D();


  private static Vector3d startPos, startRots;



  public KeysBehavior(TransformGroup tg)
  { 
    viewTG = tg;
    try {
      viewTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
      viewTG.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
    }
    catch(Exception e) { }

	WakeupCriterion[] wakeEvents = new WakeupCriterion[1];
    wakeEvents[0] = new WakeupOnAWTEvent(KeyEvent.KEY_PRESSED);
    wakeupCond = new WakeupOr(wakeEvents);
  } // end of KeysBehavior()


  public void initialize()
  {
    startPos = getPosition();
    startRots = getRotations();
	wakeupOn(wakeupCond);
  }  // end of initialize()



  public void processStimulus(Iterator<WakeupCriterion> wucs)
  // respond to a keypress
  {
    WakeupCriterion wakeup;
    AWTEvent[] event;

    while(wucs.hasNext()) {
      wakeup = wucs.next();
      if( wakeup instanceof WakeupOnAWTEvent ) {
        event = ((WakeupOnAWTEvent)wakeup).getAWTEvent();
        for( int i = 0; i < event.length; i++ ) {
          if(event[i].getID() == KeyEvent.KEY_PRESSED)
            processKeyEvent((KeyEvent)event[i]);
        }
      }
    }
	wakeupOn(wakeupCond);
  } // end of processStimulus()


  private void processKeyEvent(KeyEvent eventKey)
  {
    int keyCode = eventKey.getKeyCode();
    // System.out.println(keyCode);

    if(eventKey.isAltDown())   // key + <alt>
      doRotate(keyCode);
    else
      doMove(keyCode);
  } // end of processKeyEvent()



  private void doMove(int keycode)
  /* Move viewer forward or backward, up or down, 
     left or right */
  { 
    if ((keycode == KeyEvent.VK_UP) || 
        (keycode == KeyEvent.VK_W))
      move(FWD);
    else if ((keycode == KeyEvent.VK_DOWN) || 
             (keycode == KeyEvent.VK_S))
      move(BACK);
    else if ((keycode == KeyEvent.VK_LEFT) || 
             (keycode == KeyEvent.VK_A))
      move(LEFT);
    else if ((keycode == KeyEvent.VK_RIGHT) || 
             (keycode == KeyEvent.VK_D))
      move(RIGHT);
    else if ((keycode == KeyEvent.VK_PAGE_UP) || 
             (keycode == KeyEvent.VK_Q))
      move(UP);
    else if ((keycode == KeyEvent.VK_PAGE_DOWN) || 
             (keycode == KeyEvent.VK_E))
      move(DOWN);
  } // end of doMove()


  private void move(Vector3d theMove)
  {
    viewTG.getTransform(t3d);

    toMove.setTranslation(theMove);
    t3d.mul(toMove);
    viewTG.setTransform(t3d);
  } // end of move()



  private void doRotate(int keycode)
  /* Rotate viewer forward or backward, up or down, 
     left or right */
  {
    if ((keycode == KeyEvent.VK_UP) || 
        (keycode == KeyEvent.VK_W))
      rotate(ROT_AMT, Direction.X);
    else if ((keycode == KeyEvent.VK_DOWN) || 
             (keycode == KeyEvent.VK_S))
      rotate(-ROT_AMT, Direction.X);
    else if ((keycode == KeyEvent.VK_LEFT) || 
             (keycode == KeyEvent.VK_A))
      rotate(ROT_AMT, Direction.Y);
    else if ((keycode == KeyEvent.VK_RIGHT) || 
             (keycode == KeyEvent.VK_D))
      rotate(-ROT_AMT, Direction.Y);
    else if ((keycode == KeyEvent.VK_PAGE_UP) || 
             (keycode == KeyEvent.VK_Q))
      rotate(ROT_AMT, Direction.Z);
    else if ((keycode == KeyEvent.VK_PAGE_DOWN) || 
             (keycode == KeyEvent.VK_E))
      rotate(-ROT_AMT, Direction.Z);
  }  // end of doRotate()


  private void rotate(double radians, Direction d)
  { 
    viewTG.getTransform(t3d);   

    if (d == Direction.X)
      toRot.rotX(radians);
    else if (d == Direction.Y)
      toRot.rotY(radians);
    else  // z direction
      toRot.rotZ(radians);

    t3d.mul(toRot);
    viewTG.setTransform(t3d);
  } // end of rotate()



  public void reset()
  {
    moveTo(startPos);
    setRotations(startRots);
  } // end of reset()



  // position ---------------------------


  public void moveTo(double x, double y, double z)
  {  moveTo(new Vector3d(x,y,z));  }


  public void moveTo(Vector3d pos)
  {  Simple3D.moveTo(viewTG, pos);  }



  public Vector3d getPosition()
  {  return Simple3D.getPosition(viewTG);  }




  // get/set rotations -----------------------------


  public void setRotations(double xDegree, double yDegree, 
                                           double zDegree)
  {  setRotations( new Vector3d(xDegree, yDegree, zDegree));  }



  public void setRotations(Vector3d angles)
  // use the given Euler angles (in degrees)
  {
    angles.scale(Math.PI/180);    // convert to radians

    viewTG.getTransform(t3d);  

    Transform3D t1 = new Transform3D();
    t1.setEuler(angles);

    Vector3d pos = new Vector3d();
    t3d.get(pos);

    t1.setTranslation(pos);
    t1.setScale(t3d.getScale());

    viewTG.setTransform(t1);
  }  // end of setRotations()



  public Vector3d getRotations()
  {  return Simple3D.getRotations(viewTG);  }


}  // end of KeysBehavior class
