
// MyMouseTranslate.java
// Sun Microsystems, Inc., 2007 
// Modified by Andrew Davison, ad@fivedots.coe.psu.ac.th, Dec. 2021


/* Modified version of Java 3D's MouseTranslate to respond based
   on which button is pressed rather than on keycodes.

   Let users control the
   translation (X, Y) of an object via a mouse drag motion with the first
   mouse button. The default is to translate along the x- and y- axes at
   the same time, but if the isControlled flag is set then translation will
   only be along the x-axis. It's necessary to press
   down <CTRL> to switch movement along the y-axis.
*/


import java.awt.*;
import java.awt.event.*;
import java.util.*;

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;

import org.jogamp.java3d.utils.behaviors.mouse.*;



public class MyMouseTranslate extends MouseBehavior
{
  private static double STEP = 0.01;  
      // move in increments of 0.01 units


  private double xFactor = STEP;
  private double yFactor = STEP;
  private Vector3d translation = new Vector3d();
  private boolean isControlled = false;

  private MouseBehaviorCallback callback = null;



  public MyMouseTranslate(TransformGroup tg)
  {  super(tg);   } 

  public MyMouseTranslate()
  {  super(0);   } 

  public MyMouseTranslate(int flags)
  {  super(flags);   } 

  public MyMouseTranslate(Component c)
  {  super(c, 0);   } 

  public MyMouseTranslate(Component c, TransformGroup tg)
  {  super(c, tg);   } 

  public MyMouseTranslate(Component c, int flags)
  {  super(c, flags);   } 



  @Override
  public void initialize()
  {
    super.initialize();
    if ((flags & INVERT_INPUT) == INVERT_INPUT) {
      invert = true;
      xFactor *= -1;
      yFactor *= -1;
    }
  } /* initialize */



  public double getXFactor()
  {  return xFactor;   } 

  public double getYFactor()
  {  return yFactor;   } 

  public void setFactor(double f)
  {  xFactor = yFactor = f;   } 

  public void setFactor(double xf, double yf)
  {  xFactor = xf;  yFactor = yf;  }

  public void setControl(boolean b)
  {  isControlled = b;   } 

  public boolean setControl()
  {  return isControlled;  }



  @Override
  public void processStimulus (Iterator<WakeupCriterion> criteria)
  {
    WakeupCriterion wakeup;
    AWTEvent[] events;
    MouseEvent evt;

    while (criteria.hasNext()) {
      wakeup = criteria.next();

      if (wakeup instanceof WakeupOnAWTEvent) {
        events = ((WakeupOnAWTEvent)wakeup).getAWTEvent();
        if (events.length > 0) {
          evt = (MouseEvent) events[events.length-1];
          doProcess(evt);
        }
      }

      else if (wakeup instanceof WakeupOnBehaviorPost)
        while (true) {
          // access to the queue must be synchronized
          synchronized (mouseq) {
            if (mouseq.isEmpty()) 
              break;

            evt = (MouseEvent)mouseq.remove(0);
            // consolidate MOUSE_DRAG events
            while ((evt.getID() == MouseEvent.MOUSE_DRAGGED) &&
                   !mouseq.isEmpty() &&
                   (((MouseEvent)mouseq.get(0)).getID() ==
                      MouseEvent.MOUSE_DRAGGED))
              evt = (MouseEvent)mouseq.remove(0);
          }
          doProcess(evt);
        }
    }
    wakeupOn(mouseCriterion);
  } /* processStimulus */



  private void doProcess(MouseEvent evt)
  {
    int id;
    int dx = 0;
    int dy = 0;

    processMouseEvent(evt);

    if (((buttonPress) && ((flags & MANUAL_WAKEUP) == 0)) ||
        ((wakeUp) && ((flags & MANUAL_WAKEUP) != 0))) {
      id = evt.getID();

      if ((id == MouseEvent.MOUSE_DRAGGED) &&
          ((evt.getModifiersEx() & InputEvent.BUTTON1_DOWN_MASK) != 0)) {      // CHANGED
         // !evt.isAltDown() && evt.isControlDown()) {  // CHANGED
          // !evt.isAltDown() && evt.isMetaDown()) {
        x = evt.getX();
        y = evt.getY();

        if (isControlled) {     // CHANGED
          if (evt.isControlDown()) 
            dy = y - y_last;
          else
            dx = x - x_last;
        }
        else {
          dy = y - y_last;
          dx = x - x_last;
        }

        if ((!reset) && ((Math.abs(dy) < 50) && 
                         (Math.abs(dx) < 50))) {
          // System.out.println("dx " + dx + " dy " + dy);
          transformGroup.getTransform(currXform);

          translation.x = dx * xFactor;
          translation.y = -dy * yFactor;
          transformX.set(translation);

          if (invert)
            currXform.mul(currXform, transformX);
          else
            currXform.mul(transformX, currXform);

          transformGroup.setTransform(currXform);
          transformChanged(currXform);

          if (callback != null)
            callback.transformChanged(
                      MouseBehaviorCallback.TRANSLATE, currXform);
               // 2-arg version of transformChanged() in pick behavior
        }
        else
          reset = false;

        x_last = x;
        y_last = y;
      }
      else if (id == MouseEvent.MOUSE_PRESSED) {
        x_last = evt.getX();
        y_last = evt.getY();
      }
    }
  } /* doProcess */



  /*
     Users can overload this 1-arg version of transformChanged()
     which is called every time the Behavior updates the transform.
     Default implementation does nothing
   */
  public void transformChanged(Transform3D transform) 
  { }


  public void setupCallback(MouseBehaviorCallback callback)
  {  this.callback = callback;   } 

} // end of MyMouseTranslate class

