
// OBJWriter.java
// Emmanuel PUYBARET, 18 Sept. 2008

// Reformatted for Java 3D 1.7, Andrew Davison, Oct 2021
// Removed uses of OBJMaterial

/*
   Sweet Home 3D, Copyright (c) 2008 Emmanuel PUYBARET / eTeks <info@eteks.com>

   An output stream that writes Java 3D nodes in OBJ + MTL format.
   
   Usage:

    try {
      OBJWriter ow = new OBJWriter("MyObject.obj"); 
      ow.writeNode(java3DNode1); 
      ow.writeNode(java3DNode2); 
       ... 
      ow.close();
    }
    catch(IOException e)
    {  System.out.println(e);  }


   Writes all the 3D shapes children of node.

   If there are transformation groups on the path from node to its shapes,
   they'll be applied to the coordinates written on output.

   The node shouldn't be alive or if it's alive it should have the
   capabilities to read its children, the geometries and the app of its shapes.

   Only geometries which are instances of GeometryArray will be written.
*/



import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;
import java.util.zip.*;

import java.awt.image.RenderedImage;
import javax.imageio.*;
import javax.imageio.stream.ImageInputStream;


// Java3D 1.7 changes package names

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;



public class OBJWriter extends FilterWriter 
{
  private final NumberFormat defaultNF =
      new DecimalFormat("0.#######", 
                      new DecimalFormatSymbols(java.util.Locale.US));
  private final NumberFormat numFormat;
  private final String  header;

  private boolean firstNode = true;
  private String  mtlFnm;

  private int shapeIndex = 1;

  private Map<Point3f, Integer> vertIdxs = new HashMap<Point3f, Integer>();
  private Map<Vector3f, Integer> normalIdxs = new HashMap<Vector3f, Integer>();
  private Map<TexCoord2f, Integer> texCoordIdxs = 
                                     new HashMap<TexCoord2f, Integer>();

  private Map<ComparableAppearance, String> apps =
      new LinkedHashMap<ComparableAppearance, String>();

  private Map<Texture, File> textures = new HashMap<Texture, File>();
  private List<URL> copiedTextures = new ArrayList<URL>();



  public OBJWriter(File objFile) 
            throws FileNotFoundException, IOException 
  {
    this(objFile, null, -1);
  }


  /**
Create an OBJ writer for the given file.
With a header comment at start of the OBJ file and its MTL counterpart
   */
  public OBJWriter(File objFile, String header, int maxFracDigits) 
                throws FileNotFoundException, IOException 
  {
    this(objFile.toString(), header, maxFracDigits);
  }



  public OBJWriter(String objFnm) 
               throws FileNotFoundException, IOException 
  {
    this(objFnm, null, -1);
  }



  public OBJWriter(String objFnm, String header, int maxFracDigits) 
                        throws FileNotFoundException, IOException 
  {
    this(new FileOutputStream(objFnm), header, maxFracDigits);
    System.out.println("Writing to " + objFnm);
    
    if (objFnm.toLowerCase().endsWith(".obj"))
      mtlFnm = objFnm.substring(0, objFnm.length() - 4) + ".mtl";
    else
      mtlFnm = objFnm + ".mtl";

    // Remove spaces in MTL file name
    mtlFnm = new File(new File(mtlFnm).getParent(),
        new File(mtlFnm).getName().replace(' ', '_')).toString();

    // Ensure MTL file is using only ASCII codes
    String name = new File(mtlFnm).getName();
    for (int i = 0; i < name.length(); i++) {
      if (name.charAt(i) >= 128) {
        mtlFnm = new File(new File(mtlFnm).getParent(),
            "mats.mtl").toString();
        break;
      }
    }
  }


  public OBJWriter(OutputStream out) throws IOException 
  {
    this(out, null, -1);
  }



  public OBJWriter(OutputStream out, String header,
                   int maxFracDigits) throws IOException 
  {
    this(new OutputStreamWriter(
                new BufferedOutputStream(out), "US-ASCII"), 
         header, maxFracDigits);
  }



  public OBJWriter(Writer out) throws IOException 
  {
    this(out, null, -1);
  }



  public OBJWriter(Writer out, String header,
                   int maxFracDigits) throws IOException 
  {
    super(out);
    if (maxFracDigits >= 0) {
      numFormat = NumberFormat.getNumberInstance(java.util.Locale.US);
      numFormat.setMinimumFractionDigits(0);
      numFormat.setMaximumFractionDigits(maxFracDigits);
    } 
    else
      numFormat = null;
    this.header = header;
    writeHeader(out);
  }



  private void writeHeader(Writer writer) throws IOException 
  {
    if (header != null) {
      if (!header.startsWith("#"))
        writer.write("# ");
      writer.write(header.replace("\n", "\n# "));
      writer.write("\n");
    }
  }


  // Write a single character in a comment
  @Override
  public void write(int c) throws IOException 
  {
    out.write("# ");
    out.write(c);
    out.write("\n");
  }


  // Write a portion of an array of characters in a comment
  @Override
  public void write(char cbuf[], int off, int len) throws IOException 
  {
    out.write("# ");
    out.write(cbuf, off, len);
    out.write("\n");
  }


  // Write a portion of a string in a comment
  @Override
  public void write(String str, int off, int len) throws IOException 
  {
    out.write("# ");
    out.write(str, off, len);
    out.write("\n");
  }


  // Write a string in a comment=
  @Override
  public void write(String str) throws IOException 
  {
    out.write("# ");
    out.write(str, 0, str.length());
    out.write("\n");
  }



  private void checkCurrentThreadIsntInterrupted() 
                             throws InterruptedIOException 
  {
    if (Thread.interrupted()) {
      mtlFnm = null;
      throw new InterruptedIOException("Current thread interrupted");
    }
  }


  // Writes all the 3D shapes children of node
  public void writeNode(Node node) 
                    throws IOException, InterruptedIOException 
  {
    writeNode(node, null);
  }



  public void writeNode(Node node, String nodeName) 
                         throws IOException, InterruptedIOException 
  {
    if (firstNode) {
      if (mtlFnm != null)
        out.write("mtllib " + new File(mtlFnm).getName() + "\n");
      firstNode = false;
    }
    writeNode(node, nodeName, new Transform3D());
  }



  private void writeNode(Node node, String nodeName, Transform3D parentT3d) 
                                            throws IOException 
  {
    if (node instanceof Group) {
      if (node instanceof TransformGroup) {
        parentT3d = new Transform3D(parentT3d);
        Transform3D t3d = new Transform3D();
        ((TransformGroup)node).getTransform(t3d);
        parentT3d.mul(t3d);
      }

      // Write all children
      Iterator<Node> kids = ((Group)node).getAllChildren();
      while (kids.hasNext()) {
        writeNode( kids.next(), nodeName, parentT3d);
      }
    } 
    else if (node instanceof Link) {
      writeNode(((Link)node).getSharedGroup(), nodeName, parentT3d);
    } 
    else if (node instanceof Shape3D) {
      Shape3D shape = (Shape3D)node;
      Appearance app = shape.getAppearance();
      RenderingAttributes ra = app != null
          ? app.getRenderingAttributes() : null;
      if (shape.numGeometries() >= 1
          && (ra == null || ra.getVisible())) {
        // Build a unique human readable object name
        String objNm = "";
        if (accept(nodeName)) {
          objNm = nodeName + "_";
        }

        String shapeName = null;
        if (shape.getUserData() instanceof String)
          shapeName = (String)shape.getUserData();
        if (accept(shapeName))
          objNm += shapeName + "_";

        objNm += String.valueOf(shapeIndex++);

        // Start a new object at OBJ format
        out.write("g " + objNm + "\n");

        TexCoordGeneration tcGen = null;
        Transform3D texT3d = new Transform3D();
        if (mtlFnm != null) {
          if (app != null) {
            tcGen = app.getTexCoordGeneration();
            TextureAttributes ta = app.getTextureAttributes();
            if (ta != null) {
              ta.getTextureTransform(texT3d);
            }
            ComparableAppearance compareApp = new ComparableAppearance(app);
            String appNm = apps.get(compareApp);
            if (appNm == null) {
              // Store app
              try {
                appNm = app.getName();
              } 
              catch (NoSuchMethodError ex) {
                // Don't reuse app name with Java 3D < 1.4 where getName was added
              }

              if (appNm == null || !accept(appNm)) {
                appNm = objNm;
              } 
              else {
                // Find a unique app name among apps
                Collection<String> appearanceNames = apps.values();
                String baseName = appNm + "_" + objNm;
                for (int i = 0; appearanceNames.contains(appNm); i++) {
                  if (i == 0)
                    appNm = baseName;
                  else
                    appNm = baseName + "_" + i;
                }
              }
              apps.put(compareApp, appNm);

              Texture tex = app.getTexture();
              if (tex != null) {
                File textureFile = textures.get(tex);
                if (textureFile == null) {
                  String fileExtension = "png";
                  URL textureUrl = (URL)tex.getUserData();
                  if (textureUrl instanceof URL) {
                    InputStream in = null;
                    try {
                      // Find the format of the tex image
                      in = openStream(textureUrl);
                      ImageInputStream imageIn = ImageIO.createImageInputStream(in);
                      Iterator<ImageReader> imageReaders = ImageIO.getImageReaders(imageIn);
                      if (imageReaders.hasNext()) {
                        ImageReader reader = (ImageReader)imageReaders.next();
                        fileExtension = reader.getFormatName().toLowerCase();
                        // Store the URL to copy its image content directly when apps are saved
                        copiedTextures.add(textureUrl);
                      }
                    } 
                    catch (IOException ex) {
                      if (in != null) {
                        in.close();
                      }
                    }
                  }

                  // Find a unique tex file name which is not case sensitive
                  String textureFileBaseName = mtlFnm.substring(0, mtlFnm.length() - 4)
                      + "_" + appNm;
                  Collection<File> textureFiles = textures.values();
                  boolean fileExists = true;
                  for (int i = 0; fileExists; i++) {
                    if (i == 0) {
                      textureFile = new File(textureFileBaseName + "." + fileExtension);
                    } else {
                      textureFile = new File(textureFileBaseName + "_" + i + "." + fileExtension);
                    }

                    fileExists = false;
                    for (File file : textureFiles) {
                      if (textureFile.getName().equalsIgnoreCase(file.getName())) {
                        fileExists = true;
                        break;
                      }
                    }
                  }
                  // Store texture with associated file
                  System.out.println("Saving texture to " + textureFile.getName());
                  textures.put(tex, textureFile);
                }
              }
            }
            out.write("usemtl " + appNm + "\n");
          }
        }

        int cullFace = PolygonAttributes.CULL_BACK;
        boolean backFaceNormalFlip = false;
        if (app != null) {
          PolygonAttributes pa = app.getPolygonAttributes();
          if (pa != null) {
            cullFace = pa.getCullFace();
            backFaceNormalFlip = pa.getBackFaceNormalFlip();
          }
        }

        // Write object geometries
        for (int i = 0, n = shape.numGeometries(); i < n; i++) {
          writeNodeGeometry(shape.getGeometry(i), parentT3d, tcGen,
              texT3d, cullFace, backFaceNormalFlip);
        }
      }
    }
  }  // end of writeNode()



  private InputStream openStream(URL url) throws IOException 
  {
    URLConnection conn = url.openConnection();
    if (System.getProperty("os.name").startsWith("Windows")
        && (conn instanceof JarURLConnection)) {
      JarURLConnection urlConnection = (JarURLConnection)conn;
      URL jarFileUrl = urlConnection.getJarFileURL();
      if (jarFileUrl.getProtocol().equalsIgnoreCase("file")) {
        try {
          File file;
          try {
            file = new File(jarFileUrl.toURI());
          } 
          catch (IllegalArgumentException ex) {
            // Try a second way to be able to access to files on Windows servers
            file = new File(jarFileUrl.getPath());
          }
          if (file.canWrite()) {
            // Refuse to use caches to be able to delete the writable files accessed with jar protocol under Windows,
            // as suggested in http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6962459
            conn.setUseCaches(false);
          }
        } 
        catch (URISyntaxException ex) {
          IOException ex2 = new IOException();
          ex2.initCause(ex);
          throw ex2;
        }
      }
    }
    return conn.getInputStream();
  }  // end of openStream()



  // Returns true if name contains only letters, digits and underscores.
  private boolean accept(String name) 
  {
    if (name == null)
      return false;

    for (int i = 0; i < name.length(); i++) {
      char car = name.charAt(i);
      if (!(car >= 'a' && car <= 'z'
            || car >= 'A' && car <= 'Z'
            || car >= '0' && car <= '9'
            || car == '_')) {
        return false;
      }
    }
    return true;
  }


  // Writes a 3D geometry 
  private void writeNodeGeometry(Geometry geometry,
                                 Transform3D parentT3d,
                                 TexCoordGeneration tcGen,
                                 Transform3D texT3d,
                                 int cullFace,
                                 boolean backFaceNormalFlip) throws IOException 
  {
    if (geometry instanceof GeometryArray) {
      GeometryArray geomArr = (GeometryArray)geometry;

      int[] vertIdxSubs = new int[geomArr.getVertexCount()];

      boolean normalsDefined = (geomArr.getVertexFormat() & GeometryArray.NORMALS) != 0;
      StringBuilder normalsBuffer;
      List<Vector3f> addedNormals;
      if (normalsDefined) {
        normalsBuffer = new StringBuilder(geomArr.getVertexCount() * 3 * 10);
        addedNormals = new ArrayList<Vector3f>();
      } 
      else {
        normalsBuffer = null;
        addedNormals = null;
      }

      int[] normalIndexSubs = new int [geomArr.getVertexCount()];
      int[] oppositeSideNormalIndexSubs;
      if (cullFace == PolygonAttributes.CULL_NONE) {
        oppositeSideNormalIndexSubs = new int[geomArr.getVertexCount()];
      } 
      else {
        oppositeSideNormalIndexSubs = null;
      }

      boolean haveTexCoords = (geomArr.getVertexFormat() & GeometryArray.TEXTURE_COORDINATE_2) != 0;
      int[] texCoordsIdxSubs = new int[geomArr.getVertexCount()];

      boolean texCoordsGenerated = false;
      Vector4f planeS = null;
      Vector4f planeT = null;
      if (tcGen != null) {
        texCoordsGenerated = tcGen.getGenMode() == TexCoordGeneration.OBJECT_LINEAR
            && tcGen.getEnable()
            && !(geomArr instanceof IndexedLineArray)
            && !(geomArr instanceof IndexedLineStripArray)
            && !(geomArr instanceof LineArray)
            && !(geomArr instanceof LineStripArray);
        if (texCoordsGenerated) {
          planeS = new Vector4f();
          planeT = new Vector4f();
          tcGen.getPlaneS(planeS);
          tcGen.getPlaneT(planeT);
        }
      }

      checkCurrentThreadIsntInterrupted();

      if ((geomArr.getVertexFormat() & GeometryArray.BY_REFERENCE) != 0) {
        if ((geomArr.getVertexFormat() & GeometryArray.INTERLEAVED) != 0) {
          float [] vertexData = geomArr.getInterleavedVertices();
          int vertexSize = vertexData.length / geomArr.getVertexCount();

          // Write vertices coordinates
          for (int index = 0, i = vertexSize - 3, n = geomArr.getVertexCount();
               index < n; index++, i += vertexSize) {
            Point3f vertex = new Point3f(vertexData [i], vertexData [i + 1], vertexData [i + 2]);
            writeVertex(parentT3d, vertex, index, vertIdxSubs);
          }

          // Write tex coordinates
          if (tcGen != null) {
            if (texCoordsGenerated) {
              for (int index = 0, i = vertexSize - 3, n = geomArr.getVertexCount();
                    index < n; index++, i += vertexSize) {
                TexCoord2f texCoords = genTexCoords(
                    vertexData [i], vertexData [i + 1], vertexData [i + 2], planeS, planeT);
                writeTexCoords(texCoords, texT3d, index, texCoordsIdxSubs);
              }
            }
          } 
          else if (haveTexCoords) {
            for (int index = 0, i = 0, n = geomArr.getVertexCount();
                  index < n; index++, i += vertexSize) {
              TexCoord2f texCoords = new TexCoord2f(vertexData [i], vertexData [i + 1]);
              writeTexCoords(texCoords, texT3d, index, texCoordsIdxSubs);
            }
          }

          // Write normals
          if (normalsDefined) {
            for (int index = 0, i = vertexSize - 6, n = geomArr.getVertexCount();
                 normalsDefined && index < n; index++, i += vertexSize) {
              Vector3f normal = new Vector3f(vertexData [i], vertexData [i + 1], vertexData [i + 2]);
              normalsDefined = writeNormal(normalsBuffer, parentT3d, normal, index, normalIndexSubs,
                  oppositeSideNormalIndexSubs, addedNormals, cullFace, backFaceNormalFlip);
            }
          }
        } 
        else {
          // Write vertices coordinates
          float [] vertCoords = geomArr.getCoordRefFloat();
          for (int index = 0, i = 0, n = geomArr.getVertexCount(); index < n; index++, i += 3) {
            Point3f vertex = new Point3f(vertCoords [i], vertCoords [i + 1], vertCoords [i + 2]);
            writeVertex(parentT3d, vertex, index,
                vertIdxSubs);
          }

          // Write tex coordinates
          if (tcGen != null) {
            if (texCoordsGenerated) {
              for (int index = 0, i = 0, n = geomArr.getVertexCount(); index < n; index++, i += 3) {
                TexCoord2f texCoords = genTexCoords(
                    vertCoords [i], vertCoords [i + 1], vertCoords [i + 2], planeS, planeT);
                writeTexCoords(texCoords, texT3d, index, texCoordsIdxSubs);
              }
            }
          } 
          else if (haveTexCoords) {
            float [] textureCoordinatesArray = geomArr.getTexCoordRefFloat(0);
            for (int index = 0, i = 0, n = geomArr.getVertexCount(); index < n; index++, i += 2) {
              TexCoord2f texCoords = new TexCoord2f(textureCoordinatesArray [i], textureCoordinatesArray [i + 1]);
              writeTexCoords(texCoords, texT3d, index, texCoordsIdxSubs);
            }
          }

          // Write normals
          if (normalsDefined) {
            float [] normalCoordinates = geomArr.getNormalRefFloat();
            for (int index = 0, i = 0, n = geomArr.getVertexCount(); normalsDefined && index < n; index++, i += 3) {
              Vector3f normal = new Vector3f(normalCoordinates [i], normalCoordinates [i + 1], normalCoordinates [i + 2]);
              normalsDefined = writeNormal(normalsBuffer, parentT3d, normal, index, normalIndexSubs,
                  oppositeSideNormalIndexSubs, addedNormals, cullFace, backFaceNormalFlip);
            }
          }
        }
      } 
      else {
        // Write vertices coordinates
        for (int index = 0, n = geomArr.getVertexCount(); index < n; index++) {
          Point3f vertex = new Point3f();
          geomArr.getCoordinate(index, vertex);
          writeVertex(parentT3d, vertex, index,
              vertIdxSubs);
        }

        // Write tex coordinates
        if (tcGen != null) {
          if (texCoordsGenerated) {
            for (int index = 0, n = geomArr.getVertexCount(); index < n; index++) {
              Point3f vertex = new Point3f();
              geomArr.getCoordinate(index, vertex);
              TexCoord2f texCoords = genTexCoords(
                  vertex.x, vertex.y, vertex.z, planeS, planeT);
              writeTexCoords(texCoords, texT3d, index, texCoordsIdxSubs);
            }
          }
        } 
        else if (haveTexCoords) {
          for (int index = 0, n = geomArr.getVertexCount(); index < n; index++) {
            TexCoord2f texCoords = new TexCoord2f();
            geomArr.getTextureCoordinate(0, index, texCoords);
            writeTexCoords(texCoords, texT3d, index, texCoordsIdxSubs);
          }
        }

        // Write normals
        if (normalsDefined) {
          for (int index = 0, n = geomArr.getVertexCount(); normalsDefined && index < n; index++) {
            Vector3f normal = new Vector3f();
            geomArr.getNormal(index, normal);
            normalsDefined = writeNormal(normalsBuffer, parentT3d, normal, index, normalIndexSubs,
                oppositeSideNormalIndexSubs, addedNormals, cullFace, backFaceNormalFlip);
          }
        }
      }

      if (normalsDefined) {
        // Write normals only if they all contain valid values
        out.write(normalsBuffer.toString());
      } 
      else if (addedNormals != null) {
        // Remove ignored normals
        for (Vector3f normal : addedNormals) {
          normalIdxs.remove(normal);
        }
      }

      checkCurrentThreadIsntInterrupted();

      // Write lines, triangles or quadrilaterals depending on the geometry
      if (geomArr instanceof IndexedGeometryArray) {
        if (geomArr instanceof IndexedLineArray) {
          IndexedLineArray lineArray = (IndexedLineArray)geomArr;
          for (int i = 0, n = lineArray.getIndexCount(); i < n; i += 2) {
            writeIndexedLine(lineArray, i, i + 1, vertIdxSubs, texCoordsIdxSubs);
          }
        } 
        else if (geomArr instanceof IndexedTriangleArray) {
          IndexedTriangleArray triArr = (IndexedTriangleArray)geomArr;
          for (int i = 0, n = triArr.getIndexCount(); i < n; i += 3) {
            writeIndexedTriangle(triArr, i, i + 1, i + 2,
                vertIdxSubs, normalIndexSubs, oppositeSideNormalIndexSubs,
                normalsDefined, texCoordsIdxSubs, texCoordsGenerated, cullFace);
          }
        } 
        else if (geomArr instanceof IndexedQuadArray) {
          IndexedQuadArray quadArray = (IndexedQuadArray)geomArr;
          for (int i = 0, n = quadArray.getIndexCount(); i < n; i += 4) {
            writeIndexedQuad(quadArray, i, i + 1, i + 2, i + 3,
                vertIdxSubs, normalIndexSubs, oppositeSideNormalIndexSubs,
                normalsDefined, texCoordsIdxSubs, texCoordsGenerated, cullFace);
          }
        } 
        else if (geomArr instanceof IndexedGeometryStripArray) {
          IndexedGeometryStripArray geometryStripArray = (IndexedGeometryStripArray)geomArr;
          int[] stripIndexCounts = new int [geometryStripArray.getNumStrips()];
          geometryStripArray.getStripIndexCounts(stripIndexCounts);
          int initialIndex = 0;

          if (geometryStripArray instanceof IndexedLineStripArray) {
            for (int strip = 0; strip < stripIndexCounts.length; strip++) {
              for (int i = initialIndex, n = initialIndex + stripIndexCounts [strip] - 1; i < n; i++) {
                writeIndexedLine(geometryStripArray, i, i + 1,
                    vertIdxSubs, texCoordsIdxSubs);
              }
              initialIndex += stripIndexCounts [strip];
            }
          } 
          else if (geometryStripArray instanceof IndexedTriangleStripArray) {
            for (int strip = 0; strip < stripIndexCounts.length; strip++) {
              for (int i = initialIndex, n = initialIndex + stripIndexCounts [strip] - 2, j = 0; i < n; i++, j++) {
                if (j % 2 == 0) {
                  writeIndexedTriangle(geometryStripArray, i, i + 1, i + 2,
                      vertIdxSubs, normalIndexSubs, oppositeSideNormalIndexSubs,
                      normalsDefined, texCoordsIdxSubs, texCoordsGenerated, cullFace);
                } else { // Vertices of odd triangles are in reverse order
                  writeIndexedTriangle(geometryStripArray, i, i + 2, i + 1,
                      vertIdxSubs, normalIndexSubs, oppositeSideNormalIndexSubs,
                      normalsDefined, texCoordsIdxSubs, texCoordsGenerated, cullFace);
                }
              }
              initialIndex += stripIndexCounts [strip];
            }
          } 
          else if (geometryStripArray instanceof IndexedTriangleFanArray) {
            for (int strip = 0; strip < stripIndexCounts.length; strip++) {
              for (int i = initialIndex, n = initialIndex + stripIndexCounts [strip] - 2; i < n; i++) {
                writeIndexedTriangle(geometryStripArray, initialIndex, i + 1, i + 2,
                    vertIdxSubs, normalIndexSubs, oppositeSideNormalIndexSubs,
                    normalsDefined, texCoordsIdxSubs, texCoordsGenerated, cullFace);
              }
              initialIndex += stripIndexCounts [strip];
            }
          }
        }
      } 
      else {
        if (geomArr instanceof LineArray) {
          LineArray lineArray = (LineArray)geomArr;
          for (int i = 0, n = lineArray.getVertexCount(); i < n; i += 2) {
            writeLine(lineArray, i, i + 1, vertIdxSubs, texCoordsIdxSubs);
          }
        } 
        else if (geomArr instanceof TriangleArray) {
          TriangleArray triArr = (TriangleArray)geomArr;
          for (int i = 0, n = triArr.getVertexCount(); i < n; i += 3) {
            writeTriangle(triArr, i, i + 1, i + 2,
                vertIdxSubs, normalIndexSubs, oppositeSideNormalIndexSubs,
                normalsDefined, texCoordsIdxSubs, texCoordsGenerated, cullFace);
          }
        } 
        else if (geomArr instanceof QuadArray) {
          QuadArray quadArray = (QuadArray)geomArr;
          for (int i = 0, n = quadArray.getVertexCount(); i < n; i += 4) {
            writeQuad(quadArray, i, i + 1, i + 2, i + 3,
                vertIdxSubs, normalIndexSubs, oppositeSideNormalIndexSubs,
                normalsDefined, texCoordsIdxSubs, texCoordsGenerated, cullFace);
          }
        } 
        else if (geomArr instanceof GeometryStripArray) {
          GeometryStripArray geometryStripArray = (GeometryStripArray)geomArr;
          int[] stripVertexCounts = new int [geometryStripArray.getNumStrips()];
          geometryStripArray.getStripVertexCounts(stripVertexCounts);
          int initialIndex = 0;

          if (geometryStripArray instanceof LineStripArray) {
            for (int strip = 0; strip < stripVertexCounts.length; strip++) {
              for (int i = initialIndex, n = initialIndex + stripVertexCounts [strip] - 1; i < n; i++) {
                writeLine(geometryStripArray, i, i + 1, vertIdxSubs, texCoordsIdxSubs);
              }
              initialIndex += stripVertexCounts [strip];
            }
          } 
          else if (geometryStripArray instanceof TriangleStripArray) {
            for (int strip = 0; strip < stripVertexCounts.length; strip++) {
              for (int i = initialIndex, n = initialIndex + stripVertexCounts [strip] - 2, j = 0; i < n; i++, j++) {
                if (j % 2 == 0) {
                  writeTriangle(geometryStripArray, i, i + 1, i + 2,
                      vertIdxSubs, normalIndexSubs, oppositeSideNormalIndexSubs,
                      normalsDefined, texCoordsIdxSubs, texCoordsGenerated, cullFace);
                } else { // Vertices of odd triangles are in reverse order
                  writeTriangle(geometryStripArray, i, i + 2, i + 1,
                      vertIdxSubs, normalIndexSubs, oppositeSideNormalIndexSubs,
                      normalsDefined, texCoordsIdxSubs, texCoordsGenerated, cullFace);
                }
              }
              initialIndex += stripVertexCounts [strip];
            }
          } 
          else if (geometryStripArray instanceof TriangleFanArray) {
            for (int strip = 0; strip < stripVertexCounts.length; strip++) {
              for (int i = initialIndex, n = initialIndex + stripVertexCounts [strip] - 2; i < n; i++) {
                writeTriangle(geometryStripArray, initialIndex, i + 1, i + 2,
                    vertIdxSubs, normalIndexSubs, oppositeSideNormalIndexSubs,
                    normalsDefined, texCoordsIdxSubs, texCoordsGenerated, cullFace);
              }
              initialIndex += stripVertexCounts [strip];
            }
          }
        }
      }
    }
  } // end of writeNodeGeometry()



  // Returns tex coordinates generated with tcGen
  private TexCoord2f genTexCoords(float x, float y, float z,
                    Vector4f planeS, Vector4f planeT) 
  {
    return new TexCoord2f(
        x * planeS.x + y * planeS.y + z * planeS.z + planeS.w,
        x * planeT.x + y * planeT.y + z * planeT.z + planeT.w);
  }



  /**
Applies to vertex the given transformation, and writes it in
a line v at OBJ format, if the vertex wasn't written yet.
   */
  private void writeVertex(Transform3D transformationToParent,
                           Point3f vertex, int index,
                           int[] vertIdxSubs) throws IOException 
  {
    transformationToParent.transform(vertex);
    Integer vertIdx = vertIdxs.get(vertex);
    if (vertIdx == null) {
      vertIdxSubs [index] = vertIdxs.size() + 1;
      vertIdxs.put(vertex, vertIdxSubs [index]);
      // Write only once unique vertices
      out.write("v " + format(vertex.x)
          + " " + format(vertex.y)
          + " " + format(vertex.z) + "\n");
    } else {
      vertIdxSubs [index] = vertIdx;
    }
  }


  /**
Formats a float number to a string as fast as possible depending on the
format chosen in constructor.
   */
  private String format(float number) 
  {
    if (numFormat != null) {
      return numFormat.format(number);
    } else {
      String numberString = String.valueOf((float)number);
      if (numberString.indexOf('E') != -1) {
        // Avoid scientific notation
        return defaultNF.format(number);
      } else {
        return numberString;
      }
    }
  }


  /**
Applies to normal the given transformation, and appends to normalsBuffer
its values in a line vn at OBJ format, if the normal wasn't written yet.
@return true if the written normal doens't contain any NaN value
   */
  private boolean writeNormal(StringBuilder normalsBuffer,
                      Transform3D transformationToParent,
                      Vector3f normal, int index,
                      int[] normalIndexSubs,
                      int[] oppositeSideNormalIndexSubs,
                      List<Vector3f> addedNormals,
                      int cullFace, boolean backFaceNormalFlip) throws IOException 
  {
    if (Float.isNaN(normal.x) || Float.isNaN(normal.y) || Float.isNaN(normal.z)) {
      return false;
    }

    if (backFaceNormalFlip) {
      normal.negate();
    }

    if (normal.x != 0 || normal.y != 0 || normal.z != 0) {
      transformationToParent.transform(normal);
      normal.normalize();
    }

    Integer normalIndex = normalIdxs.get(normal);
    if (normalIndex == null) {
      normalIndexSubs [index] = normalIdxs.size() + 1;
      normalIdxs.put(normal, normalIndexSubs [index]);
      addedNormals.add(normal);
      // Write only once unique normals
      normalsBuffer.append("vn " + format(normal.x)
          + " " + format(normal.y)
          + " " + format(normal.z) + "\n");
    } 
    else {
      normalIndexSubs [index] = normalIndex;
    }

    if (cullFace == PolygonAttributes.CULL_NONE) {
      Vector3f oppositeNormal = new Vector3f();
      oppositeNormal.negate(normal);
      // Fill opposite side normal index substitutes array
      return writeNormal(normalsBuffer, transformationToParent, oppositeNormal, index, oppositeSideNormalIndexSubs,
          null, addedNormals, PolygonAttributes.CULL_FRONT, false);
    } 
    else {
      return true;
    }
  }  // end of writeNormal()



  /**
Writes texCoords in a line vt
if the tex coordinates wasn't written yet.
   */
  private void writeTexCoords(TexCoord2f texCoords, Transform3D texT3d,
                           int index, int[] texCoordsIdxSubs) throws IOException 
  {
    if (texT3d.getBestType() != Transform3D.IDENTITY) {
      Point3f transformedCoords = new Point3f(texCoords.x, texCoords.y, 0);
      texT3d.transform(transformedCoords);
      texCoords = new TexCoord2f(transformedCoords.x, transformedCoords.y);
    }
    Integer texCoordsIdx = texCoordIdxs.get(texCoords);
    if (texCoordsIdx == null) {
      texCoordsIdxSubs [index] = texCoordIdxs.size() + 1;
      texCoordIdxs.put(texCoords, texCoordsIdxSubs [index]);
      // Write only once unique tex coordinates
      out.write("vt " + format(texCoords.x)
          + " " + format(texCoords.y) + " 0\n");
    } else {
      texCoordsIdxSubs [index] = texCoordsIdx;
    }
  }  // end of writeTexCoords() 



  /**
Writes the line indices given at vertIdx1, vertIdx2,
in a line l at OBJ format.
   */
  private void writeIndexedLine(IndexedGeometryArray geomArr,
                         int vertIdx1, int vertIdx2,
                         int[] vertIdxSubs,
                         int[] texCoordsIdxSubs) throws IOException 
  {
    if ((geomArr.getVertexFormat() & GeometryArray.TEXTURE_COORDINATE_2) != 0) {
      out.write("l " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
          + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx1)])
          + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
          + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx2)]) + "\n");
    } 
    else {
      out.write("l " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
          + " "  + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)]) + "\n");
    }
  }  // end of writeIndexedLine()



  /**
Writes the triangle indices given at vertIdx1, vertIdx2, vertIdx3,
in a line f at OBJ format.
   */
  private void writeIndexedTriangle(IndexedGeometryArray geomArr,
                          int vertIdx1, int vertIdx2, int vertIdx3,
                          int[] vertIdxSubs,
                          int[] normalIndexSubs,
                          int[] oppositeSideNormalIndexSubs,
                          boolean normalsDefined,
                          int[] texCoordsIdxSubs,
                          boolean texCoordsGenerated, int cullFace) throws IOException 
  {
    if (cullFace == PolygonAttributes.CULL_FRONT) {
      // Reverse vertex order
      int tmp = vertIdx1;
      vertIdx1 = vertIdx3;
      vertIdx3 = tmp;
    }

    if (texCoordsGenerated) {
      if (normalsDefined) {
        out.write("f " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx1)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx2)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx3)]) + "\n");
      } 
      else {
        out.write("f " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx3)]) + "\n");
      }
    } 
    else if ((geomArr.getVertexFormat() & GeometryArray.TEXTURE_COORDINATE_2) != 0) {
      if (normalsDefined) {
        out.write("f " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx1)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx1)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx2)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx2)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx3)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx3)]) + "\n");
      } 
      else {
        out.write("f " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx1)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx2)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx3)]) + "\n");
      }
    } 
    else {
      if (normalsDefined) {
        out.write("f " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + "//" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx1)])
            + " "  + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + "//" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx2)])
            + " "  + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + "//" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx3)]) + "\n");
      } 
      else {
        out.write("f " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + " "  + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + " "  + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx3)]) + "\n");
      }
    }

    if (cullFace == PolygonAttributes.CULL_NONE) {
      // Use opposite side normal index substitutes array
      writeIndexedTriangle(geomArr, vertIdx1, vertIdx2, vertIdx3,
          vertIdxSubs, oppositeSideNormalIndexSubs, null,
          normalsDefined, texCoordsIdxSubs, texCoordsGenerated, PolygonAttributes.CULL_FRONT);
    }
  }  // end of writeIndexedTriangle()



  /**
Writes the quadrilateral indices given at vertIdx1, vertIdx2, vertIdx3, vertIdx4,
in a line f at OBJ format.
   */
  private void writeIndexedQuad(IndexedGeometryArray geomArr,
                       int vertIdx1, int vertIdx2, int vertIdx3, int vertIdx4,
                       int[] vertIdxSubs,
                       int[] normalIndexSubs,
                       int[] oppositeSideNormalIndexSubs,
                       boolean normalsDefined,
                       int[] texCoordsIdxSubs,
                       boolean texCoordsGenerated, int cullFace) throws IOException 
  {
    if (cullFace == PolygonAttributes.CULL_FRONT) {
      // Reverse vertex order
      int tmp = vertIdx2;
      vertIdx2 = vertIdx3;
      vertIdx3 = tmp;
      tmp = vertIdx1;
      vertIdx1 = vertIdx4;
      vertIdx4 = tmp;
    }

    if (texCoordsGenerated) {
      if (normalsDefined) {
        out.write("f " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx1)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx2)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx3)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx4)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx4)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx4)]) + "\n");
      } 
      else {
        out.write("f " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx4)])
            + "/" + (texCoordsIdxSubs [geomArr.getCoordinateIndex(vertIdx4)]) + "\n");
      }
    } 
    else if ((geomArr.getVertexFormat() & GeometryArray.TEXTURE_COORDINATE_2) != 0) {
      if (normalsDefined) {
        out.write("f " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx1)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx1)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx2)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx2)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx3)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx3)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx4)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx4)])
            + "/" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx4)]) + "\n");
      } 
      else {
        out.write("f " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx1)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx2)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx3)])
            + " " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx4)])
            + "/" + (texCoordsIdxSubs [geomArr.getTextureCoordinateIndex(0, vertIdx4)]) + "\n");
      }
    } 
    else {
      if (normalsDefined) {
        out.write("f " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + "//" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx1)])
            + " "  + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + "//" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx2)])
            + " "  + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + "//" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx3)])
            + " "  + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx4)])
            + "//" + (normalIndexSubs [geomArr.getNormalIndex(vertIdx4)]) + "\n");
      } 
      else {
        out.write("f " + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx1)])
            + " "  + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx2)])
            + " "  + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx3)])
            + " "  + (vertIdxSubs [geomArr.getCoordinateIndex(vertIdx4)]) + "\n");
      }
    }

    if (cullFace == PolygonAttributes.CULL_NONE) {
      // Use opposite side normal index substitutes array
      writeIndexedQuad(geomArr, vertIdx1, vertIdx2, vertIdx3, vertIdx4,
          vertIdxSubs, oppositeSideNormalIndexSubs, null,
          normalsDefined, texCoordsIdxSubs, texCoordsGenerated, PolygonAttributes.CULL_FRONT);
    }
  }  // end of writeIndexedQuad()



  /**
Writes the line indices given at vertIdx1, vertIdx2,
in a line l at OBJ format.
   */
  private void writeLine(GeometryArray geomArr,
                   int vertIdx1, int vertIdx2,
                   int[] vertIdxSubs,
                   int[] texCoordsIdxSubs) throws IOException 
  {
    if ((geomArr.getVertexFormat() & GeometryArray.TEXTURE_COORDINATE_2) != 0) {
      out.write("l " + (vertIdxSubs [vertIdx1])
          + "/" + (texCoordsIdxSubs [vertIdx1])
          + " " + (vertIdxSubs [vertIdx2])
          + "/" + (texCoordsIdxSubs [vertIdx2]) + "\n");
    } 
    else {
      out.write("l " + (vertIdxSubs [vertIdx1])
          + " "  + (vertIdxSubs [vertIdx2]) + "\n");
    }
  }



  /**
Writes the triangle indices given at vertIdx1, vertIdx2, vertIdx3,
in a line f 
   */
  private void writeTriangle(GeometryArray geomArr,
                        int vertIdx1, int vertIdx2, int vertIdx3,
                        int[] vertIdxSubs,
                        int[] normalIndexSubs,
                        int[] oppositeSideNormalIndexSubs,
                        boolean normalsDefined,
                        int[] texCoordsIdxSubs,
                        boolean texCoordsGenerated, int cullFace) 
                                              throws IOException 
  {
    if (cullFace == PolygonAttributes.CULL_FRONT) {
      // Reverse vertex order
      int tmp = vertIdx1;
      vertIdx1 = vertIdx3;
      vertIdx3 = tmp;
    }

    if (texCoordsGenerated
        || (geomArr.getVertexFormat() & GeometryArray.TEXTURE_COORDINATE_2) != 0) {
      if (normalsDefined) {
        out.write("f " + (vertIdxSubs [vertIdx1])
            + "/" + (texCoordsIdxSubs [vertIdx1])
            + "/" + (normalIndexSubs [vertIdx1])
            + " " + (vertIdxSubs [vertIdx2])
            + "/" + (texCoordsIdxSubs [vertIdx2])
            + "/" + (normalIndexSubs [vertIdx2])
            + " " + (vertIdxSubs [vertIdx3])
            + "/" + (texCoordsIdxSubs [vertIdx3])
            + "/" + (normalIndexSubs [vertIdx3]) + "\n");
      } 
      else {
        out.write("f " + (vertIdxSubs [vertIdx1])
            + "/" + (texCoordsIdxSubs [vertIdx1])
            + " " + (vertIdxSubs [vertIdx2])
            + "/" + (texCoordsIdxSubs [vertIdx2])
            + " " + (vertIdxSubs [vertIdx3])
            + "/" + (texCoordsIdxSubs [vertIdx3]) + "\n");
      }
    } 
    else {
      if (normalsDefined) {
        out.write("f " + (vertIdxSubs [vertIdx1])
            + "//" + (normalIndexSubs [vertIdx1])
            + " "  + (vertIdxSubs [vertIdx2])
            + "//" + (normalIndexSubs [vertIdx2])
            + " "  + (vertIdxSubs [vertIdx3])
            + "//" + (normalIndexSubs [vertIdx3]) + "\n");
      } 
      else {
        out.write("f " + (vertIdxSubs [vertIdx1])
            + " "  + (vertIdxSubs [vertIdx2])
            + " "  + (vertIdxSubs [vertIdx3]) + "\n");
      }
    }

    if (cullFace == PolygonAttributes.CULL_NONE) {
      // Use opposite side normal index substitutes array
      writeTriangle(geomArr, vertIdx1, vertIdx2, vertIdx3,
          vertIdxSubs, oppositeSideNormalIndexSubs, null,
          normalsDefined, texCoordsIdxSubs, texCoordsGenerated, PolygonAttributes.CULL_FRONT);
    }
  }  // end of writeTriangle()



  /**
Writes the quadrilateral indices given at vertIdx1, vertIdx2, vertIdx3, vertIdx4,
in a line f
   */
  private void writeQuad(GeometryArray geomArr,
                     int vertIdx1, int vertIdx2, int vertIdx3, int vertIdx4,
                     int[] vertIdxSubs,
                     int[] normalIndexSubs,
                     int[] oppositeSideNormalIndexSubs,
                     boolean normalsDefined,
                     int[] texCoordsIdxSubs,
                     boolean texCoordsGenerated, int cullFace) 
                                             throws IOException 
  {
    if (cullFace == PolygonAttributes.CULL_FRONT) {
      // Reverse vertex order
      int tmp = vertIdx2;
      vertIdx2 = vertIdx3;
      vertIdx3 = tmp;
      tmp = vertIdx1;
      vertIdx1 = vertIdx4;
      vertIdx4 = tmp;
    }

    if (texCoordsGenerated
        || (geomArr.getVertexFormat() & GeometryArray.TEXTURE_COORDINATE_2) != 0) {
      if (normalsDefined) {
        out.write("f " + (vertIdxSubs [vertIdx1])
            + "/" + (texCoordsIdxSubs [vertIdx1])
            + "/" + (normalIndexSubs [vertIdx1])
            + " " + (vertIdxSubs [vertIdx2])
            + "/" + (texCoordsIdxSubs [vertIdx2])
            + "/" + (normalIndexSubs [vertIdx2])
            + " " + (vertIdxSubs [vertIdx3])
            + "/" + (texCoordsIdxSubs [vertIdx3])
            + "/" + (normalIndexSubs [vertIdx3])
            + " " + (vertIdxSubs [vertIdx4])
            + "/" + (texCoordsIdxSubs [vertIdx4])
            + "/" + (normalIndexSubs [vertIdx4]) + "\n");
      } 
      else {
        out.write("f " + (vertIdxSubs [vertIdx1])
            + "/" + (texCoordsIdxSubs [vertIdx1])
            + " " + (vertIdxSubs [vertIdx2])
            + "/" + (texCoordsIdxSubs [vertIdx2])
            + " " + (vertIdxSubs [vertIdx3])
            + "/" + (texCoordsIdxSubs [vertIdx3])
            + " " + (vertIdxSubs [vertIdx4])
            + "/" + (texCoordsIdxSubs [vertIdx4]) + "\n");
      }
    } 
    else {
      if (normalsDefined) {
        out.write("f " + (vertIdxSubs [vertIdx1])
            + "//" + (normalIndexSubs [vertIdx1])
            + " "  + (vertIdxSubs [vertIdx2])
            + "//" + (normalIndexSubs [vertIdx2])
            + " "  + (vertIdxSubs [vertIdx3])
            + "//" + (normalIndexSubs [vertIdx3])
            + " "  + (vertIdxSubs [vertIdx4])
            + "//" + (normalIndexSubs [vertIdx4]) + "\n");
      } 
      else {
        out.write("f " + (vertIdxSubs [vertIdx1])
            + " "  + (vertIdxSubs [vertIdx2])
            + " "  + (vertIdxSubs [vertIdx3])
            + " "  + (vertIdxSubs [vertIdx4]) + "\n");
      }
    }

    if (cullFace == PolygonAttributes.CULL_NONE) {
      // Use opposite side normal index substitutes array
      writeQuad(geomArr, vertIdx1, vertIdx2, vertIdx3, vertIdx4,
          vertIdxSubs, oppositeSideNormalIndexSubs, null,
          normalsDefined, texCoordsIdxSubs, texCoordsGenerated, PolygonAttributes.CULL_FRONT);
    }
  }  // end of writeQuad()




  /**
Closes this writer and writes MTL file and its tex images,
if this writer was created from a file.
   */
  @Override
  public void close() 
          throws IOException, InterruptedIOException 
  {
    super.close();
    if (mtlFnm != null) {
      writeAppsToMTL();
    }
  }


  // Exports a set of appearances to a MTL file
  private void writeAppsToMTL() throws IOException 
  {
    Writer writer = null;
    try {
      System.out.println("Writing material info. to " + mtlFnm);
      writer = new OutputStreamWriter(
                        new BufferedOutputStream(
                                new FileOutputStream(mtlFnm)), "ISO-8859-1");
      writeHeader(writer);
      for (Map.Entry<ComparableAppearance, String> appearanceEntry : apps.entrySet()) {
        checkCurrentThreadIsntInterrupted();

        Appearance app = appearanceEntry.getKey().getAppearance();
        String appNm = appearanceEntry.getValue();
        writer.write("\nnewmtl " + appNm + "\n");
        Material mat = app.getMaterial();
        if (mat != null) {
          if (mat.getShininess() > 1) {
            writer.write("illum 2\n");
          } 
          else if (mat.getLightingEnable()) {
            writer.write("illum 1\n");
          } 
          else {
            writer.write("illum 0\n");
          }

          Color3f color = new Color3f();
          mat.getAmbientColor(color);
          writer.write("Ka " + format(color.x) + " " + format(color.y) + " " + format(color.z) + "\n");
          mat.getDiffuseColor(color);
          writer.write("Kd " + format(color.x) + " " + format(color.y) + " " + format(color.z) + "\n");
          mat.getSpecularColor(color);
          writer.write("Ks " + format(color.x) + " " + format(color.y) + " " + format(color.z) + "\n");
          writer.write("Ns " + format(mat.getShininess()) + "\n");
        } 
        else {
          ColoringAttributes ca = app.getColoringAttributes();
          if (ca != null) {
            writer.write("illum 0\n");
            Color3f color = new Color3f();
            ca.getColor(color);
            writer.write("Ka " + format(color.x) + " " + format(color.y) + " " + format(color.z) + "\n");
            writer.write("Kd " + format(color.x) + " " + format(color.y) + " " + format(color.z) + "\n");
            writer.write("Ks " + format(color.x) + " " + format(color.y) + " " + format(color.z) + "\n");
          }
        }
        TransparencyAttributes ta = app.getTransparencyAttributes();
        if (ta != null) {
          writer.write("Ni 1\n");
          writer.write("d " + format(1f - ta.getTransparency()) + "\n");
        }
        Texture tex = app.getTexture();
        if (tex != null) {
          writer.write("map_Kd " + textures.get(tex).getName() + "\n");
        }
      }

      for (Map.Entry<Texture, File> textureEntry : textures.entrySet()) {
        Texture tex = textureEntry.getKey();
        Object textureUrl = tex.getUserData();
        if (copiedTextures.contains(textureUrl)) {
          // Copy tex image file directly
          InputStream in = null;
          OutputStream out = null;
          try {
            in = openStream((URL)textureUrl);
            out = new FileOutputStream(textureEntry.getValue());
            byte [] buffer = new byte [8192];
            int size;
            while ((size = in.read(buffer)) != -1) {
              out.write(buffer, 0, size);
            }
          } 
          finally {
            if (in != null) {
              in.close();
            }
            if (out != null) {
              out.close();
            }
          }
        } 
        else {
          ImageComponent2D imageComponent = (ImageComponent2D)tex.getImage(0);
          RenderedImage image = imageComponent.getRenderedImage();
          ImageIO.write(image, "png", textureEntry.getValue());
        }
      }
    } 
    finally {
      if (writer != null) {
        writer.close();
      }
    }
  }  // end of writeAppsToMTL()



  /**
Writes node in an entry at OBJ format of the given zip file
along with its MTL file and tex images.
   */
  public static void writeNodeInZIPFile(Node node,
                                File zipFile,
                                int comLvl,
                                String entryName,
                                String header) throws IOException 
  {
    writeNodeInZIPFile(node, null, zipFile, comLvl, entryName, header);
  }


  /**
Writes node in an entry at OBJ format of the given zip file
along with its MTL file and tex images.
Once saved, matAppearances will contain the apps matching
each mat saved in the MTL file. Material names used as keys maybe be different
of the app names to respect MTL specifications.
   */
  public static void writeNodeInZIPFile(Node node,
                                Map<String, Appearance> matAppearances,
                                File zipFile,
                                int comLvl,
                                String entryName,
                                String header) throws IOException 
  {
    // Create a temporary folder
    File tempFolder = null;
    for (int i = 0; i < 10 && tempFolder == null; i++) {
      tempFolder = File.createTempFile("obj", "tmp");
      tempFolder.delete();
      if (!tempFolder.mkdirs()) {
        tempFolder = null;
      }
    }
    if (tempFolder == null) {
      throw new IOException("Couldn't create a temporary folder");
    }

    ZipOutputStream zipOut = null;
    try {
      // Write model in an OBJ file
      OBJWriter writer = new OBJWriter(new File(tempFolder, entryName), header, -1);
      writer.writeNode(node);
      writer.close();
      // Create a ZIP file containing temp folder files (OBJ + MTL + tex files)
      zipOut = new ZipOutputStream(new FileOutputStream(zipFile));
      zipOut.setLevel(comLvl);
      for (File tempFile : tempFolder.listFiles()) {
        if (tempFile.isFile()) {
          InputStream tempIn = null;
          try {
            zipOut.putNextEntry(new ZipEntry(tempFile.getName()));
            tempIn = new FileInputStream(tempFile);
            byte [] buffer = new byte [8096];
            int size;
            while ((size = tempIn.read(buffer)) != -1) {
              zipOut.write(buffer, 0, size);
            }
            zipOut.closeEntry();
          } 
          finally {
            if (tempIn != null) {
              tempIn.close();
            }
          }
        }
      }

      if (matAppearances != null) {
        // Update mat / app map
        for (Map.Entry<ComparableAppearance, String> appearanceEntry : writer.apps.entrySet()) {
          matAppearances.put(appearanceEntry.getValue(), appearanceEntry.getKey().getAppearance());
        }
      }
    } 
    finally {
      if (zipOut != null) {
        zipOut.close();
      }
      // Empty tempFolder
      for (File tempFile : tempFolder.listFiles()) {
        if (tempFile.isFile()) {
          tempFile.delete();
        }
      }
      tempFolder.delete();
    }
  }  // end of writeNodeInZIPFile()



  // ----------------------------------------------------------------

  /**
An Appearance wrapper able to compare
if two apps are equal for MTL format.
   */
  private static class ComparableAppearance 
  {
    private Appearance app;


    public ComparableAppearance(Appearance app) 
    {
      this.app = app;
    }

    public Appearance getAppearance() 
    {
      return app;
    }


    /**
     * Returns true if this app and the one of obj
     * describe the same colors, ta and tex.
     */
    @Override
    public boolean equals(Object obj) 
    {
      if (obj instanceof ComparableAppearance) {
        Appearance app2 = ((ComparableAppearance)obj).app;
        // Compare coloring attributes
        ColoringAttributes ca1 = app.getColoringAttributes();
        ColoringAttributes ca2 = app2.getColoringAttributes();
        if ((ca1 == null) ^ (ca2 == null)) {
          return false;
        } 
        else if (ca1 != ca2) {
          Color3f color1 = new Color3f();
          Color3f color2 = new Color3f();
          ca1.getColor(color1);
          ca2.getColor(color2);
          if (!color1.equals(color2)) {
            return false;
          }
        }

        // Compare mat colors
        Material mat1 = app.getMaterial();
        Material mat2 = app2.getMaterial();
        if ((mat1 == null) ^ (mat2 == null)) {
          return false;
        } 
        else if (mat1 != mat2) {
          Color3f color1 = new Color3f();
          Color3f color2 = new Color3f();
          mat1.getAmbientColor(color1);
          mat2.getAmbientColor(color2);
          if (!color1.equals(color2)) {
            return false;
          } 
          else {
            mat1.getDiffuseColor(color1);
            mat2.getDiffuseColor(color2);
            if (!color1.equals(color2)) {
              return false;
            } 
            else {
              mat1.getEmissiveColor(color1);
              mat2.getEmissiveColor(color2);
              if (!color1.equals(color2)) {
                return false;
              } 
              else {
                mat1.getSpecularColor(color1);
                mat2.getSpecularColor(color2);
                if (!color1.equals(color2)) {
                  return false;
                } 
                else if (mat1.getShininess() != mat2.getShininess()) {
                  return false;
                } 
                else if (mat1.getClass() != mat2.getClass()) {
                  return false;
                } 
              }
            }
          }
        }

        // Compare ta
        TransparencyAttributes ta1 = app.getTransparencyAttributes();
        TransparencyAttributes ta2 = app2.getTransparencyAttributes();
        if ((ta1 == null) ^ (ta2 == null)) {
          return false;
        } 
        else if (ta1 != ta2) {
          if (ta1.getTransparency() != ta2.getTransparency()) {
            return false;
          }
        }

        // Compare tex
        Texture texture1 = app.getTexture();
        Texture texture2 = app2.getTexture();
        if ((texture1 == null) ^ (texture2 == null)) {
          return false;
        } 
        else if (texture1 != texture2) {
          if (texture1.getImage(0) != texture2.getImage(0)) {
            return false;
          }
        }

        // Compare name
        try {
          String name1 = app.getName();
          String name2 = app2.getName();
          if ((name1 == null) ^ (name2 == null)) {
            return false;
          } 
          else if (name1 != name2
                     && !name1.equals(name2)) {
            return false;
          }
        } 
        catch (NoSuchMethodError ex) {
          // Don't compare name with Java 3D < 1.4 where getName was added
        }

        return true;
      }
      return false;
    }  // end of equals()



    @Override
    public int hashCode() {
      int code = 0;
      ColoringAttributes ca = app.getColoringAttributes();
      if (ca != null) {
        Color3f color = new Color3f();
        ca.getColor(color);
        code += color.hashCode();
      }

      Material mat = app.getMaterial();
      if (mat != null) {
        Color3f color = new Color3f();
        mat.getAmbientColor(color);
        code += color.hashCode();
        mat.getDiffuseColor(color);
        code += color.hashCode();
        mat.getEmissiveColor(color);
        code += color.hashCode();
        mat.getSpecularColor(color);
        code += color.hashCode();
        code += Float.floatToIntBits(mat.getShininess());
      }

      TransparencyAttributes ta = app.getTransparencyAttributes();
      if (ta != null) {
        code += Float.floatToIntBits(ta.getTransparency());
      }

      Texture tex = app.getTexture();
      if (tex != null) {
        code += tex.getImage(0).hashCode();
      }

      try {
        String name = app.getName();
        if (name != null) {
          code += name.hashCode();
        }
      } 
      catch (NoSuchMethodError ex) {
        // Don't take name into account with Java 3D < 1.4 where getName was added
      }
      return code;
    }  // end of hashCode()


  } // end of ComparableAppearance class



}  // end of OBJWriter class
