
// PickBehavior.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, Nov. 2021

/*  PickBehavior is triggered when the mouse is clicked on a 
    pick-enabled shape (or primitive) in the 3D scene. 
    Its processStimulus() method
    is called which in turn calls updateScene() which is defined in
    this subclass.

    updateScene() calls PickUpdater.picked() with the id of the
    selected shape and the intersection point on the shape where 
    the mouse was clicked (if that info is available).
*/


import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;
import org.jogamp.java3d.utils.geometry.*;

import org.jogamp.java3d.utils.picking.*;
import org.jogamp.java3d.utils.picking.behaviors.*;



public class PickBehavior extends PickMouseBehavior
{
  private PickUpdater pu;


  public PickBehavior(PickUpdater pu, Canvas3D canvas, BranchGroup sceneBG)
  {
    super(canvas, sceneBG, Simple3D.INFINITE_BOUNDS);
    this.pu = pu;
    setSchedulingBounds(Simple3D.INFINITE_BOUNDS);
  }


  public void updateScene(int xpos, int ypos)
  /* Use the mouse click (x,y) and the camera's position
     to project a pick ray into the scene. Return the first 
     pickable shape as a PickResult object.

     Extract the name (id) of the shape, and try to get the
     3D point where the ray intersects with the shape.

     Call PickUpdater.picked() with the id and the point.
  */
  {
    pickCanvas.setShapeLocation(xpos, ypos);
         // register mouse pointer location on the screen (canvas)

    Point3d cameraPos = pickCanvas.getStartPosition();
            // get the camera's's location

    PickResult pickResult = pickCanvas.pickClosest();
         // get the intersected shape closest to the camera
    if (pickResult != null) {
      String id = Simple3D.pickID(pickResult);

      // now try to get ray intersect
      Point3d intercept = null;
      try {
        PickIntersection pi = pickResult.getClosestIntersection(cameraPos);
        intercept = pi.getPointCoordinatesVW();
      }
      catch(Exception e)  // intersection may fail
      { //System.out.println("No intercept found for " +  id);
      }
      
      pu.picked(id, intercept);
    }
  }  // end of updateScene()


}  // end of PickBehavior class
