
// RotPosPath.java
// Based on code by Marco13
// Modified by Andrew Davison, ad@fivedots.coe.psu.ac.th, October 2021

// https://stackoverflow.com/questions/29609439/making-a-correct-animation-with-java3d-and-rotpospathinterpolator

/* Convert a list of (x,y,z) points and y-axis degree 
   rotations into a series of Point3d and quaternions 
   suitable for supplying to the constructor of an
   RotPosPathInterpolator. The length of the Points list
   is also used to generate a sequence of Alpha changes
   from 0 to 1.
*/


import java.util.*;

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;


public class RotPosPath
{
  private int axisType;

  private List<Point3f> locs = new ArrayList<Point3f>();
  private List<Quat4f> rots = new ArrayList<Quat4f>();


  public RotPosPath()
  {  this(Simple3D.Y_AXIS);  }


  public RotPosPath(int axisType)
  {  this.axisType = axisType;  }


  public void add(float x, float y, float z, float angleDeg)
  {
    locs.add( new Point3f(x,y,z));

    Quat4f quat = null;

    if (axisType == Simple3D.Z_AXIS) 
      quat = Simple3D.aa2Quat(0,0,1, angleDeg);   // along +z-axis
    else if (axisType == -Simple3D.Z_AXIS) 
      quat = Simple3D.aa2Quat(0,0,-1, angleDeg);   // -z-axis

    else if (axisType == Simple3D.X_AXIS) 
      quat = Simple3D.aa2Quat(1,0,0, angleDeg);   // along +x-axis
    else if (axisType == -Simple3D.X_AXIS) 
      quat = Simple3D.aa2Quat(-1,0,0, angleDeg);   // -x-axis

    else if (axisType == Simple3D.Y_AXIS) 
      quat = Simple3D.aa2Quat(0,1,0, angleDeg);   // along y-axis
    else if (axisType == -Simple3D.Y_AXIS) 
      quat = Simple3D.aa2Quat(0,-1,0, angleDeg);   // -y-axis

    else
      System.out.println("axisType unknown: " + axisType);
      
    rots.add(quat);
  } /* add */





  public TransformGroup getTG(Node obj, int period)
  /* Return a new TransformGroup from which hangs the supplied
     node and the RotPosPathInterpolator:

       tg --> obj
        ^
        |
        ----> rot pos path interpolator
  */
  {
    TransformGroup tg = Simple3D.createTG();
    tg.addChild(obj);

    Alpha alpha = new Alpha(-1, period);
    alpha.setStartTime( System.currentTimeMillis()); // start now

    RotPosPathInterpolator interp = new RotPosPathInterpolator(
        alpha, tg, new Transform3D(), 
        getAlphas(), getRots(), getLocs());  
      
    interp.setSchedulingBounds(Simple3D.INFINITE_BOUNDS);

    tg.addChild(interp);
    return tg;
  }  // end of rotPosPath()


  private float[] getAlphas()
  {
    float alphas[] = new float[locs.size()];
    float delta = 1.0f / (alphas.length - 1);
    for (int i=0; i < alphas.length; i++)
      alphas[i] = i * delta;
    return alphas;
  } /* getAlphas */


  private Quat4f[] getRots()
  {  return rots.toArray(new Quat4f[0]);  }


  private Point3f[] getLocs()
  {  return locs.toArray(new Point3f[0]);  }


}  // end of RotPosPath class