
// Simple3D.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, Dec. 2021

/* A class to simplify the use of Java3D for 
   new programmers.

   This is done in three ways:

    1) The Simple3D constructor builds all of the top-level
       scenegraph of the Java3D program using a SimpleUniverseG
       object, and creates a default background, set of lights,
       and an optional grid-based floor and set of x-, y-, z- axes.
       The camera behavior is augmented to allow both mouse-based
       and key-based navigation.

    2) The Java3D scenegraph is embedded in a JFrame window that 
       contains a JOGL panel to hold the Canvas3D object, and 
       includes a menu bar. 

       The JOGL panel seems more stable than using a
       Swing JPanel, and the JFrame uses a few coding tricks which 
       seem to improve rendering stability across multiple monitors
       and graphics cards. 

       The menu bar includes a screenshot feature,
       information about the camera controls, and a menu item to 
       reset the camera position to its starting point.

    3) A collection of static methods that simplify common tasks
       that a Java3D beginner usually finds complicated:
          * translations, rotations, scaling
          * orbit behaviour
          * node creation
          * colors and textures for the Appearance node
          * loading OBJ models   
          * setting up shaders


   ------------------------
   Utilizes JOGL and Java3D 1.7:

   I've been using JOGL from Aug 2021:
   https://jogamp.org/deployment/v2.4.0-rc-20210111/fat/


   Java 3D info:
   Note: Java3D 1.7 has renamed packages to fit in with the
   rest of the JOGL hierarchy
   https://jogamp.org/wiki/index.php/Downloading_and_installing_Java3D
   https://gouessej.wordpress.com/2012/08/01/java-3d-est-de-retour-java-3d-is-back/

    Java3D (1.5.1) docs:
      https://download.java.net/media/java3d/javadoc/1.5.1/
    
    Java 3D tutorial
      -- available on my website
    
    Java 3D Forum at jogamp
    https://forum.jogamp.org/java3d-f3728156.html
    
    Selman's Java3D book:
    https://webserver2.tecgraf.puc-rio.br/~ismael/Cursos/Cidade_CG/labs/Java3D/Java3D_onlinebook_selman/onlinebook.html
    

   Compilation/execution requires the installation of:
     jogamp-fat.jar       // JOGL JAR files

     j3dcore.jar    \
     j3dutils.jar   |----- the Java3D JAR files
     vecmath.jar    /

   Compilation and execution of my code:
     Use the compile.bat and run.bat files on MS Windows
*/


import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.net.*;
import java.awt.image.BufferedImage;
import javax.imageio.ImageIO;

// JOGL stuff

import com.jogamp.opengl.*;
import com.jogamp.opengl.awt.GLJPanel;

import jogamp.nativewindow.x11.X11Util;
// import com.jogamp.newt.opengl.GLWindow;
import com.jogamp.nativewindow.NativeSurface;
import com.jogamp.nativewindow.x11.X11GraphicsDevice;
import com.jogamp.nativewindow.AbstractGraphicsDevice;

// Java3D 1.7 changes package names

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;

import org.jogamp.java3d.utils.universe.*;
import org.jogamp.java3d.utils.geometry.*;
import org.jogamp.java3d.utils.image.*;
import org.jogamp.java3d.utils.behaviors.vp.*;
import org.jogamp.java3d.utils.picking.*;

import org.jogamp.java3d.loaders.objectfile.ObjectFile;
import org.jogamp.java3d.loaders.*;

import org.jogamp.java3d.utils.shader.StringIO;



public class Simple3D implements ActionListener
{
  // axis names
  public static final int X_AXIS = 1;
  public static final int Y_AXIS = 2;
  public static final int Z_AXIS = 3;

  // bounding sphere used by behaviours, lighting, etc.
  public static final BoundingSphere INFINITE_BOUNDS =
                  new BoundingSphere( new Point3d(0,0,0), 
                                      Double.MAX_VALUE);

  public SimpleUniverse su;

  public Canvas3D canvas3D;
       // has to be public for picking behavior code

  public BranchGroup rootBG;      
       // node that connects a user's scenegraph to the rest
       // of the Java3D graph 


  private static final int PWIDTH = 600;   // ibitial size of panel
  private static final int PHEIGHT = 600; 

  // initial viewpoint position
  private static final Point3d USERPOS = new Point3d(0,1,5);


  // Default clipping distances for rendering
  private static final double FRONT_CLIP = 0.01;
  private static final double BACK_CLIP = 100;

  // size of floor along the +x and +z axis
  private static final int FLOOR_LEN = 10;


  // GUI Components
  private JFrame frame;
  private JMenuBar menuBar;
  private JMenu fileMenu;
  private JMenuItem saveMI, quitMI, resetMI;
  private static JCheckBoxMenuItem helpCheckBox;

  private GLJPanel panel;

  // private Random rand = new Random();

  // Camera 
  private static MyOrbitBehavior orbitBeh;
  private static KeysBehavior keysBeh;
  private static boolean showHelp;
  private static BufferedImage navHelp;

  private static int ssCount = 1;   
          // a counter for uniquely naming screenshot files

  private static Font texFont = new Font("Serif", Font.BOLD, 48);
                       // font used to draw text in a texture image



  public Simple3D()
  {  this("Simple3D", null, false, false);  }

  public Simple3D(String title)
  {  this(title, null, false, false);  }


  public Simple3D(String title, double x, double y, double z)
  {  this(title, new Point3d(x,y,z), false, false);  }

  public Simple3D(String title, boolean hasFloor, boolean hasAxes)
  {  this(title, null, hasFloor, hasAxes);  }


  public Simple3D(String title, double x, double y, double z,
                        boolean hasFloor, boolean hasAxes)
  {  this(title, new Point3d(x,y,z), hasFloor, hasAxes);  }



  public Simple3D(String title, Point3d pos, boolean hasFloor, 
                                             boolean hasAxes)
  /* the constructor can include a title for the window frame, a
     starting position for the camera, and flags for displaying
     the grid-based floor and the x-, y-, z- axes
  */
  {
    System.out.println("Loading JOGL, Java 3D, Simple3D...");
    
    GLJPanel canvasPanel = initCanvas();   // a JOGL Panel
    initShaderUniv();
    initFrame(title, canvasPanel);

    rootBG = new BranchGroup();
    rootBG.setCapability(BranchGroup.ALLOW_CHILDREN_READ);
    rootBG.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
    rootBG.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
    rootBG.setUserData("rootBG");

    lightScene();     // add lights
    addBackground();  // add sky

    if (pos != null)
      initCamera(pos, canvas3D);
    else
      initCamera(USERPOS, canvas3D);

    initViewer();

    if (hasFloor)
     rootBG.addChild( makeFloorGrid(FLOOR_LEN));
    if (hasAxes)
      rootBG.addChild( makeAxes(FLOOR_LEN));

    su.addBranchGraph(rootBG);
       /* rootBG is the node that connects a user's scenegraph 
          to the rest of the Java3D graph 
       */

    frame.setVisible(true);
    frame.toFront();

    canvas3D.setFocusable(true);   // give focus to the 3D canvas
    canvas3D.requestFocus();
  }  // end of Simple3D()



  private GLJPanel initCanvas()
  // add 3D Canvas to a JOGL panel
  {
    GLProfile glProfile = GLProfile.getDefault();
    GLCapabilities glCaps = new GLCapabilities(glProfile);
    // System.out.println("GLCapabilities: " + glCaps);

    // GraphicsDevice gDevice = GraphicsEnvironment.
    //           getLocalGraphicsEnvironment().getDefaultScreenDevice();
    // System.out.println("GraphicsDevice: " + gDevice);
    X11Util.markAllDisplaysUnclosable();  // for Ubuntu

    // GLJPanel 
    panel = new GLJPanel(glCaps);   
    // JPanel panel = new JPanel();
                      // the Swing JPanel sometimes crashes 

    panel.setLayout(new BorderLayout());
    panel.setOpaque(false);
    panel.setPreferredSize(new Dimension(PWIDTH, PHEIGHT));

    navHelp = Simple3D.loadImage("images/navHelp.png");

    GraphicsConfiguration config =
                   SimpleUniverse.getPreferredConfiguration();
    canvas3D = new Canvas3D(config)
    {
      public void postRender()
      // display info as an overlay image
      {
        J3DGraphics2D g = this.getGraphics2D();
        if (showHelp)
          g.drawRenderedImage(displayInfo(), null); 
        g.flush(false);
      }
    };

    panel.add("Center", canvas3D);  

    su = new SimpleUniverse(canvas3D); 
    return panel;
  }  // end of initCanvas()



  private void initShaderUniv()
  {
    su.addShaderErrorListener(new ShaderErrorListener() {
      public void errorOccurred(ShaderError error) 
      {  error.printVerbose();  }
    });

    // Ensure at least 5 ms per frame (i.e., < 200Hz)
    su.getViewer().getView().setMinimumFrameCycleTime(5);
  }  // end of initShaderUniv()



  private void initFrame(String title, GLJPanel canvasPanel)
  {
    System.setProperty("sun.awt.noerasebackground", "true");
    JPopupMenu.setDefaultLightWeightPopupEnabled(false);
    ToolTipManager ttm = ToolTipManager.sharedInstance();
    ttm.setLightWeightPopupEnabled(false);
      /* Swing menus always hide behind a Canvas3D – the only workaround 
          is to use old heavyweight menus. */

    frame = new JFrame(title);
    frame.setVisible(false);
    frame.setResizable(true);
    // frame.setLocationRelativeTo(null);
    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

    frame.addWindowFocusListener(new WindowAdapter() {
      // added to fix GLJPanel going blank when another window
      // partly obsures it; the window is redrawn once the mouse
      // is moved over it
      public void windowGainedFocus(WindowEvent e) {
        // System.out.println("gained focus");
      }
    });

    frame.addComponentListener(new ComponentAdapter() {
      // added to avoid GLJPanel redrawing probs when moving 
      // the JFrame between monitors
      public void componentMoved(ComponentEvent e) {
        // System.out.println("window moved");
      }
    });

    frame.addWindowListener(new WindowAdapter() {
       /* Various unsuccessful attempts to avoid an
          X11Util.Display shutdown warning when Simple3D closes
          on Ubuntu. Also see initCanvas() above.
          This seems to relate to JOGAMP bug 705: 
          https://jogamp.org/bugzilla/show_bug.cgi?id=705
       */
       public void windowClosing(WindowEvent e) 
       {
         // System.out.println("Closing");
         frame.setVisible(false);
/*
         GraphicsDevice gDevice = GraphicsEnvironment.
             getLocalGraphicsEnvironment().getDefaultScreenDevice();
         // System.out.println("GraphicsDevice: " + gDevice);
         if (gDevice.toString().contains("X11")) {
           // long displayHandle = panel.getHandle();
           NativeSurface surface = panel.getNativeSurface();
           long displayHandle = surface.getDisplayHandle();
           X11GraphicsDevice x11Device = new X11GraphicsDevice(
                                    displayHandle, 
                                    AbstractGraphicsDevice.DEFAULT_UNIT, true);
           // x11Device.close();  // this causes a thread-related crash
         }
*/
         frame.getContentPane().remove(panel);
         panel.setEnabled(false);
         panel.destroy();
         frame.dispose();
       }
    });

    frame.add(canvasPanel);
    frame.setJMenuBar(createMenuBar());
    frame.pack();
  }  // end of initFrame()



  private BufferedImage displayInfo()
  {
    BufferedImage im = new BufferedImage(frame.getWidth(), 
                                         frame.getHeight(), 
                                  BufferedImage.TYPE_INT_ARGB);
    Graphics2D g = (Graphics2D) im.getGraphics();
    if (showHelp)
      g.drawImage(navHelp, null, 
             im.getWidth() - navHelp.getWidth() - 20, 5);

    return im;
  }  // end of displayInfo()



  private void lightScene()
  // a dim ambient light and two directional lights
  {
    Color3f white = ColorFactory.getColor("white");
    // Color3f yellow = ColorFactory.getColor("yellow");

    // Set up the ambient light
    AmbientLight ambientLightNode = 
                new AmbientLight(ColorFactory.getColor("dimgray"));
    ambientLightNode.setInfluencingBounds(INFINITE_BOUNDS);
    rootBG.addChild(ambientLightNode);

    // Set up the directional lights
    Vector3f dir1  = new Vector3f(-1f, 0f, -0.5f);
    Vector3f dir2  = new Vector3f(1f, 0f, 0.5f);

    // Vector3f dir1  = new Vector3f(-1.0f, -1.0f, -1.0f);
                                // left, down, backwards 
    // Vector3f dir2  = new Vector3f(1.0f, -1.0f, 1.0f);
                                // right, down, forwards
    DirectionalLight light1 = new DirectionalLight(white, dir1);
    light1.setInfluencingBounds(INFINITE_BOUNDS);

    DirectionalLight light2 = new DirectionalLight(white, dir2);
    light2.setInfluencingBounds(INFINITE_BOUNDS);

    rootBG.addChild(light1);
    rootBG.addChild(light2);
  }  // end of lightScene()



  private void addBackground()
  // A blue sky
  { Background back = new Background();
    back.setApplicationBounds(INFINITE_BOUNDS);
    back.setColor(0.17f, 0.65f, 0.92f);
    rootBG.addChild( back );
  }  // end of addBackground()



  private void initCamera(Point3d pos, Canvas3D c)
  {
    ViewingPlatform vp = su.getViewingPlatform();
    TransformGroup steerTG = vp.getViewPlatformTransform();

    Transform3D t3d = new Transform3D();
    steerTG.getTransform(t3d);

    // Set the camera's initial viewpoint using lookAt()
    t3d.lookAt(pos, new Point3d(0,0,0), new Vector3d(0,1,0));
        // viewer posn; where looking; up direction (the +y axis)

    try {
      t3d.invert();
    }
    catch (SingularMatrixException e) {
      // occurs if the user supplies a starting point over the + y-axis
      System.out.println("Camera position is illegal; using default");
      steerTG.getTransform(t3d);
      t3d.lookAt(USERPOS, new Point3d(0,0,0), new Vector3d(0,1,0));
      t3d.invert();
    }

    steerTG.setTransform(t3d);

    // set up controls to move the camera viewpoint

    /* OrbitBehaviour should allow the user to rotate around the 
       scene, zoom in and out, and translate with the mouse, but the
       translation part is broken. I fixed it in MyOrbitBehavior
    */
    orbitBeh = 
		new MyOrbitBehavior(c, MyOrbitBehavior.REVERSE_ALL);
    orbitBeh.setSchedulingBounds(INFINITE_BOUNDS);
    vp.setViewPlatformBehavior(orbitBeh);	   

    // keyboard controls for the camera; explained in navhelp
 	keysBeh = new KeysBehavior(steerTG);
	keysBeh.setSchedulingBounds(INFINITE_BOUNDS);
    rootBG.addChild(keysBeh);

  }  // end of initCamera()


  public void cameraMouseEnable(boolean b)
  {  orbitBeh.setEnable(b);  }



  public void setCameraPos(double x, double y, double z)
  {  keysBeh.moveTo(x, y, z);  }



  private void initViewer()
  { 
    Viewer viewer = su.getViewer();
    View view = viewer.getView();
    view.setTransparencySortingPolicy(View.TRANSPARENCY_SORT_GEOMETRY);
    view.setSceneAntialiasingEnable(false);
            // anti-aliasing is resource heavy
    view.setFrontClipDistance(FRONT_CLIP);
    view.setBackClipDistance(BACK_CLIP);
  }



  // menu bar ----------------------------


  private JMenuBar createMenuBar()
  {
    int modkey = Toolkit.getDefaultToolkit().getMenuShortcutKeyMaskEx();
    menuBar = new JMenuBar();

    // ------------ file menu ----------------
    fileMenu = new JMenu("File");

    menuBar.add(fileMenu);

    saveMI = new JMenuItem(" Take a Screenshot...");
    saveMI.addActionListener(this);
    saveMI.setAccelerator( KeyStroke.getKeyStroke(KeyEvent.VK_S, modkey));
    fileMenu.add(saveMI);

    fileMenu.addSeparator();

    quitMI = new JMenuItem(" Quit...");
    quitMI.addActionListener(this);
    quitMI.setAccelerator( KeyStroke.getKeyStroke(KeyEvent.VK_Q, modkey));
    fileMenu.add(quitMI);


    // ----------- camera menu ---------------------

    JMenu cameraMenu = new JMenu("Camera");
    menuBar.add(cameraMenu);

    resetMI = new JMenuItem(" Reset camera");
    resetMI.addActionListener(this);
    cameraMenu.add(resetMI);

    cameraMenu.addSeparator();

    helpCheckBox = new JCheckBoxMenuItem("Show Help");
    helpCheckBox.setFocusable(false);
    helpCheckBox.addActionListener(this);
    cameraMenu.add(helpCheckBox);

    return menuBar;
  }  // end of createMenuBar()



  // Action Listener -----------------------------


  public void actionPerformed(ActionEvent e)
  {
    Object source = e.getSource();
    if (source == saveMI) {
      new Thread(() -> {
        takeScreenShot();  
        // using a thread so the method's pausing doesn't affect the menu
      }).start();
    }
    else if (source == quitMI)
      frame.dispatchEvent(new WindowEvent(frame, 
                                   WindowEvent.WINDOW_CLOSING));

    // camera menu
    else if (source == resetMI)
      keysBeh.reset();
    else if (source == helpCheckBox)
      showHelp = helpCheckBox.isSelected();

  }  // end of actionPerformed()




  public void takeScreenShot()
  {
    Simple3D.pause(600);
    // so the pop-down menu has time to disappear from the screen

    int borderWidth = (frame.getWidth() - 
                       frame.getContentPane().getWidth())/2;
    int borderHeight = (frame.getHeight() - 
                        frame.getContentPane().getHeight())/2;
    Point pt = frame.getLocationOnScreen();

    Rectangle screenRect = new Rectangle( pt.x + borderWidth, pt.y, 
                         frame.getWidth() - borderWidth*2,
                         frame.getHeight() - borderHeight);

    String fnm = String.format("screenshot%02d.png", Simple3D.ssCount++);
    try {
      BufferedImage image = new Robot().createScreenCapture(screenRect);
      ImageIO.write(image, "png", new File(fnm));
      System.out.println("Saved a screenshot to " + fnm);
    }
    catch (Exception e) {
      System.out.println("Could not save a screenshot: " + e);
    }
  }  // end of takeScreenShot()


  public void exit()
  {  
    System.out.println("Simple3D closing down...");
    frame.setVisible(false);
    frame.dispose();
    System.exit(0);
  }


  // the floor grid and axes --------------------------


  private BranchGroup makeFloorGrid(int len)
  {
    BranchGroup bg = new BranchGroup();
    double[] ys = {0, 0, 0, 0};

    Shape3D tile;
    for (int i = -len; i < len; i++) {
      for (int k = -len; k < len; k++) {
         double[] xs = {i, i+1, i+1, i};
         double[] zs = {k+1, k+1, k, k};
         Point3d[] pts = toPoints(xs, ys, zs);
         if ((i+k)%2 == 0) 
           tile = polygon(pts, "darkgreen");
         else
           tile = polygon(pts, "darkblue");
         bg.addChild(tile);
      }
    }
    return bg;
  }  // end of makeFloorGrid()



  private Point3d[] toPoints(double[] xs, double[] ys, double[] zs)
  // Construct a Point3d array from the coordinate arrays
  {
    int size = xs.length;
    Point3d[] coords = new Point3d[size];
    for (int i = 0; i < size; i++)
      coords[i] = new Point3d( xs[i], ys[i], zs[i] );
    return coords;
  }



  private Shape3D polygon(Point3d[] coords, String colorNm)
  {
    GeometryArray geom = new TriangleFanArray(coords.length,
                          LineArray.COORDINATES,
                          new int[] { coords.length});
    geom.setCoordinates(0, coords);

    GeometryInfo geomInfo = new GeometryInfo((GeometryArray) geom);

    NormalGenerator ng = new NormalGenerator();
    ng.generateNormals(geomInfo);

    Appearance app = Simple3D.color(colorNm);
    Shape3D shape = new Shape3D(geomInfo.getIndexedGeometryArray(), app);

    return shape;
  }  // end of polygon()



  private BranchGroup makeAxes(int len)
  {
    BranchGroup bg = new BranchGroup();

    bg.addChild( line(-len, 0.02, 0, len, 0.02, 0));   // x-
    bg.addChild( line(0, 0.02, -len, 0, 0.02, len));   // z-
    bg.addChild( line(0, -len/2, 0, 0, len/2, 0));   // y-  (half length)

    // x-axis tick marks and labels
    for (int i = -len; i <= len; i++) {
      bg.addChild( line(i, 0, 0, i, 0.1, 0));   
      bg.addChild( Simple3D.label(i-0.1, 0.02, 0, ""+i, 2));
    }

    // y-axis tick marks and labels
    for (int j = -len/2; j <= len/2; j++) {
      bg.addChild( line(-0.1, j, 0, 0, j, 0)); 
      if (j != 0)  
        bg.addChild( Simple3D.label(0.1, j-0.05, 0, ""+j, 2));
    }

    // z-axis tick marks and labels
    for (int k = -len; k <= len; k++) {
      bg.addChild( line(0, 0, k, 0, 0.1, k));  
      if (k != 0) 
        bg.addChild( Simple3D.label(-0.07, 0.02, k+0.05, ""+k, 2));
    }

    // axis labels
    bg.addChild( Simple3D.label(len+0.5, 0, 0, "X", 3));
    bg.addChild( Simple3D.label(-len-0.7, 0, 0, "-X", 3));

    bg.addChild( Simple3D.label(0, 0, len+0.5, "Z", 3));
    bg.addChild( Simple3D.label(0, 0, -len-0.5, "-Z", 3));

    bg.addChild( Simple3D.label(0, len/2+0.2, 0, "Y", 3));
    bg.addChild( Simple3D.label(0, -len/2-0.2, 0, "-Y", 3));

    return bg;
  }  // end of makeAxes()



  private BranchGroup line(double x1, double y1, double z1,
                           double x2, double y2, double z2)
  // a line from (x1, y1, z1) to (x2, y2, z2)
  { 
    Point3d[] coords = {
           new Point3d(x1, y1, z1),  new Point3d(x2, y2, z2) };

    GeometryArray geom = new LineStripArray(coords.length,
                               LineArray.COORDINATES,
                                new int[] { coords.length});
    geom.setCoordinates(0, coords);

    Appearance app = Simple3D.color("white");
    Shape3D shape = new Shape3D(geom, app);

    BranchGroup bg = new BranchGroup();
    bg.addChild(shape);
    return bg;
  }



  // ================== static utility methods ===================

  /*  The purpise of these methods is to hide some of the complexity
      of Java3D for common use-cases.
  */

  // =============================================================


  public static void pause(int ms)
  {
    try {
      Thread.sleep(ms);
    }
    catch(InterruptedException ex) { }
  }  // end of pause()



  public static String tupleStr(Tuple3d t)
  { 
     if (t == null)
       return "null";
     else
       return String.format("( %.3f, %.3f, %.3f )", t.x, t.y, t.z);
  }



  public static void printTuple(String id, Tuple3d t)
  {  System.out.println(id + ": " + tupleStr(t));  }  



  public static void printMatrix(TransformGroup tg)
  {
    Transform3D t3d = new Transform3D();
    tg.getTransform(t3d);
    printMatrix(t3d);
  }

  public static void printMatrix(Transform3D t3d)
  {
     double[] mat = new double[16]; 
     t3d.get(mat);  // stored row-by-row

     for(int row = 0; row < 4; row++){
       System.out.print("| ");
       for(int col = 0; col < 4; col++)
         System.out.printf("%5.2f ", mat[row*4 + col]);
       System.out.println(" |");
     }
     System.out.println();
  }  // end of printMatrix()




  // ----------------------- translation  -----------
  /* These methods fall into three broad groups -- 
         1. those that create a new TransformGroup (TG); "set" methods;
         2. those that overwrite an existing one; "To" methods;
         3. those that incrementally change an existing one; "By" methods;
     In all cases the Transform3D inside the
     TG is hidden from the user of the method.
  */


  public static TransformGroup setTranslation(double x, double y, double z)
  {  return setTranslation( new Vector3d(x,y,z));  }

  public static TransformGroup setTranslation(Vector3d v)
  // Create a TransformGroup for the position
  {
    Transform3D t3d = new Transform3D();
    t3d.set(v);
    return createTG(t3d);
  }



  public static void moveTo(TransformGroup tg, double x, double y, double z)
  {  moveTo(tg, new Vector3d(x,y,z));  }


  public static void moveTo(TransformGroup tg, Vector3d pos)
  // Change the position stored in an existing TransformGroup
  {
    Transform3D t3d = new Transform3D();
    tg.getTransform(t3d);
    // t3d.set(pos);   // overwrite matrix
    t3d.setTranslation(pos);   // replace translation only
    tg.setTransform(t3d);
  }  // end of moveTo()



  public static void moveBy(TransformGroup tg, double dx, double dy, double dz)
  // Reposition the TransformGroup by offsets dx, dy and dz
  {
    moveBy(tg, new Vector3d(dx, dy, dz));  
  }  // end of moveBy()


  private static void moveBy(TransformGroup tg, Vector3d offset)
  // Move the TransformGroup by the offset amount
  {
    Transform3D t3d = new Transform3D();
    tg.getTransform(t3d);

    Transform3D moveT3d = new Transform3D();
    moveT3d.setTranslation(offset); // only translate
    t3d.mul(moveT3d);   // 'add' translation to existing one
    tg.setTransform(t3d);
  }  // end of moveBy()



  // ------------ rotations ------------------------


  public static TransformGroup rotX(double degrees)
  // Create a TransformGroup for the x-axis rotation
  {
    double rads = Math.toRadians(degrees);
    Transform3D t3d = new Transform3D();
    t3d.rotX(rads);
    return createTG(t3d);
  }

  public static TransformGroup rotY(double degrees)
  // Create a TransformGroup for the y-axis rotation
  {
    double rads = Math.toRadians(degrees);
    Transform3D t3d = new Transform3D();
    t3d.rotY(rads);
    return createTG(t3d);
  }

  public static TransformGroup rotZ(double degrees)
  // Create a TransformGroup for the z-axis rotation
  {
    double rads = Math.toRadians(degrees);
    Transform3D t3d = new Transform3D();
    t3d.rotZ(rads);
    return createTG(t3d);
  }



  public static void rotXTo(TransformGroup tg, double degrees)
  // change the rotation to be degrees around the x-axis
  {
    Transform3D rotT3d = new Transform3D();
    rotT3d.rotX( Math.toRadians(degrees) );
    tg.setTransform(rotT3d);
  }


  public static void rotYTo(TransformGroup tg, double degrees)
  // change the rotation to be degrees around the y-axis
  {
    Transform3D rotT3d = new Transform3D();
    rotT3d.rotY( Math.toRadians(degrees) );
    tg.setTransform(rotT3d);
  }


  public static void rotZTo(TransformGroup tg, double degrees)
  // change the rotation to be degrees around the z-axis
  {
    Transform3D rotT3d = new Transform3D();
    rotT3d.rotZ( Math.toRadians(degrees) );
    tg.setTransform(rotT3d);
  }



  public static void rotXBy(TransformGroup tg, double degrees)
  // rotate the existing TransformGroup by degrees around the x-axis
  {
    Transform3D t3d = new Transform3D();
    tg.getTransform(t3d);

    Transform3D rotT3d = new Transform3D();
    rotT3d.rotX( Math.toRadians(degrees) );
    t3d.mul(rotT3d);

    tg.setTransform(t3d);
  }  // end of rotXBy()


  public static void rotYBy(TransformGroup tg, double degrees)
  // rotate the existing TransformGroup by degrees around the y-axis
  {
    Transform3D t3d = new Transform3D();
    tg.getTransform(t3d);

    Transform3D rotT3d = new Transform3D();
    rotT3d.rotY( Math.toRadians(degrees) );
    t3d.mul(rotT3d);

    tg.setTransform(t3d);
  }  // end of rotYBy()


  public static void rotZBy(TransformGroup tg, double degrees)
  // rotate the existing TransformGroup by degrees around the z-axis
  {
    Transform3D t3d = new Transform3D();
    tg.getTransform(t3d);

    Transform3D rotT3d = new Transform3D();
    rotT3d.rotZ( Math.toRadians(degrees) );
    t3d.mul(rotT3d);

    tg.setTransform(t3d);
  }  // end of rotZBy()



  public static TransformGroup setRotations(double xDegree,
                          double yDegree, double zDegree)
  // Create a TransformGroup using the three axes rotations
  {  return setRotations( new Vector3d(xDegree, yDegree, zDegree));  }


  public static TransformGroup setRotations(Vector3d angles)
  /* The three rotations are applied first about the X, 
     then Y, then the Z axis. The orientation of the Y 
     axis is not affected by the X rotation, and the 
     orientation of the Z axis is not affected 
     by the X or Y rotation. 
  */
  { 
    angles.scale(Math.PI/180);    // convert to radians

    Transform3D t3d = new Transform3D();
    t3d.setEuler(angles);
    return createTG(t3d);
  }



  public static void rotsTo(TransformGroup tg, 
                   double xDegree, double yDegree, double zDegree)
  // change the  TransformGroup to be the specified degrees
  {  rotsTo(tg, new Vector3d(xDegree, yDegree, zDegree));  }


  public static void rotsTo(TransformGroup tg, Vector3d angles)
  { 
    angles.scale(Math.PI/180);    // convert to radians
    Transform3D rotT3d = new Transform3D();
    rotT3d.setEuler(angles);
    tg.setTransform(rotT3d);
  }



  public static void rotsBy(TransformGroup tg, 
                   double xDegree, double yDegree, double zDegree)
  // rotate the existing TransformGroup by the specified degrees
  {  rotsBy(tg, new Vector3d(xDegree, yDegree, zDegree));  }


  public static void rotsBy(TransformGroup tg, Vector3d angles)
  { 
    angles.scale(Math.PI/180);    // convert to radians
    Transform3D rotT3d = new Transform3D();
    rotT3d.setEuler(angles);

    Transform3D t3d = new Transform3D();
    tg.getTransform(t3d);
    t3d.mul(rotT3d);

    tg.setTransform(t3d);

  }



  public static TransformGroup setRotationAxis(double x, double y,
                                    double z, double degrees)
  /* Create a TransformGroup of degree rotations around the
     axis defined by the vector (x,y,z)
  */
  {  return setRotationAxis( new Vector3d(x,y,z), degrees);  }


  public static TransformGroup setRotationAxis(Vector3d axis, double degrees)
  { 
    double rads = Math.toRadians(degrees);
    Vector3d normalAxis = new Vector3d();
    normalAxis.normalize(axis);
    AxisAngle4d aa = new AxisAngle4d(normalAxis, rads);

    Transform3D t3d = new Transform3D();
    t3d.setRotation(aa);

    return createTG(t3d);
  }



  public static Quat4f aa2Quat(double x, double y, double z, double degrees)
  /* Create a Quaternion of degree rotations around the
     axis defined by the vector (x,y,z). This is probably the
     easiest way to create a Quaternion from scratch.
  */
  {
    double rads = Math.toRadians(degrees);
    Vector3d axis = new Vector3d(x,y,z);
    Vector3d normalAxis = new Vector3d();
    normalAxis.normalize(axis);
    AxisAngle4d aa = new AxisAngle4d(normalAxis, rads);

    Quat4f quat = new Quat4f();
    quat.set(aa);
    return quat;
  }


  public static TransformGroup setTransforms(
                    double x, double y, double z,
                    double xDegree, double yDegree, double zDegree,
                    double scaleFactor)
  {
    Vector3d trans = new Vector3d(x, y, z);

    Vector3d angles = new Vector3d(xDegree, yDegree, zDegree);
    angles.scale(Math.PI/180);    // convert to radians
    Matrix3d rotMat = eulerToMatrix(angles);

    Transform3D t3d = new Transform3D(rotMat, trans, scaleFactor);
             // translation first, then rotation;
             // The rotations are applied in the order x-, then y-, then z-
             // the scaling is applied only to the rotational component;

    return createTG(t3d);
  }  // end of setTransforms()



  public static Matrix3d eulerToMatrix(Vector3d v) 
  /* Convert Euler rotations (in radians) to a matrix.
     See https://www.learnopencv.com/rotation-matrix-to-euler-angles/
  */
  {
    double cosx = Math.cos(v.x);
    double cosy = Math.cos(v.y);
    double cosz = Math.cos(v.z);
    
    double sinx = Math.sin(v.x);
    double siny = Math.sin(v.y);
    double sinz = Math.sin(v.z);

    Matrix3d rotX = new Matrix3d( 1,    0,     0,
                                  0, cosx, -sinx,
                                  0, sinx,  cosx);

    Matrix3d rotY = new Matrix3d(cosy, 0, siny,
                                    0, 1,    0,
                                -siny, 0, cosy);

    Matrix3d rotZ = new Matrix3d(cosz, -sinz, 0,
                                 sinz,  cosz, 0,
                                    0,     0, 1);

    Matrix3d rot = new Matrix3d();
    Matrix3d interim = new Matrix3d();
    interim.mul(rotY, rotX);
    rot.mul(rotZ, interim);    // rot = rotZ * rotY * rotX
    return rot;
  }  // end of eulerToMatrix()





  // ----------- scaling ---------------------


  public static TransformGroup scale(double scaleFactor)
  // Create a TransformGroup for the scaling
  {
    Transform3D t3d = new Transform3D();
    t3d.setScale(scaleFactor);
    return createTG(t3d);
  }


  private static void scale(TransformGroup tg, double scaleFactor)
  // Set the TransformGroup to a new scaling
  {
    Transform3D t3d = new Transform3D();
    tg.getTransform(t3d);
    t3d.setScale(scaleFactor);   // overwrite old scaling
    tg.setTransform(t3d);
  }  // end of scale()




  // ------------ orbit animation -----------------------
  /* The aim here is to hide the setup of the Alpha, and to
     limit the number of arguments that need to be supplied
     to the RotationInterpolator
  */


  public static TransformGroup orbit(Node obj, int period)
  {  return orbit(obj, period, Y_AXIS);   } 


  public static TransformGroup orbit(Node obj, int period, int axisType)
  /* Return a new TransformGroup from which hangs the supplied
     node and the rotator:

       tg --> obj
        ^
        |
        ----> rotator around axisType
  */
  {
    Transform3D axisT3d = new Transform3D();  // along the +y-axis (default)

    if (axisType == Z_AXIS) 
      axisT3d.rotX( Math.PI/2);   // along +z-axis
    else if (axisType == -Z_AXIS) 
      axisT3d.rotX( Math.PI/2);   // along -z-axis

    else if (axisType == X_AXIS) 
      axisT3d.rotZ( -Math.PI/2);   // along +x-axis
    else if (axisType == -X_AXIS) 
      axisT3d.rotZ( Math.PI/2);   // along -x-axis

    else if (axisType == -Y_AXIS) 
      axisT3d.rotX( Math.PI);   // along -y-axis


    TransformGroup tg = Simple3D.createTG();
    tg.addChild(obj);

    Alpha alpha = new Alpha(-1, period);
    alpha.setStartTime( System.currentTimeMillis()); // start now

    RotationInterpolator rotator = new RotationInterpolator(alpha, tg); 
    rotator.setTransformAxis(axisT3d);
         // no min or max angle is specified, so the rotation is one complete turn
    rotator.setSchedulingBounds(Simple3D.INFINITE_BOUNDS);

    tg.addChild(rotator);

    return tg;
  }  // end of orbit()




  // -------------- get info from a TransformGroup -----------
  /* the methods split into those that get data specified in
     the coordinate system local to the node, and data converted
     to world coordinates (those used by the user).

     This information can be displayed in a window by using
     a TGViewer object.
  */


  public static Vector3d getPosition(TransformGroup tg)
  {        
    Transform3D t3d = new Transform3D();
    tg.getTransform(t3d);
    Vector3d trans = new Vector3d();
    t3d.get(trans);
    return new Vector3d( trans.x, trans.y, trans.z);
  } // end of getPosition()


  public static Vector3d getWorldPosition(TransformGroup tg)
  {        
    Transform3D t3d = new Transform3D();
    tg.getLocalToVworld(t3d);
    Vector3d trans = new Vector3d();
    t3d.get(trans);
    return new Vector3d( trans.x, trans.y, trans.z);
  } // end of getWorldPosition()



  public static Vector3d getRotations(TransformGroup tg)
  {
    Transform3D t3d = new Transform3D();
    tg.getTransform(t3d);
    Matrix3d mat = new Matrix3d();
    t3d.get(mat);
    return mat2Point(mat);
  }  // end of getRotations()


  private static Vector3d mat2Point(Matrix3d mat)
  /*  Matrix has the form:
            m00 m01 m02
            m10 m11 m12 
            m20 m21 m22
    
      Extract the X-, Y-, and Z- rotations (A, B, C),
      from the matrix mat of the following form
      (assuming the rotation order Rz x Ry x Rx) :

        cosCcosB  —sinCcosA+cosCsinBsinA   sinCsinA+cosCsinBcosA
        sinCcosB   cosCcosA+sinCsinBsinA  -cosCsinA+sinCsinBcosA
       -sinB       cosBsinA                cosBcosA
    https://mino-git.github.io/rtcw-wet-blender-model-tools/publications/EulerToMatrix.pdf
  */
  {
    double xA, yB, zC;
    yB = -Math.asin(mat.m20);    // the y-axis rotation (B)
    double cosB = Math.cos(yB);

    if (Math.abs(cosB) > 0.005) {
      xA = -Math.atan2(-mat.m21/cosB, mat.m22/cosB);  
               //atan of sin A / cos A
      zC = -Math.atan2(-mat.m10/cosB, mat.m00/cosB);  
               //atan of sin C / cos C
    }
    else {  // cosB is too small
      xA = 0;
      zC = -Math.atan2(mat.m01, mat.m11);
    }

    xA = Math.toDegrees(xA);
    yB = Math.toDegrees(yB);
    zC = Math.toDegrees(zC);
    return new Vector3d(xA, yB, zC);
  }  // end of mat2Point()


  public static Vector3d getWorldRotations(TransformGroup tg)
  {        
    Transform3D t3d = new Transform3D();
    tg.getLocalToVworld(t3d);
    Matrix3d mat = new Matrix3d();
    t3d.get(mat);
    return mat2Point(mat);
  } // end of getWorldRotations()


  public static Vector3d getScale(TransformGroup tg)
  {  
    Transform3D t3d = new Transform3D();
    tg.getTransform(t3d);
    Vector3d scale = new Vector3d();
    t3d.getScale(scale);
    return new Vector3d( scale.x, scale.y, scale.z);
  }  // end of getScale()


  public static Vector3d getWorldScale(TransformGroup tg)
  {  
    Transform3D t3d = new Transform3D();
    tg.getLocalToVworld(t3d);
    Vector3d scale = new Vector3d();
    t3d.getScale(scale);
    return new Vector3d( scale.x, scale.y, scale.z);
  } // end of getWorldScale()




  // ---------------- simplified node creation ---------------
  /* These methods mostly hide the setting of capabilities
     which let the nodes be read and written after they've
     been added to the Java3D scenegraph
  */



  public static BranchGroup createBG()
  {
    BranchGroup bg = new BranchGroup();
    bg.setCapability(BranchGroup.ALLOW_CHILDREN_READ);
    bg.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
    bg.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
    bg.setCapability(BranchGroup.ALLOW_DETACH);
    return bg;
  }

  public static BranchGroup createBG(Node node)
  {
    BranchGroup bg = new BranchGroup();
    bg.setCapability(BranchGroup.ALLOW_CHILDREN_READ);
    bg.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
    bg.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
    bg.setCapability(BranchGroup.ALLOW_DETACH);
    bg.addChild(node);
    return bg;
  }


  public static TransformGroup createTG()
  {
    TransformGroup tg = new TransformGroup();
    tg.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
    tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
    return tg;
  }


  public static TransformGroup createTG(Transform3D t1)
  {
    TransformGroup tg = new TransformGroup(t1);
    tg.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
    tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
    return tg;
  }


  private static Shape3D createShape3D(Geometry geom)
  {
    Shape3D s3d = new Shape3D(geom);
    s3d.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
    s3d.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
    s3d.setCapability(Shape3D.ALLOW_APPEARANCE_OVERRIDE_READ);
    s3d.setCapability(Shape3D.ALLOW_APPEARANCE_OVERRIDE_WRITE);
    return s3d;
  }


  private static Shape3D createShape3D(Geometry geom, Appearance app)
  {
    Shape3D s3d = new Shape3D(geom, app);
    s3d.setCapability(Shape3D.ALLOW_APPEARANCE_READ);
    s3d.setCapability(Shape3D.ALLOW_APPEARANCE_WRITE);
    s3d.setCapability(Shape3D.ALLOW_APPEARANCE_OVERRIDE_READ);
    s3d.setCapability(Shape3D.ALLOW_APPEARANCE_OVERRIDE_WRITE);
    return s3d;
  }


  public static BranchGroup label(double x, double y, double z, String text)
  {  return label(x, y, z, text, "white", 2);  }


  public static BranchGroup label(double x, double y, double z, 
                                    String text, int fontSize)
  {  return label(x, y, z, text, "white", fontSize);  }


  public static BranchGroup label(double x, double y, double z, 
                      String text, String colorNm, int fontSize)
  // draw this text centered at (x,y,z) using a reduced font size
  {
    Font3D f3d = new Font3D(
            new Font("Helvetica", Font.PLAIN, fontSize),
            new FontExtrusion());

    Text3D text3D = new Text3D(f3d, text, new Point3f(0,0,0),
                           Text3D.ALIGN_CENTER, Text3D.PATH_RIGHT);

    Shape3D shape = new Shape3D(text3D, Simple3D.color(colorNm));

    Transform3D t3d = new Transform3D();
    t3d.setTranslation(new Vector3d(x,y,z));
    t3d.setScale(0.07);    // shrink the text using a magic number
    TransformGroup tg = new TransformGroup(t3d);
    tg.addChild(shape);

    BranchGroup bg = new BranchGroup();
    bg.addChild(tg);
    return bg;
  }  // end of label()




  public static BranchGroup text3D(String text)
  {  return text3D(text, "yellow");  }


  public static BranchGroup text3D(String text, String colorNm)
  /*
        bg --> Shape --> text
                 |
                 -----> colored appearance

  */
  {
    Font3D f3d = new Font3D(new Font("Helvetica", Font.PLAIN, 2),
                                     new FontExtrusion());

    Text3D text3D = new Text3D(f3d, text, new Point3f(0,0,0), 
                           Text3D.ALIGN_CENTER, Text3D.PATH_RIGHT);

    Shape3D shape = new Shape3D(text3D, Simple3D.color(colorNm));

    BranchGroup bg = new BranchGroup();
    bg.addChild(shape);
    return bg;
  }




  public static Shape3D tetra(double width, double height,
                                    double length, String colorNm)
  {
    return new Shape3D( tetra(width, height, length), 
                        Simple3D.color(colorNm) );
  }


  public static Shape3D texTetra(double width, double height, 
                                    double length, String texFnm)
  {
    return new Shape3D( tetra(width, height, length), 
                        Simple3D.texture(texFnm));
  }



  public static GeometryArray tetra(double width, 
                                 double height, double length)
  /* a triangular tetrahedron of height, -width to width on the x-axis,
     and length along the z-axis
  */
  {
    GeometryInfo gi = new GeometryInfo(GeometryInfo.TRIANGLE_ARRAY);

    Point3d[] pts = new Point3d[4];
    pts[0] = new Point3d(0, 0, length);
    pts[1] = new Point3d(-width, 0, 0);
    pts[2] = new Point3d(width, 0, 0);
    pts[3] = new Point3d(0, height, 0);
    gi.setCoordinates(pts);

    int[] indices = {0,1,2, 0,3,1, 0,2,3, 2,1,3};
                   // base, left,  right,  back    in ccw order
    gi.setCoordinateIndices(indices);

    NormalGenerator ng = new NormalGenerator();
    ng.generateNormals(gi);

    return gi.getGeometryArray();
  }  // end of tetra()




  // ----------------- color and textures ------------------------
  /* Helper methods for creating Appearance nodes for coloring 
     or texturing the geometries they are associated with
   */


  public static Color3f toColor3f(Color c)
  // convert a Java Color to a Java3D Color3f
  {
    float x = (c.getRed() / 255.0F);
    float y = (c.getGreen() / 255.0F);
    float z = (c.getBlue() / 255.0F);
    return new Color3f(x, y, z);
  }


  public static Sphere texSphere(float radius, String texFnm)
  {
    Appearance app = texture(texFnm, false);   
                // texture() does not need to generate tex coords
    return new Sphere(radius, Primitive.GENERATE_TEXTURE_COORDS |
                              Primitive.GENERATE_NORMALS, 50, app);
  }


  public static Cylinder texCylinder(float radius, float height, String texFnm)
  {
    Appearance app = texture(texFnm, false);
    return new Cylinder(radius, height, Primitive.GENERATE_TEXTURE_COORDS |
                                        Primitive.GENERATE_NORMALS, app);
  }


  public static Cone texCone(float radius, float height, String texFnm)
  {
    Appearance app = texture(texFnm, false);
    return new Cone(radius, height, Primitive.GENERATE_TEXTURE_COORDS |
                                    Primitive.GENERATE_NORMALS, app);
  }


  public static org.jogamp.java3d.utils.geometry.Box 
                       texBox(float x, float y, float z, String texFnm)
  {
    Appearance app = texture(texFnm, false);
    return new org.jogamp.java3d.utils.geometry.Box(
                   x, y, z, Primitive.GENERATE_TEXTURE_COORDS |
                            Primitive.GENERATE_NORMALS, app);
  }



  public static Appearance color(String colorNm)
  {
    Color3f col = ColorFactory.getColor(colorNm);
    Appearance app = new Appearance();
    app.setCapability(Appearance.ALLOW_MATERIAL_READ);
    app.setCapability(Appearance.ALLOW_MATERIAL_WRITE);

    // set culling to NONE (show both sides of shapes)
    PolygonAttributes pa = new PolygonAttributes();
    pa.setCullFace(PolygonAttributes.CULL_NONE);
    app.setPolygonAttributes(pa);

    ColoringAttributes ca = new ColoringAttributes(col, 
                                   ColoringAttributes.NICEST);
    app.setColoringAttributes(ca);

    Material mat = new Material(col, ColorFactory.getColor("blackish"),
                                col,  ColorFactory.getColor("gray"),  64f);
                        // sets ambient, emissive, 
                        //     diffuse, specular, shininess
    mat.setCapability(Material.ALLOW_COMPONENT_READ);
    mat.setCapability(Material.ALLOW_COMPONENT_WRITE);
    mat.setCapability(Material.AMBIENT_AND_DIFFUSE);

    app.setMaterial(mat);

    return app;
  }  // end of color()



  public static Appearance texture(String fnm)
  {  return texture(fnm, true);  }  // generate texture coordinates


  public static Appearance texture(String fnm, boolean needsTexCoords)
  {  return texture(loadTexture(fnm), needsTexCoords);  } 



  public static Texture2D loadTexture(String fnm)
  {
    TextureLoader loader = null;  
    try {
       loader = new TextureLoader(fnm, null);
    }
    catch(Exception e) 
    { System.out.println(e); }
       
    if (loader == null) {
      System.out.println("Unable to load texture from " + fnm);
      return null;
    }
    else
      return (Texture2D)loader.getTexture();
  }  // end of loadTexture()



  public static Appearance texture(Texture2D tex, boolean needsTexCoords)
  /*  A textured apearance requires:
          * a 2D texture image (loaded from a file)
          * texture attributes to say how it combines with face colors
          * tex coord generation (optional) if the geometry
            doesn't include texture coordinates
  */
  {
    Appearance app = new Appearance();
    app.setCapability(Appearance.ALLOW_MATERIAL_READ);
    app.setCapability(Appearance.ALLOW_MATERIAL_WRITE);

    // set culling to NONE (show both sides of shapes)
    PolygonAttributes pa = new PolygonAttributes();
    pa.setCullFace(PolygonAttributes.CULL_NONE);
    app.setPolygonAttributes(pa);

    if (tex == null)
      System.out.println("Texture is null");
    else
      app.setTexture(tex);

    // combine texture and lighting
    TextureAttributes ta = new TextureAttributes();
    ta.setTextureMode(TextureAttributes.COMBINE);
    app.setTextureAttributes(ta);

    Color3f col = ColorFactory.getColor("white");
    Color3f black = ColorFactory.getColor("black");
    Material mat = new Material(col,black, col,col,25f);
    app.setMaterial(mat);

    if (needsTexCoords) {
      TexCoordGeneration tcg =  new TexCoordGeneration(
                     TexCoordGeneration.OBJECT_LINEAR,
                     TexCoordGeneration.TEXTURE_COORDINATE_2);
      tcg.setPlaneS( new Vector4f(1,0,1,0));
      tcg.setPlaneT( new Vector4f(0,1,1,0));
      app.setTexCoordGeneration(tcg);
    }

    return app;
  }  // end of texture()



  public static Texture2D strToTex(String msg)
  // convert a string into a texture
  {
    int texSize = calcTextureSize(msg);

    BufferedImage bi = new BufferedImage(texSize, texSize,
                                BufferedImage.TYPE_INT_ARGB);
    Graphics2D g2 = (Graphics2D)bi.getGraphics();
    g2.setColor(Color.WHITE);
    g2.fillRect(0, 0, texSize, texSize);

    g2.setFont(texFont);
    g2.setColor(Color.BLACK);
    drawCenteredString(msg, texSize, texSize, g2);

    ImageComponent2D im = new ImageComponent2D( 
                           ImageComponent2D.FORMAT_RGBA, bi);

    Texture2D texture = new Texture2D(Texture.BASE_LEVEL, Texture.RGBA,
                                      texSize, texSize);
    texture.setImage(0, im);
    texture.setMagFilter(Texture.BASE_LEVEL_LINEAR);

    return texture;
  }  // end of strToTex()


  private static int calcTextureSize(String msg)
  {
    Canvas c = new Canvas();
    FontMetrics fm = c.getFontMetrics(texFont);
    int width = fm.stringWidth(msg);
    int size = 60;   // magic numbers
    while (size <= width)
      size += 20;
    return size;
  }
  


  private static void drawCenteredString(String s, int w, int h, Graphics2D g2) 
  {
    FontMetrics fm = g2.getFontMetrics();
    int x = (w - fm.stringWidth(s)) / 2;
    int y = (fm.getAscent() + (h - (fm.getAscent() + fm.getDescent())) / 2);
    g2.drawString(s, x, y);
  }



  // ----- change color / texture of shape(s) in a branch ----------------


  public static void colorShape(Node node, String colorNm)
  {
    Appearance app = Simple3D.color(colorNm);
    appShapes(node, app, false);   // not a texture
  }  


  public static void textureShape(Node node, String fnm)
  {
    Appearance app = Simple3D.texture(fnm);
    appShapes(node, app, true);
  }  


  private static void appShapes(Node node, Appearance app, boolean isTexture)
  /*  If the node is a Group then recursively visit its children.
      Otherwise if the node is a Shape3D then apply the new appearance.
  */
  {
    if(node instanceof Group) {
      Group g = (Group) node;
      Iterator<Node> kids = g.getAllChildren();
      while(kids.hasNext())   // visit children
        appShapes(kids.next(), app, isTexture);
    }
    else if (node instanceof Shape3D) {
      Shape3D shape = (Shape3D) node;
      if (isTexture) {
        if (!singleGeom(shape)) {
          System.out.println("Shape cannot be textured");
          return;
        }
      }
      shape.setAppearance(app);
    }
  }  // end of appShapes()



  private static boolean singleGeom(Shape3D shape)
  // Is the shape a single GeometryArray?
  { if (shape.numGeometries() == 1) {
      Geometry g = shape.getGeometry();
      if (g instanceof GeometryArray) 
        return true;
    }
    return false;
  } 


  // --------- picking -------------------------


  public static String pickID(PickResult pickResult)
  {
    if (pickResult == null)
      return null;

    String id = getPrimitiveID(pickResult);
    if (id == null)
      id = getShapeID(pickResult);
    return id;
  } // end of pickID()


  private static String getPrimitiveID(PickResult pickResult)
  // extract ID from a Primitive object in pickResult
  {
    Primitive pickedShape = 
             (Primitive) pickResult.getNode(PickResult.PRIMITIVE);
    if (pickedShape == null)
      return null;
    else {
      String id = pickedShape.getName();
      if (id == null)
        return "primitive";
      else
        return id;
    }
  } // end of getPrimitiveID()



  private static String getShapeID(PickResult pickResult)
  // extract ID from a Shape3D object in pickResult
  {
    Shape3D pickedShape = 
             (Shape3D) pickResult.getNode(PickResult.SHAPE3D);
    if (pickedShape == null)
      return null;
    else {
      String id = pickedShape.getName();
      if (id == null)
        return "shape3D";
      else
        return id;
    }
  } // end of getShapeID()



  // loading OBJ models -----------------------

  /* Loads an OBJ model using Java3D's rather old and simple loader,
     and optionally suspends the shape before a TransformGroup that
     place it at the origin, facing along the +z axis. 

     This information is loaded from a "XXXZero.txt" configuration file 
     which the user places in the folder with the XXX ".obj" file.
  */  


  public static BranchGroup loadModel(String fnm)
  { 
    Vector3d trans = new Vector3d();
    Vector3d angles = new Vector3d();
    Vector3d scale = new Vector3d();

    boolean configFound = loadZeroConfig(stripExt(fnm) + "Zero.txt", 
                                       trans, angles, scale);
    if (configFound)
      return loadModel(fnm, trans, angles, scale);
    else
      return loadBasicModel(fnm);
  }


  private static String stripExt(String fnm)
  {
    int dotPos = fnm.lastIndexOf('.');
    if (dotPos == -1)
      return fnm;
    else
      return fnm.substring(0, dotPos);
  }



  private static BranchGroup loadModel(String fnm, Vector3d toMove,
                  Vector3d angles, Vector3d scale)
  /* Returns:
        bg --> tg --> modelBG
  */
  { 
    BranchGroup modelBG = loadBasicModel(fnm);
    if (modelBG == null)
      return null;

    TransformGroup tg = new TransformGroup();
    tg.addChild(modelBG);

    // create a transform group for the object
    // rotate then scale then translate
    Transform3D t3d = new Transform3D();
    tg.getTransform(t3d);
    angles.scale(Math.PI/180);   // convert degrees to radians
    t3d.setEuler(angles);   // rotX, rotY, rotZ
    t3d.setScale(scale);
    t3d.setTranslation(toMove);   // x, y, z
    tg.setTransform(t3d);
    
    BranchGroup bg = new BranchGroup();
    bg.addChild(tg);
    return bg;
  }  // end of loadModel()



  public static BranchGroup loadBasicModel(String fnm)
  /* load the OBJ model stored in fnm

     The most common problem I've found with this old OBJ
     loader is it not understanding the "o" OBJ command. The solution
     is to open the OBJ file in a text editor, and comment out
     those lines. (Add # to the start of the line.)

     Another problem is that the OBJ file may refer to its MTL file
     with the wrong path/filename, and the MTL file may refer to its JPG
     files with the wrong path/filename. Fortunately both the OBJ and
     MTL files are text, so easily edited.
  */
  {
    // System.out.println("Loading OBJ model from " + fnm);
    File file = new File(fnm);
    if (!file.isFile()) {
      System.out.println("Could not find " + fnm);
      return null;
    }

    /* Convert the filename to a URL, so the OBJ file can 
       find MTL and image files in the Models/ subdirectory
       at runtime. */
    URL url = null;
    try {
      url = file.toURI().toURL();
    }
    catch(Exception e) {
      System.out.println(e);
      return null;
    }

    // read in the geometry from the file
    int params = ObjectFile.RESIZE ; // | Loader.LOAD_ALL;
    ObjectFile loader = new ObjectFile(params);

    Scene scene = null;
    try {
      scene = loader.load(url);
    }
	catch (FileNotFoundException e) {
	  System.out.println("Could not find " + fnm);
	  return null;
	}
	catch (ParsingErrorException e) {
	  System.out.println("Could not parse the contents of " + fnm);
	  System.out.println(e.getMessage());
	  return null;
	}
	catch (IncorrectFormatException e) {
	  System.out.println("Incorrect format in " + fnm);
	  System.out.println(e);
	  return null;
	}

    // return the model's BranchGroup
    if (scene == null) {
      System.out.println("Loaded scene is null");
      return null;
    }

    BranchGroup modelBG = scene.getSceneGroup();

    modelBG.setCapability( BranchGroup.ALLOW_BOUNDS_READ );
    modelBG.setCapability( BranchGroup.ALLOW_CHILDREN_READ );

    // fix face culling problem and set user data for picking
    for (int i = 0; i < modelBG.numChildren(); i++) {
      Node child = modelBG.getChild(i);
      if (child instanceof Shape3D) {
        Shape3D shape = (Shape3D) child;
        
        shape.setName(fnm);
/*
        PickTool.setCapabilities(shape, PickTool.INTERSECT_FULL);
        shape.setPickable(true);
        shape.setCapability(Node.ENABLE_PICK_REPORTING);
        shape.setCapability(Node.ALLOW_PICKABLE_READ);
        shape.setCapability(Shape3D.ALLOW_GEOMETRY_READ);

        Geometry geom = shape.getGeometry();
        geom.setCapability(Geometry.ALLOW_INTERSECT);
*/
        Appearance ap = shape.getAppearance();
        ap.setCapability(Appearance.ALLOW_MATERIAL_READ);
        ap.setCapability(Appearance.ALLOW_MATERIAL_WRITE);
        PolygonAttributes pa = ap.getPolygonAttributes();
        if (pa == null)
          pa = new PolygonAttributes();
        pa.setCullFace(PolygonAttributes.CULL_NONE);
        ap.setPolygonAttributes(pa);
      }
    }

    // listNamedObjects(scene);

    return modelBG;
  }  // end of loadBasicModel()




  private static boolean loadZeroConfig(String fnm,
           Vector3d trans, Vector3d angles, Vector3d scale)
  /* The format of the input lines are:
        pos:  double double double
        rots: double double double  (in degrees)
        scale: scale
     and blank lines and comment lines.
  */
  { 
    File file = new File(fnm);
    if (!file.isFile())
      return false;

    boolean hasConfig = true;
    try {
      System.out.println("Loading config file: " + fnm);
      BufferedReader br = new BufferedReader( new FileReader(fnm));
      String line;
      char ch;
      while((line = br.readLine()) != null) {
        line = line.trim();
        if (line.length() == 0)  // blank line
          continue;
        if (line.startsWith("//"))   // comment
          continue;

        if (line.startsWith("pos:"))
          trans.set( getCoord( line.substring(4).trim() ));
        else if (line.startsWith("rots:"))
          angles.set( getCoord(line.substring(5).trim() ));
        else if (line.startsWith("scale:")) {
          double s = getDouble( line.substring(6).trim() );
          scale.set(s,s,s);   // uniform scaling
        }
      }
      br.close();
    } 
    catch (IOException e) 
    { 
      System.out.println("  " + e);
	  hasConfig = false;
    }
    return hasConfig;
  }  // end of loadZeroConfig()



  private static Point3d getCoord(String line)
  {
    double x = 0;
    double y = 0;
    double z = 0;
    String[] valStr = line.split("\\s+");
    if (valStr.length != 3)
      System.out.println("Wrong number of values (" 
                         + valStr.length + ") in " + line);
    else {
      x = getDouble(valStr[0]);
      y = getDouble(valStr[1]);
      z = getDouble(valStr[2]);
    }
    return new Point3d(x,y, z);
  }  // end of getCoord()



  public static double getDouble(String s)
  { 
    double num = 0;
    try {
      num = Double.parseDouble(s);
    }
    catch (NumberFormatException ex){ 
      System.out.println("Incorrect format for double " + s); 
    }
    return num;
  }  // end of getDouble()


  public static int getInt(String s)
  { 
    int num = 0;
    try {
      num = Integer.parseInt(s);
    }
    catch (NumberFormatException ex){ 
      System.out.println("Incorrect format for int " + s); 
    }
    return num;
  }  // end of getInt()



  private static void listNamedObjects(Scene scene)
  {
    // Hashtable<String,Object> namedObjects = scene.getNamedObjects();
    // Enumeration<String> e = namedObjects.keys();

    Hashtable namedObjects = scene.getNamedObjects();
    if (namedObjects.isEmpty())
      System.out.println("No mamed objects in model");
    else {
      Enumeration e = namedObjects.keys();
      String name;
      System.out.println("Named objects in model:");
      while(e.hasMoreElements()) {
        name = (String)e.nextElement(); 
        Object obj = namedObjects.get(name);
        if (obj instanceof Shape3D) {
          Shape3D shape = (Shape3D) obj;
          System.out.println("  " + name + 
                   ": Shape3D; no. geometries: " + shape.numGeometries());
        }
        else
          System.out.println("  " + name + ": " + obj);
      }
    }
  }  // end of listNamedObjects()



  public static BufferedImage loadImage(String fnm)
  {
    BufferedImage im = null;
    try {
      im = ImageIO.read(new File(fnm));
    } 
    catch (IOException e) 
    {   System.out.println("Could not load image from " + fnm); }
    return im;
  }  // end of loadImage()



  // ----------- shaders ----------------------------
  /* Methods that hide the complexity of loading vertex and
     fragment shaders, and the assignment of values to their
     parameters (attributes).
  */



  public static ShaderAppearance makeShaderApp(
                         String vertexName, String fragmentName)
  {  return makeShaderApp(vertexName, fragmentName, null);  }


  public static ShaderAppearance makeShaderApp(
                         String vertexName, String fragmentName,  
                         Map<String,Object> attributes)
  // Set vertex and fragment shaders in the appearance
  {
    ShaderProgram sp =  Simple3D.loadShaders(vertexName, fragmentName);  
    if (sp == null)
      return null;

    // add attribute names to the shader program
    if (attributes != null)
      Simple3D.nameAttributes(sp, attributes);

    // Set shader program as an appearance
    ShaderAppearance app = new ShaderAppearance();
    Material mat = new Material();
    app.setShaderProgram(sp);
    app.setMaterial(mat);

    // set culling to NONE (show both sides of shapes)
    PolygonAttributes pa = new PolygonAttributes();
    pa.setCullFace(PolygonAttributes.CULL_NONE);
    app.setPolygonAttributes(pa);


    // set attribute values 
    if (attributes != null)
      app.setShaderAttributeSet( 
             Simple3D.assignAttributes(attributes) );

    return app;
  }  // end of makeShaderApp()




  public static ShaderProgram loadShaders(String vertexName, 
                                          String fragmentName)
  // load the vertex and fragment shaders
  {
    int numShaders = 0;

    // read in the shaders
    String vertexProg = null;
    if (vertexName != null) {
      try {
        vertexProg = StringIO.readFully(vertexName);
        numShaders++;
      }
      catch (IOException e) {
        System.out.println("Could not read the vertex shader: " + vertexName);
      }
    }

    String fragmentProg = null;
    if (fragmentName != null) {
      try {
        fragmentProg = StringIO.readFully(fragmentName);
        numShaders++;
      }
      catch (IOException e) {
        System.out.println("Could not read fragment shader: " + fragmentName);
      }
    }

    if (numShaders == 0)
      return null;

    // create the shader programs
    Shader[] shaders = new Shader[numShaders];
    int i = 0;
    if (vertexProg != null) {
      shaders[i] = new SourceCodeShader(Shader.SHADING_LANGUAGE_GLSL,
                           Shader.SHADER_TYPE_VERTEX, vertexProg);
      i++;
    }

    if (fragmentProg != null) {
      shaders[i] = new SourceCodeShader(Shader.SHADING_LANGUAGE_GLSL,
                           Shader.SHADER_TYPE_FRAGMENT, fragmentProg);
      i++;
    }

    ShaderProgram shaderProg = new GLSLShaderProgram();
    try {
      shaderProg.setShaders(shaders);
    }
    catch(Exception e) {  // occurs on Ubuntu sometimes
      System.out.println("Could not link in the shaders");
      return null;
    }
    return shaderProg;
  }  // end of loadShaders()



  public static void nameAttributes(ShaderProgram sp, 
                                        Map<String,Object> attrs)
  {
    if (attrs == null)
      System.out.println("No attributes found");
    else {
      Set<String> nmsSet = attrs.keySet();
      String[] nms = nmsSet.toArray(new String[nmsSet.size()]);
      sp.setShaderAttrNames(nms);
    }
  }  // end of nameAttributes()



  public static ShaderAttributeSet assignAttributes(Map<String,Object> attrs)
  {
    ShaderAttributeSet sas = null;
    if (attrs == null)
      System.out.println("No attribute values found");
    else {
      sas = new ShaderAttributeSet();
      for (Map.Entry<String, Object> attr : attrs.entrySet())
        sas.put( new ShaderAttributeValue(attr.getKey(), attr.getValue()) );
    }
    return sas;
  }  // end of assignAttributes()



  public static void applyShader(Group g, ShaderAppearance app)
  // Recursively set shader for all children of the group
  {
    Iterator<Node> children = g.getAllChildren();
    while (children.hasNext()) {
      Node node = children.next();
      if (node instanceof Group)
        applyShader((Group)node, app);
      else if (node instanceof Shape3D)
        ((Shape3D)node).setAppearance(app);
    }
  } /* applyShader */



  // ----------------- test rig --------------------


  public static void main(String[] args)
  {
    Simple3D j3d = new Simple3D("Simple3D");

    BranchGroup bg = Simple3D.createBG(new ColorCube(0.4));
    j3d.rootBG.addChild(bg);  
            // only a BG can be linked to rootBG since it is live
  } /* main */


}

