
// Dice.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, November 2021

/* Create a textured dice in 4 ways:
     * use Simple3D.texBox()
     * use TextureCubeMap
     * use GeometryInfo
     * use IndexedQuadArray
*/

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;
import org.jogamp.java3d.utils.geometry.*;



public class Dice
{

  public static void main(String[] args)
  {
    Simple3D j3d = new Simple3D("A Dice", false, true);

    // Box dice = Simple3D.texBox(1, 1, 1, "images/diceFaces.png");  // 1
             // with this approach the texture mapping isn't good enough

    // Box dice = diceCubeMap();   // 2


    // Geometry geom = diceGeometryGI();   // 3

    Geometry geom = diceGeometryQuad();   // 4

    Appearance app = Simple3D.texture("images/diceFaces.png", false);
                // texture() does not need to generate tex coords
    Shape3D dice = new Shape3D(geom, app);

    j3d.rootBG.addChild( Simple3D.createBG(dice));
  }  // end of main()




  private static Box diceCubeMap()
  {
    // build the textured box
    Box dice = new Box(0.5f, 0.5f, 0.5f,
                           Box.GENERATE_TEXTURE_COORDS, 
                                         Simple3D.color("green"));

    addPic(dice, Box.RIGHT, "images/pic1.jpg");    
    addPic(dice, Box.LEFT, "images/pic6.jpg");

    // addPic(dice, Box.TOP, "images/pic2.jpg"); 
              // show color if no texture
    addPic(dice, Box.BOTTOM, "images/pic5.jpg");

    addPic(dice, Box.FRONT, "images/pic3.jpg");
    addPic(dice, Box.BACK, "images/pic4.jpg");

    return dice;
  } // end of diceCubeMap()


  private static void addPic(Box b, int partId, String texFnm)
  {
    Shape3D face = b.getShape(partId);
    face.setAppearance( Simple3D.texture(texFnm, false) );
             // texture() should not generate texture coords
  }



  private static GeometryArray diceGeometryGI()
  // using GeometryInfo
  // 6 square faces; 8 vertices
  {
    GeometryInfo gi = new GeometryInfo(GeometryInfo.QUAD_ARRAY);
    
    // 8 vertices
    Point3f[] coords = { 
          new Point3f(0,0,0), new Point3f(1,0,0),
          new Point3f(1,1,0), new Point3f(0,1,0),
          new Point3f(0,0,1), new Point3f(1,0,1),
          new Point3f(1,1,1), new Point3f(0,1,1)
    };
    
    // coordinates defining 6 faces
    int[] coordIndices = {
           0,1,2,3,  0,1,5,4,    // back, bottom
           1,2,6,5,  2,3,7,6,    // right, top
           3,0,4,7,  4,5,6,7     // left, front
    };
    
    gi.setCoordinates(coords);
    gi.setCoordinateIndices(coordIndices);

    // 12 coords of a 3 x 2 grid division of the texture
    TexCoord2f[] tex = {    
          new TexCoord2f(0, 1), new TexCoord2f(1f/3, 1),
          new TexCoord2f(2f/3, 1), new TexCoord2f(1, 1),
          new TexCoord2f(0, 0.5f),new TexCoord2f(1f/3, 0.5f),
          new TexCoord2f(2f/3, 0.5f), new TexCoord2f(1, 0.5f),
          new TexCoord2f(0, 0), new TexCoord2f(1f/3, 0),
          new TexCoord2f(2f/3, 0), new TexCoord2f(1, 0)
    };
    
    // 6 faces from the 3 x 2 grid
    int[] texIndices = {
           0,4,5,1,   1,5,6,2,      // one, two
           2,6,7,3,   5,9,10,6,     // three, five
           6,10,11,7, 8,9,5,4       // four, six  (4,8,9,5)
    };
    
    gi.setTextureCoordinateParams(1, 2); // 1 set of 2d tex coords
    gi.setTextureCoordinates(0, tex);
    gi.setTextureCoordinateIndices(0, texIndices);

    NormalGenerator ng = new NormalGenerator();
    ng.generateNormals(gi);

    return gi.getGeometryArray();
  }  // end of diceGeometryGI()



  private static Geometry diceGeometryQuad() 
  // using IndexedQuadArray
  {
    IndexedQuadArray qa = new IndexedQuadArray(12,
                                QuadArray.COORDINATES|
                                QuadArray.TEXTURE_COORDINATE_2, 24);
    
    Point3f[] coords = { 
          new Point3f(0,0,0), new Point3f(1,0,0),
          new Point3f(1,1,0), new Point3f(0,1,0),
          new Point3f(0,0,1), new Point3f(1,0,1),
          new Point3f(1,1,1), new Point3f(0,1,1)
    };
    
    int[] coordIndices = {
           0,1,2,3,  0,1,5,4, 
           1,2,6,5,  2,3,7,6, 
           3,0,4,7,  4,5,6,7
    };
    
    qa.setCoordinates(0, coords);
    qa.setCoordinateIndices(0, coordIndices);
    
    TexCoord2f[] tex = { 
          new TexCoord2f(0, 1),    new TexCoord2f(1f/3, 1),
          new TexCoord2f(2f/3, 1), new TexCoord2f(1, 1),
          new TexCoord2f(0, 0.5f), new TexCoord2f(1f/3, 0.5f),
          new TexCoord2f(2f/3, 0.5f), new TexCoord2f(1, 0.5f),
          new TexCoord2f(0, 0),    new TexCoord2f(1f/3, 0),
          new TexCoord2f(2f/3, 0), new TexCoord2f(1, 0)
    };
    
    int[] texIndices = {
           0,1,5,4, 1,2,6,5, 
           2,3,7,6, 5,6,10,9, 
           6,7,11,10, 4,5,9,8
    };
    
    qa.setTextureCoordinates(0,0,tex);
    qa.setTextureCoordinateIndices(0,0,texIndices);

    return qa;
  }  // end of diceGeometryQuad()




}  // end of Dice class
