
// MorphCube.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, November 2021

/*  Use a TimeBehavior to continuous morph between three geometries
    in the order
        cube --> truncated pyramid --> cube --> inverted truncated pyramid

    This requires the creation of an object so that
    the class can implement TimeUpdater. Its tick() method
    is called every DELAY ms by the Behavior.

    -----
    Note: the Morph class was deprecated in Java 3D 1.4;
    For a complete deprecation list, see
      https://jogamp.org/deployment/java3d/1.6.0-final/javadoc/deprecated-list.html
   
    As a result, the compilation reports 3 deprecation warnings 
    for the Morph class.
*/

import java.util.*;

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;
import org.jogamp.java3d.utils.geometry.*;



public class MorphCube implements TimeUpdater
{
  private static final int DELAY = 50;
  private static final int NUM_FRAMES = 4;

  private BranchGroup sceneBG;

  // globals for animation
  private Morph morph;
  private Alpha alpha;

  private  Simple3D j3d;



  public MorphCube(Simple3D j3d)
  {
    this.j3d = j3d;    // required so the morph shape can be reliably redrawn

    /* The morph contains 4 'frames':
       morph --> cube
             --> truncated pyramid
             --> cube
             --> inverted truncated pyramid
    */

    GeometryArray[] geoms = new GeometryArray[NUM_FRAMES];
    geoms[0] = createCuboid(cubeCoords());
    geoms[1] = createCuboid(pyraCoords());
    geoms[2] = createCuboid(cubeCoords());
    geoms[3] = createCuboid(ipyraCoords());

    morph = new Morph(geoms, Simple3D.color("gray"));
    morph.setCapability(Morph.ALLOW_WEIGHTS_WRITE);

    alpha = new Alpha(-1, 4000); // 4 sec period before repeating

    sceneBG = Simple3D.createBG(morph);
  }  // end of MorphCube()


  public BranchGroup getScene()
  {  return sceneBG;  }



  private static GeometryArray createCuboid(Point3d[] coords)
  /* Use supplied coordinates to generate faces, and 
     also normals, for the geometry
  */
  {
    GeometryInfo gi = new GeometryInfo(GeometryInfo.QUAD_ARRAY);
    gi.setCoordinates(coords);

    int indices[] = {
           0,1,2,3,    // front (CCW)
           4,5,6,7,    // back
           0,7,6,1,    // right
           3,2,5,4,    // left
           1,6,5,2,    // top
           4,7,0,3     // bottom
    };
    gi.setCoordinateIndices(indices);

    NormalGenerator ng = new NormalGenerator();
    ng.generateNormals(gi);

    return gi.getGeometryArray();
  }  // end of createCuboid()




  private static Point3d[] cubeCoords()
  /* coordinates for a 1 x 1 x 1 cube, 
     centered at origin, on XZ floor  */
  {
    Point3d[] coords = {   // 8 vertices
       new Point3d(0.5, 0, 0.5),   // front 
       new Point3d(0.5, 1, 0.5),
       new Point3d(-0.5, 1, 0.5),
       new Point3d(-0.5, 0, 0.5),

       new Point3d(-0.5, 0, -0.5),  // back
       new Point3d(-0.5, 1, -0.5),
       new Point3d(0.5, 1, -0.5),
       new Point3d(0.5, 0, -0.5)
    };
    return coords;
  }


  private static Point3d[] pyraCoords()
  /* coordinates for a truncated pyramid base of 2 x 2 x 2, 
     centered at origin, on XZ floor  */
  {
    Point3d[] coords = {   // 8 vertices
       new Point3d(1, 0, 1),   // front 
       new Point3d(0.5, 1, 0.5),
       new Point3d(-0.5, 1, 0.5),
       new Point3d(-1, 0, 1),

       new Point3d(-1, 0, -1),  // back
       new Point3d(-0.5, 1, -0.5),
       new Point3d(0.5, 1, -0.5),
       new Point3d(1, 0, -1)
    };
    return coords;
  }


  private static Point3d[] ipyraCoords()
  /* coordinates for a truncated inverted pyramid with top of 2 x 2 x 2, 
     centered at origin, on XZ floor */
  {
    Point3d[] coords = {   // 8 vertices
       new Point3d(0.5, 0, 0.5),   // front 
       new Point3d(1, 1, 1),
       new Point3d(-1, 1, 1),
       new Point3d(-0.5, 0, 0.5),

       new Point3d(-0.5, 0, -0.5),  // back
       new Point3d(-1, 1, -1),
       new Point3d(1, 1, -1),
       new Point3d(0.5, 0, -0.5)
    };
    return coords;
  }




  public void tick(int delay)
  /* animation code. The morphing is expressed in the weights of
     adjacent frames which add up to a single 'shape' being
     drawn for this tick() call
   */
  {
    double[] weights = new double[NUM_FRAMES];  // all values are 0
    
    float scaledAlpha = (float) (NUM_FRAMES * alpha.value()); 
                               // a float between 0 and NUM_FRAMES
    int idx = (int) scaledAlpha; 
          // truncate to convert into an index for weights[]
    
    // increase weight stored at next index; decrease current one;
    double nextWeight = (double)scaledAlpha - (double)idx;
    double currWeight = 1.0 - nextWeight;       // total of all weights must == 1

    /* the rendered shape is these weights applied to adjacent shapes
       in the Morph object */
    weights[idx] = currWeight;
    weights[(idx+1)%NUM_FRAMES] = nextWeight;    // use modulo to wraparound in array
/*
    // useful for debugging purposes
    System.out.println("scaledAlpha: " + scaledAlpha + 
                       "; index: " + idx);
    System.out.println("weights: " + Arrays.toString(weights));
*/
    morph.setWeights(weights);

    j3d.canvas3D.getView().repaint();  
        // this line is necessary to force the morphed shape to be redrawn
  }  // end of tick()



  // ------------------------------------

  public static void main(String[] args)
  /*
       rootBG --> sceneBG
            |
            ----> TimeBehavior
  */
  {
    Simple3D j3d = new Simple3D("Morph a Cube", true, true);

    MorphCube mc = new MorphCube(j3d);
    TimeBehavior tb = new TimeBehavior(DELAY, mc);

    //assemble scene graph
    BranchGroup bg = Simple3D.createBG();
    bg.addChild(mc.getScene());
    bg.addChild(tb);
    j3d.rootBG.addChild(bg);
  }  // end of main()


}  // end of MorphCube class
