
// MovingModels2.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, October 2021

/*  Use a TimeBehavior to create: 
      * a rotating Earth
      * a circling car
      * levitating text

    See MovingModels.java for a DIY animation version.

    This version requires the creation of an object so that
    the class can implement TimeUpdater. Its tick() method
    is called every DELAY ms by the Behavior.
*/

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;
import org.jogamp.java3d.utils.geometry.*;



public class MovingModels2 implements TimeUpdater
{
  private static final int DELAY= 50;  // time between updates in ms

  private static final double CAR_RADIUS = 2.5;  // angle step on each update
  private static final double CAR_ANGLE = 5;
  private static final double EARTH_ANGLE = 5;
  private static final int MAX_LEV_STEPS = 20;

  private BranchGroup sceneBG;

  // globals for animation
  private TransformGroup posCarTG, rotCarTG, 
                         rotSphereTG, posTextTG;

  private double carAngle = 180; 
  private double levDist = 0.1;
  private int levStep = 0;



  public MovingModels2()
  {  
    /*
       sceneBG --> posCarTG --> rotCarTG --> car model
            |
            -----> pos2 --> rotSphereTG --> sphere
            |
            -----> posTextTG --> scale --> text
    */

    // a moveable car
    BranchGroup modelBG = Simple3D.loadModel("models/car/car.obj");
    rotCarTG = Simple3D.createTG(); 
    rotCarTG.addChild(modelBG); 
    posCarTG = Simple3D.createTG(); 
    posCarTG.addChild(rotCarTG); 

    // a rotating, positioned sphere 
    Sphere earth = Simple3D.texSphere(1, "images/earth1.jpg");
    rotSphereTG = Simple3D.createTG(); 
    rotSphereTG.addChild(earth); 
    TransformGroup pos2TG = Simple3D.setTranslation(-2,2,0);
    pos2TG.addChild(rotSphereTG);  

    // scaled, positioned text
    BranchGroup text = Simple3D.text3D("Simple3D");
    TransformGroup scaleTG = Simple3D.scale(0.3);
    scaleTG.addChild(text);  
    posTextTG = Simple3D.setTranslation(3,0.1,0);
    posTextTG.addChild(scaleTG);   

    // build top-level scene
    sceneBG = Simple3D.createBG();
    sceneBG.addChild(posCarTG);
    sceneBG.addChild(pos2TG);
    sceneBG.addChild(posTextTG); 
  }  // end of movingModels2()


  public BranchGroup getScene()
  {  return sceneBG;  }


  public void tick(int delay)
  // animation code; equivalent to animate() in MovingModels
  {
    // rotate the sphere
    Simple3D.rotYBy(rotSphereTG, EARTH_ANGLE);  
   
    // drive the car in a circle
    carAngle = (carAngle + CAR_ANGLE)%360;
    double rads = Math.toRadians(carAngle);
    Simple3D.moveTo(posCarTG, 
                  CAR_RADIUS+CAR_RADIUS*Math.cos(rads), 0, 
                 -CAR_RADIUS*Math.sin(rads)); 
    Simple3D.rotYBy(rotCarTG, CAR_ANGLE); 
   
    // move the text up or down
    Simple3D.moveBy(posTextTG, 0, levDist, 0);
    levStep++;
    if (levStep > MAX_LEV_STEPS) {  // toggle direction
      levDist = -levDist;
      levStep = 0;
    }
  }  // end of tick()

 

  // --------------------------------

  public static void main(String[] args)
  /*
       rootBG --> sceneBG
            |
            ----> TimeBehavior
  */
  {  
    Simple3D j3d = new Simple3D("Moving Models 2", 0,1,10, true, true);

    MovingModels2 mm = new MovingModels2();
    TimeBehavior tb = new TimeBehavior(DELAY, mm);

    // build top-level scene
    BranchGroup bg = Simple3D.createBG();
    bg.addChild(mm.getScene());
    bg.addChild(tb);

    j3d.rootBG.addChild(bg);
  }  // end of main()

}  // end of MovingModels2 class
