
// PenguinWalks.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, November 2021

/* Four ways to have a penguin walk in a loop of radius 2
   counter clockwise around the origin on the XZ plane. Each
   loop takes 6 secs to complete, and keeps repeating.
*/


import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;


public class PenguinWalks 
{

  public static void main(String[] args)
  {
    Simple3D j3d = new Simple3D("Walking Penguin", true, true);
    BranchGroup penguinBG = Simple3D.loadModel("models/penguin/penguin.obj");

    TransformGroup rot1TG = Simple3D.rotY(90);  // face right
    rot1TG.addChild(penguinBG);

    // four ways to use an interpolator
    // TransformGroup pathTG = orbitY(rot1TG, 2);               // 1 
    // TransformGroup pathTG = squarePath(rot1TG, 2);        // 2
    TransformGroup pathTG = squareRotPath(rot1TG, 2);     // 3
    // TransformGroup pathTG = pathRotateY(rot1TG, 2);       // 4

    // build the scene
    BranchGroup bg = Simple3D.createBG(pathTG);
    j3d.rootBG.addChild(bg);
  }  // end of main()



  private static TransformGroup orbitY(TransformGroup obj, double radius)
  {
    TransformGroup posTG = Simple3D.setTranslation(0,0,radius);
    posTG.addChild(obj);

    TransformGroup rotTG = Simple3D.orbit(posTG, 6000);
    return rotTG;
  }  // end of orbitY()




  private static TransformGroup squarePath(TransformGroup obj, double radius)
  /*
     A square path with no rotation of the penguin.

       pathTG --> obj
        ^
        |
        ----> pos path
  */
  {
    TransformGroup pathTG = Simple3D.createTG();
    pathTG.addChild(obj);

    float r = (float)radius;
    Point3f[] positions = { new Point3f(0,0,r),   // point 0
                            new Point3f(r,0,r),
                            new Point3f(r,0,0),
                            new Point3f(r,0,-r),
                            new Point3f(0,0,-r),
                            new Point3f(-r,0,-r),
                            new Point3f(-r,0,0),
                            new Point3f(-r,0,r),
                            new Point3f(0,0,r)    // back at start; point 8
         };

    float[] knots = {  0.0f, 0.1f, 0.2f, 0.3f, 0.4f, 
                       0.5f, 0.6f, 0.7f, 1.0f 
                    };

    Alpha alpha = new Alpha(-1, 6000);
    alpha.setStartTime( System.currentTimeMillis()); // start now

    Transform3D axis = new Transform3D();   // + y-axis
    PositionPathInterpolator ppInterp =
       new PositionPathInterpolator(alpha, pathTG, axis, knots, positions);

    ppInterp.setSchedulingBounds(Simple3D.INFINITE_BOUNDS);

    pathTG.addChild(ppInterp);
    return pathTG;
  }  // end of squarePath()



  private static TransformGroup squareRotPath(TransformGroup obj, double radius)
  /*
     A square path with penguin rotations.

       tg --> obj
        ^
        |
        ----> rot pos path
  */
  {
    float r = (float)radius;

    RotPosPath rpPath = new RotPosPath(Simple3D.Y_AXIS);
    rpPath.add(0,0,r, 0.0f);  // datum 0
    rpPath.add(r,0,r, 0.0f);  // unchanged rot, so only a move
    rpPath.add(r,0,r, 90.0f); // unchanged pos, so only a rotation

    rpPath.add(r,0,0,  90.0f);  
    rpPath.add(r,0,-r, 90.0f);  
    rpPath.add(r,0,-r, 180.0f);

    rpPath.add(0,0,-r,  180.0f);
    rpPath.add(-r,0,-r, 180.0f);
    rpPath.add(-r,0,-r, 270.0f);

    rpPath.add(-r,0,0, 270.0f);
    rpPath.add(-r,0,r, 270.0f);
    rpPath.add(-r,0,r, 0.0f);

    rpPath.add(0,0,r, 0.0f);   // datum 12

    return rpPath.getTG(obj, 6000);
  }  // end of squareRotPath()



  private static TransformGroup pathRotateY(TransformGroup obj, double radius)
  // A path that closely matches a circle.
  {
    RotPosPath rpPath = new RotPosPath();
    for (float i = 0.0f; i <= 360.0f; i = i + 10.0f) {
      float z = (float)(radius*Math.cos( Math.toRadians(i)));
      float x = (float)(radius*Math.sin( Math.toRadians(i)));
      rpPath.add(x,0.0f,z, i);   // (x,y,z), angle (both changing)
    }

    return rpPath.getTG(obj, 6000);
  }  // end of pathRotateY()


} // end of class PenguinWalks
