
// PickActs.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, Nov. 2021

/*
  An example using PickingBehavior. Clicking the cube will make it
  rotate. Clicking the sphere will make it shrink or grow alternatingly.

  The rotation, growing, and shrinking are controlled by PulseAlpha timers
  connected to various interpolators.
 
  Based on code by Frank Klawonn.

  This version requires the creation of an object so that
  the class can implement PickUpdater. Its picked() method
  is called whenever a pick is detected by the Behavior.
*/

import java.util.*;

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;
import org.jogamp.java3d.utils.geometry.*;


public class PickActs implements PickUpdater
{

  private BranchGroup sceneBG;

  // globals for animation
  private FireAlpha boxAlpha, shrinkAlpha, growAlpha;

  // should the sphere shrink or grow?
  private boolean isShrinking = true;


  public PickActs()
  {
    // box sub-branch
    /*   posTG -->  boxTG --> box
                      |
                      ----> spinning interpolator
    */
    Box box = new Box(0.5f, 0.5f, 0.5f, Simple3D.color("red"));
    box.setName("box");

    TransformGroup boxTG = new TransformGroup();
    boxTG.addChild(box);
    animateBox(boxTG);

    TransformGroup posTG = Simple3D.setTranslation(0,0.5,0);
    posTG.addChild(boxTG);


    // sphere sub-branch
    /*   posSTG -->  sphereTG --> sphere
                        |
                        ----> shrinking interpolator
                        |
                        ----> growing interpolator
    */
    Sphere sphere = new Sphere(1f, Simple3D.color("green"));
    sphere.setName("sphere");

    TransformGroup sphereTG = new TransformGroup();
    sphereTG.addChild(sphere);
    animateSphere(sphereTG);  // builds both interpolators

    // position the sphere
    TransformGroup posSTG = Simple3D.setTranslation(2,1,-2);
    posSTG.addChild(sphereTG);


    // top-level of the scene graph
    /*   sceneBG --> posTG 
                 |
                 --> posSTG
    */
    sceneBG = Simple3D.createBG();
    sceneBG.addChild(posTG);
    sceneBG.addChild(posSTG);

  }  // end of PickActs()



  public BranchGroup getScene()
  {  return sceneBG;  }



  private void animateBox(TransformGroup boxTG)
  /* make the box spin when picked using a
     RotationInterpolator controlled by a FireAlpha object
  */
  {
    boxAlpha = new FireAlpha(2000);  // 2 sec duration
    RotationInterpolator spinning = 
            new RotationInterpolator(boxAlpha.getAlpha(), boxTG);
             // default is one complete rotation around the y-axis
    spinning.setSchedulingBounds(Simple3D.INFINITE_BOUNDS);

    boxTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
    boxTG.addChild(spinning);
  }  // end of animateBox()




  private void animateSphere(TransformGroup sphereTG)
  /* make the sphere shrink or grow when picked using two
     ScaleInterpolators controlled by FireAlpha objects
  */
  {
    // shrinking animation: radius goes from 1 to 0.5
    shrinkAlpha = new FireAlpha(2000); // 2 sec duration
    ScaleInterpolator shrinking = 
              new ScaleInterpolator(shrinkAlpha.getAlpha(), sphereTG,
                                    new Transform3D(), 1.0f, 0.5f);
                       // scaling applied to an identity matrix 
    shrinking.setSchedulingBounds(Simple3D.INFINITE_BOUNDS);

    sphereTG.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
    sphereTG.addChild(shrinking);

    // growing animation: radius goes from 0.5 to 1
    growAlpha = new FireAlpha(2000); // 2 sec duration
    ScaleInterpolator growing = 
              new ScaleInterpolator(growAlpha.getAlpha(), sphereTG,
                                  new Transform3D(), 0.5f, 1.0f);
    growing.setSchedulingBounds(Simple3D.INFINITE_BOUNDS);
    sphereTG.addChild(growing);
  }  // end of animateSphere()



  public void picked(String id, Point3d pt)
  // the method called by PickBehavior
  {
     if (id == null)
       return;

    if (id.equals("box"))
      boxAlpha.fire();   // rotate the cube
    else if (id.equals("sphere")) {
      // grow or shrink the cube (when previous change has finished)
      if (isShrinking && !growAlpha.isRunning()) {
        shrinkAlpha.fire();
        isShrinking = false;  // Next time the sphere should grow
      }
      else if (!isShrinking && !shrinkAlpha.isRunning()) {
        growAlpha.fire();
        isShrinking = true; // Next time the sphere should shrink
      }
    }
  }  // end of picked()


  // ---------------------------------------------

  public static void main(String[] args)
  /*
       rootBG --> sceneBG
            |
            ----> PickBehavior
  */
  {
    Simple3D j3d = new Simple3D("Pick Acts", true, true);

    PickActs pa = new PickActs();
    PickBehavior pb =  new PickBehavior(pa, j3d.canvas3D, pa.getScene());

    // build the scene graph
    BranchGroup bg = Simple3D.createBG();
    bg.addChild(pa.getScene());
    bg.addChild(pb);

    // Add everything to the universe
    j3d.rootBG.addChild(bg);

  }  // end of main()



}  // end of PickActs class
