
// PickCubes.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, November 2021

/*  Use a PickBehavior to use the mouse to click on three boxes and
    a car model.

    This version requires the creation of an object so that
    the class can implement PickUpdater. Its picked() method
    is called whenever a pick is detected by the Behavior.
*/

import java.util.*;

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;
import org.jogamp.java3d.utils.geometry.*;


public class PickCubes implements PickUpdater
{

  private BranchGroup sceneBG;


  public PickCubes()
  {
    /*   sceneBG --> pos1TG --> red box (pickable)
                 |
                 --> pos2TG --> blue box (pickable)
                 |
                 --> pos3TG --> green box (pickable)
                 |
                 --> pos4TG --> car (a loaded model is always pickable)
    */

    // to be pickable, primitives must have names set with setName()

    Box redBox = new Box(0.5f, 0.5f, 0.5f, Simple3D.color("red"));
    redBox.setName("red box");
    TransformGroup pos1TG = Simple3D.setTranslation(-2,0.5,0);
    pos1TG.addChild(redBox);

    Box blueBox = new Box(0.5f, 0.5f, 0.5f, Simple3D.color("blue"));
    blueBox.setName("blue box");
    TransformGroup pos2TG = Simple3D.setTranslation(0,0.5,0);
    pos2TG.addChild(blueBox);

    Box greenBox = new Box(0.5f, 0.5f, 0.5f, Simple3D.color("green"));
    greenBox.setName("green box");
    TransformGroup pos3TG = Simple3D.setTranslation(2,0.5,0);
    pos3TG.addChild(greenBox);

    BranchGroup modelBG = Simple3D.loadModel("models/car/car.obj");
    TransformGroup pos4TG = Simple3D.setTranslation(0,0,2);
    pos4TG.addChild(modelBG);
       // loadModel() automatically calls setName() with the name of the file

    // build the scene graph
    sceneBG = Simple3D.createBG();
    sceneBG.addChild(pos1TG);
    sceneBG.addChild(pos2TG);
    sceneBG.addChild(pos3TG);
    sceneBG.addChild(pos4TG);
  }  // end of PickCubes()


  public BranchGroup getScene()
  {  return sceneBG;  }



  public void picked(String id, Point3d pt)
  // the method called by PickBehavior
  // Note: the intercept may be null, as in the case of the car model
  {
    if (pt == null)
      System.out.println(id + " picked");
    else
      Simple3D.printTuple(id + " picked at ", pt);
  }  // end of picked()


  // ------------------------------------

  public static void main(String[] args)
  /*
       rootBG --> sceneBG
            |
            ----> PickBehavior
  */
  {
    Simple3D j3d = new Simple3D("Pick Cubes", true, true);

    PickCubes pc = new PickCubes();
    PickBehavior pb =  new PickBehavior(pc, j3d.canvas3D, pc.getScene());

    // build the scene graph
    BranchGroup bg = Simple3D.createBG();
    bg.addChild(pc.getScene());
    bg.addChild(pb);

    j3d.rootBG.addChild(bg);
  }  // end of main()

}  // end of PickCubes class
