
// PickMoveCubes.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, Dec. 2021

/*  Use the mouse to click on three boxes:
      - dragging with the left mouse button pressed causes 
        a translation along the x-axis. If the <ctrl> key is
        held down then the selected box moves along the y-axis

      - rotating the clicked mouse wheel causes a translation along
        the z-axis;

      - dragging the right mouse button pressed causes 
        a rotation around the y-axis. If the <ctrl> key is
        held down then the selected box rotates around the x-axis

*/

import java.util.*;

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;
import org.jogamp.java3d.utils.geometry.*;

import org.jogamp.java3d.utils.picking.behaviors.*; 


public class PickMoveCubes implements PickMoverListener
{
  public PickMoveCubes()
  {
    Simple3D j3d = new Simple3D("Select Cubes", true, true);
    /*
         sceneBG --> 3 pickable boxes
              |
              ----> PickMoverBeh
    */
    // build the scene graph
    BranchGroup sceneBG = Simple3D.createBG();
    sceneBG.addChild( makeBox(-2, 0, "red"));    // (x,z) position on floor
    sceneBG.addChild( makeBox(0, 0, "blue"));
    sceneBG.addChild( makeBox(2, 0, "green"));

    j3d.cameraMouseEnable(false);  
          // disables mouse effects on the camera

    PickMoverBeh pickMove =  new PickMoverBeh(j3d.canvas3D, sceneBG, this);
         // callbacks to transformChanged() come here
    sceneBG.addChild(pickMove); 

    j3d.rootBG.addChild(sceneBG);
  }  // end of PickMoveCubes()



  private TransformGroup makeBox(double x, double z, String colorStr)
  /* 
      posTG --> pickTG --> box (pickable)
  */
  {
    // to be pickable, primitives must have names set with setName()
    Box b = new Box(0.5f, 0.5f, 0.5f, Simple3D.color(colorStr));
    b.setName(colorStr + " box");

    // position the box on the XZ plane
    TransformGroup posTG = Simple3D.setTranslation(x,0.5,z);

    // Create a second transform group node  
    // which can report its picking changes.
    TransformGroup pickTG = Simple3D.createTG();
    pickTG.setCapability(TransformGroup.ENABLE_PICK_REPORTING);

    // build scene graph
    pickTG.addChild(b);
    posTG.addChild(pickTG);
    return posTG;
  }  // end of makeBox()



  public void transformChanged(String id, int type, Transform3D t3d) 
  // called by PickMoverBeh when there is a move
  {
    if (id != null) {
      if (type != PickMoverListener.NO_PICK) {
        System.out.println(id + ": " + getPickType(type));
        if (t3d != null)
          Simple3D.printMatrix(t3d);
        else
          System.out.println("Transform is null");
      }
    }
  }  // end of transformChanged()



  private String getPickType(int t)
  {
    if (t == PickMoverListener.ROTATE)
      return "rotate";
    else if (t == PickMoverListener.TRANSLATE)
      return "translate";
    else if (t == PickMoverListener.ZOOM)
      return "zoom";
    else if (t == PickMoverListener.NO_PICK)
      return "no pick";
    else
      return "??";
  }  // end of getPickType()


  // ----------------------------------------

  public static void main(String[] args)
  {  new PickMoveCubes();   } 

}  // end of PickMoveCubes class
