
// PickMoverBeh.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, Dec. 2021

/*
   Replaces PickTranslateBehavior, PicRotateBehavior, and PickZoomBehavior

   - dragging with the left mouse button pressed causes 
     a translation along the x-axis. If the <ctrl> key is
     held down then the selected box moves along the y-axis

   - rotating the clicked mouse wheel causes a translation along
     the z-axis;

   - dragging the right mouse button pressed causes 
     a rotation around the y-axis. If the <ctrl> key is
     held down then the selected box rotates around the x-axis
*/

import java.awt.event.*;

import org.jogamp.java3d.*;
import org.jogamp.java3d.utils.picking.*;
import org.jogamp.java3d.utils.picking.behaviors.*;
import org.jogamp.java3d.utils.behaviors.mouse.*;


public class PickMoverBeh extends PickMouseBehavior
                          implements MouseBehaviorCallback
{
  private MyMouseTranslate mouseTrans; 
  private MouseWheelZoom mouseZoom;   // part of Java 3D's utilities
  private MyMouseRotate mouseRot; 

  private PickMoverListener pmListener = null;

  private String currID = null;



  public PickMoverBeh(Canvas3D canvas, BranchGroup root, PickMoverListener pml)
  // Initialize mouse behaviors for moving a picked object
  {
    super(canvas, root, Simple3D.INFINITE_BOUNDS);
    pmListener = pml;

    mouseTrans = new MyMouseTranslate(MouseBehavior.MANUAL_WAKEUP); 
    // mouseTrans.setControl(true);   // use <CTRL> key
    mouseTrans.setTransformGroup(currGrp);
    mouseTrans.setupCallback(this);
    currGrp.addChild(mouseTrans);

    mouseZoom = new MouseWheelZoom(MouseBehavior.MANUAL_WAKEUP); 
    mouseZoom.setTransformGroup(currGrp);
    mouseZoom.setupCallback(this);
    currGrp.addChild(mouseZoom);

    mouseRot = new MyMouseRotate(MyMouseRotate.MANUAL_WAKEUP); 
    mouseRot.setControl(true);   // use <CTRL> key
    mouseRot.setTransformGroup(currGrp);
    mouseRot.setupCallback(this);
    currGrp.addChild(mouseRot);

    mouseTrans.setSchedulingBounds(Simple3D.INFINITE_BOUNDS);
    mouseZoom.setSchedulingBounds(Simple3D.INFINITE_BOUNDS);
    mouseRot.setSchedulingBounds(Simple3D.INFINITE_BOUNDS);

    this.setSchedulingBounds(Simple3D.INFINITE_BOUNDS);
  }  // end of PickMoverBeh()



  @Override
  public void updateScene(int xpos, int ypos)
  /* Examine the mouse to decide whether to use
     the mouse behavior for translation, zooming, or rotation
     on the picked TransformGroup.
   */
  {
    TransformGroup tg = null;

    pickCanvas.setShapeLocation(xpos, ypos);
    PickResult pr = pickCanvas.pickClosest();
    currID = Simple3D.pickID(pr);     // get ID of picked object

    if ((pr != null) &&
        ((tg = (TransformGroup)pr.getNode(PickResult.TRANSFORM_GROUP)) != null) &&
         (tg.getCapability(TransformGroup.ALLOW_TRANSFORM_READ)) &&
         (tg.getCapability(TransformGroup.ALLOW_TRANSFORM_WRITE))) {
              // get the pickable TransformGroup of the picked shape

      if ((mevent.getModifiersEx() & InputEvent.BUTTON1_DOWN_MASK) != 0) {
        // left button pressed
        mouseTrans.setTransformGroup(tg);
        mouseTrans.wakeup();
      }

      else if ((mevent.getModifiersEx() & InputEvent.BUTTON2_DOWN_MASK) != 0) { 
        // middle button (the mouse wheel) pressed
        mouseZoom.setTransformGroup(tg);
        mouseZoom.wakeup();
      }

      else if ((mevent.getModifiersEx() & InputEvent.BUTTON3_DOWN_MASK) != 0) { 
        // right button pressed
        mouseRot.setTransformGroup(tg);
        mouseRot.wakeup();
      }
    }
    else if (pmListener != null)
      pmListener.transformChanged(currID, PickingCallback.NO_PICK, null);

  } /* updateScene */



  @Override
  public void transformChanged(int type, Transform3D t3d)
  // called by the mouse behaviors; 
  // this method calls the 3-arg version of transformChanged()
  {  pmListener.transformChanged(currID, type, t3d);  } 


}  // end of PickMoverBeh class

