
// Platonics.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, October 2021

/* Create a scene containing one of the Platonic solids:

   Usage:
    > run Platonics  [t | c | o | d | i]
        t : tetrahedron
        c : cube
        o : octahedron
        d : dodecahedron
        i : icosahedron

   See https://en.wikipedia.org/wiki/Platonic_solid
*/

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;
import org.jogamp.java3d.utils.geometry.*;



public class Platonics
{
  private static final double phi = 0.5*(Math.sqrt(5) + 1);



  public static void main(String[] args)
  {
    if (args.length != 1)
      System.exit(1);

    String shapeNm = args[0].toLowerCase().trim();
    char shapeCh = shapeNm.charAt(0);

    Simple3D j3d = new Simple3D("Platonic Shapes", false, true);

    BranchGroup coordsBG = null;

    Shape3D shape = null;
    if (shapeCh == 't') {
      System.out.println("Shape: tetrahedron");
      coordsBG = labelCoords(tetraCoords());
      shape = createTexShape(tetraCoords(), tetraIndices(), 
                          tetraStrips(), "images/granite.jpg");
    }
    else if (shapeCh == 'c') {
      System.out.println("Shape: cube");
      coordsBG = labelCoords(cubeCoords());
      shape = createTexShape(cubeCoords(), cubeIndices(), 
                          cubeStrips(), "images/marble.jpg");
    }
    else if (shapeCh == 'o') {
      System.out.println("Shape: octahedron");
      coordsBG = labelCoords(octaCoords());
      shape = createTexShape(octaCoords(), octaIndices(), 
                          octaStrips(), "images/grass.jpg");
    }
    else if (shapeCh == 'd') {
      System.out.println("Shape: dodecahedron");
      coordsBG = labelCoords(dodecCoords());
      shape = createTexShape(dodecCoords(), dodecIndices(), 
                          dodecStrips(),  "images/grid.jpg");
    }
    else if (shapeCh == 'i') {
      System.out.println("Shape: icosahedron");
      coordsBG = labelCoords(icosCoords());
      shape = createTexShape(icosCoords(), icosIndices(), 
                          icosStrips(),  "images/stone.jpg");
    }
    else {
      System.out.println("Unknown letter: '" + shapeCh + "'");
      System.out.println("Usage: run Platonics  [t | c | o | d | i]");
      // j3d.exit();
    }

    // build scene graph
    BranchGroup sceneBG = new BranchGroup();
    sceneBG.addChild(shape);
    sceneBG.addChild(coordsBG);

    j3d.rootBG.addChild(sceneBG);
  }  // end of main()



  // ======== SHAPES ===============


  private static Shape3D createShape(Point3d[] coords, 
                              int[] indices, int[] stripCounts,
                              String colorNm)
  {
    GeometryArray geom = createGeom(coords, indices, stripCounts);
    return new Shape3D(geom, Simple3D.color(colorNm) );
  }


  private static Shape3D createTexShape(Point3d[] coords, 
                              int[] indices, int[] stripCounts,
                              String texFnm)
  {
    GeometryArray geom = createGeom(coords, indices, stripCounts);
    return new Shape3D(geom, Simple3D.texture(texFnm));
  }



  private static BranchGroup labelCoords(Point3d[] coords)
  {
    BranchGroup bg = new BranchGroup();
    for (int i =0; i < coords.length; i++)
      bg.addChild( Simple3D.label(coords[i].x, coords[i].y, coords[i].z, 
                                                     ""+i, "yellow", 3));
    return bg;
  }


  // ----- Geometry --------------


  private static GeometryArray createGeom(Point3d[] coords, 
                              int[] indices, int[] stripCounts)
  {
    GeometryInfo gi = new GeometryInfo(GeometryInfo.POLYGON_ARRAY);
    gi.setCoordinates(coords);
    gi.setCoordinateIndices(indices);
    gi.setStripCounts(stripCounts);

    NormalGenerator ng = new NormalGenerator();
    ng.setCreaseAngle( Math.toRadians(35));  // for icosahedron
    // System.out.println("Crease angle: " + 
    //          Math.toDegrees(ng.getCreaseAngle()));
    ng.generateNormals(gi);

    return gi.getGeometryArray();
  }  // end of createGeom()



  private static int[] strips(int numFaces, int numVertsPerFace)
  /* For example, 6 square faces would
     generate {4,4,4,4,4,4}
  */
  { int[] counts = new int[numFaces];
    for (int i=0; i < numFaces; i++)
      counts[i] = numVertsPerFace;
    return counts;
  }





  // ----------------- tetrahedron ---------------------
  // 4 triangular faces; 4 vertices; 6 edges


  private static Point3d[] tetraCoords()
  /*  The coordinates are obtained by starting with an 
      equilateral triangle and adding an apex equidistant 
      from the triangle's vertices. This creates a
      Triangular Pyramid.

      Given an origin-centered equilateral triangle having edge length 2 
      with the coordinates (+/-1, −1/√3), (0, 2/√3),  we add a 
      fourth vertex equidistant from these coordinates, of the 
      form (0, 0, k). 
  */
  {
    double z =  -1 / 3.0;
    double r =  Math.sqrt(1 - z*z);

    double x0 = r * Math.cos(0);
    double y0 = r * Math.sin(0);

    double x1 = r * Math.cos(Math.toRadians(120));
    double y1 = r * Math.sin(Math.toRadians(120));

    double x2 = r * Math.cos(Math.toRadians(240));
    double y2 = r * Math.sin(Math.toRadians(240));

    Point3d[] coords = {   // 4 vertices
      new Point3d(x0, y0, z),        // coord 0
      new Point3d(x1, y1, z),
      new Point3d(x2, y2, z),
      new Point3d(0, 0, 1)
    };

    return coords;
  }


  private static int[] tetraIndices()  // 4 triangular faces
  {
    int[] indices = {2,1,0,  0,1,3,  1,2,3,  2,0,3};   // (CCW)
    return indices;
  }


  private static int[] tetraStrips()  // 4 triangular faces
  {  return new int[] {3, 3, 3, 3};  }


  // ----------------- cube ---------------------
  // 6 square faces; 8 vertices; 12 edges


  private static Point3d[] cubeCoords()
  {
    Point3d[] coords = {   // 8 vertices
       new Point3d(1, -1, 1),         // coord 0
       new Point3d(1, 1, 1),
       new Point3d(-1, 1, 1),
       new Point3d(-1, -1, 1),

       new Point3d(-1, -1, -1),
       new Point3d(-1, 1, -1),
       new Point3d(1, 1, -1),
       new Point3d(1, -1, -1)
    };
    return coords;
  }


  private static int[] cubeIndices()  // 6 square faces
  {
    int indices[] = {
           0,1,2,3,    // front (CCW)
           4,5,6,7,    // back
           0,7,6,1,    // right
           3,2,5,4,    // left
           1,6,5,2,    // top
           4,7,0,3     // bottom
    };
    return indices;
  }


  private static int[] cubeStrips()  // 6 square faces
  {  return new int[]{4,4,4,4,4,4};  }



  // ----------------- octahedron ---------------------
  // 8 triangular faces; 6 vertices; 12 edges


  private static Point3d[] octaCoords()
  /* The coordinates of the regular octahedron are 
     permutations of:
         (+/-1,0,0)

     which create an octahedron of edge length sqrt(2). 
  */
  {
    Point3d[] coords = {   // 6 vertices
       new Point3d(0, 1, 0),    // coord 0
       new Point3d(-1, 0, 0),
       new Point3d(0, 0, -1),
       new Point3d(1, 0, 0),
       new Point3d(0, 0, 1),
       new Point3d(0, -1, 0)
    };
    return coords;
  }


  private static int[] octaIndices()  // 8 triangular faces
  {
    int indices[] = {
           0,1,4,  0,2,1,  0,3,2,  0,4,3,      // (CCW)
           1,2,5,  4,1,5,  2,3,5,  3,4,5
    };
    return indices;
  }


  private static int[] octaStrips()  // 8 triangular faces
  {  return strips(8,3);  }   // {3,3,3,3, 3,3,3,3}




  // -------------------- dodecahedron ------------------------
  // 12 pentagonal faces; 20 vertices; 30 edges


  private static Point3d[] dodecCoords()
  /*  The Cartesian coordinates for the regular dodecahedron, 
       centered on the origin with edge length 2, are:
           (+/-phi, +/-phi, +/-phi)
           (0, +/-phi^2, +/-1)
           (+/-phi^2, +/-1, 0)
           (+/-1, 0, +/-phi^2)

      In the code below, these have been scaled by multiplying
      by 1/phi
  */
  {
    Point3d[] coords = {    // 20 vertices
          new Point3d(1, 1, 1),                  // coord 0
          new Point3d(0, 1/phi, phi),
          new Point3d(phi, 0, 1/phi),
          new Point3d(1/phi, phi, 0),
          new Point3d(-1, 1, 1),

          new Point3d(0, -1/phi, phi),           // coord 5
          new Point3d(1, -1, 1),
          new Point3d(phi, 0, -1/phi),
          new Point3d(1, 1, -1),
          new Point3d(-1/phi, phi, 0),

          new Point3d(-phi, 0, 1/phi),           // coord 10
          new Point3d(-1, -1, 1),
          new Point3d(1/phi, -phi, 0),
          new Point3d(1, -1, -1),
          new Point3d(0, 1/phi, -phi),

          new Point3d(-1, 1, -1),                // coord 15
          new Point3d(-1/phi, -phi, 0),
          new Point3d(-phi, 0, -1/phi),
          new Point3d(0, -1/phi, -phi),
          new Point3d(-1, -1, -1)};
    return coords;
  }


  private static int[] dodecIndices()   // 12 pentagonal faces
  {
    int[] indices = {
         0,1,5,6,2,      0,2,7,8,3,      0,3,9,4,1,     // (CCW)
         1,4,10,11,5,    2,6,12,13,7,    3,8,14,15,9,
         5,11,16,12,6,   7,13,18,14,8,   9,15,17,10,4,
         19,16,11,10,17, 19,17,15,14,18, 19,18,13,12,16};
    return indices;
  }


  private static int[] dodecStrips()  // 12 pentagonal faces
  {  return strips(12,5);  }   // {5,5,5, 5,5,5, 5,5,5, 5,5,5}



  // ----------------- icosahedron ---------------------
  // 20 triangular faces; 12 vertices; 30 edges
  // dihedral angle = 138 degrees


  private static Point3d[] icosCoords()
  /* The Cartesian coordinates of an origin-centered regular 
     icosahedron of edge length 2 are:
            (0, +/-1, +/-phi)
            (+/-1, +/-phi, 0)
            (+/-phi, 0, +/-1)
  */
  {
    Point3d[] coords = {    // 12 vertices
       new Point3d( 0, 1, phi),           // coord 0
       new Point3d( 0, -1, phi),
       new Point3d( 0, 1, -phi),
       new Point3d( 0, -1, -phi),
       new Point3d(1, phi, 0),
       new Point3d(1, -phi, 0),

       new Point3d(-1, phi, 0),           // coord 6
       new Point3d(-1, -phi, 0),
       new Point3d(phi, 0, 1),
       new Point3d(-phi, 0, 1),
       new Point3d(phi, 0, -1),
       new Point3d(-phi, 0, -1)
    };
    return coords;
  }


  private static int[] icosIndices()
  {
    int indices[] = {    // 20 triangular faces, (CCW)
         0, 1, 8,
         1, 0, 9,
         8, 1, 5,
         5, 1, 7,
         1, 9, 7,

         7, 9, 11,
         7, 11, 3,
         11, 6, 2,
         11, 9, 6,
         11, 2, 3,

         3, 2, 10,
         10, 2, 4,
         10, 4, 8,
         8, 4, 0,
         0, 4, 6,

         6, 4, 2,
         5, 3, 10,
         0, 6, 9,
         5, 7, 3,
         10, 8, 5
    };
    return indices;
  }


  private static int[] icosStrips()  // 20 triangular faces
  {  return strips(20,3);  }
    //  {3,3,3,3,3, 3,3,3,3,3, 3,3,3,3,3, 3,3,3,3,3};


}  // end of Platonics class
