
// PulseCube.java
// Andrew Davison, November 2021, ad@fivedots.coe.psu.ac.th

/*  Uses a time-based Behavior (GeometryUpdaterBeh) to change
    the geometry of a cube so that its top, front, right corner
    repeatedly extends up and then reverts to normal. 

    It is NOT possible to use Simple3D's TimeBehavior here since the
    Behavior must call GeometryArray.updateData(GeometryUpdater gu)
    which will cause Java 3D to call GeometryUpdater.updateData(Geometry geo)
    which is defined here.
*/

import java.util.*;

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;
import org.jogamp.java3d.utils.geometry.*;



public class PulseCube implements GeometryUpdater
{
  private static final int NUM_COORDS = 8;   // in the cube

  private static final int X_COORD = 0;
  private static final int Y_COORD = 1;
  private static final int Z_COORD = 2;

  // apply MAX_STEPS distance changes before reversing the changes
  private static final int MAX_STEPS = 10;  
  private static final double CHG_DIST = 0.05;  

  private Geometry geom;
  private Shape3D shape;
  private double[] coords;  // so available to updateData()

  private int currStep = 0;
  private boolean isGrowing = true;



  public PulseCube()
  {
    geom = createCube();
    shape = new Shape3D(geom, Simple3D.color("gray"));
  } 


  public Shape3D getShape()
  { return shape;  }


  public Geometry getGeom()
  { return geom;  }


  private Geometry createCube()
  // 1 x 1 x 1 cube, centered at origin, on XZ floor
  // with normals applied to the faces
  // based on http://www.java2s.com/Code/Java/3D/SimpleIndexedQuad.htm
  {
    IndexedQuadArray qa = new IndexedQuadArray(NUM_COORDS,
                                IndexedQuadArray.COORDINATES|
                                IndexedQuadArray.NORMALS |
                                IndexedQuadArray.BY_REFERENCE, NUM_COORDS*3);
       /* geometry updating requires that the code be stored by
          reference so that changes to the coords[] array affect the
          shape */

    // the referenced data can be read and written
    qa.setCapability(IndexedQuadArray.ALLOW_REF_DATA_READ);
    qa.setCapability(IndexedQuadArray.ALLOW_REF_DATA_WRITE);

    coords = new double[] { 
               0.5, 0, 0.5,   // 8 vertices; 3 values/coord
               0.5, 1, 0.5,   // 4 coords at the front
              -0.5, 1, 0.5,
              -0.5, 0, 0.5,
	          
              -0.5, 0, -0.5,  // 4 coords at the back
              -0.5, 1, -0.5,
               0.5, 1, -0.5,
               0.5, 0, -0.5
    };
    qa.setCoordRefDouble(coords);

    int indices[] = {
           0,1,2,3,    // front (CCW)
           4,5,6,7,    // back
           0,7,6,1,    // right
           3,2,5,4,    // left
           1,6,5,2,    // top
           4,7,0,3     // bottom
    };
    qa.setCoordinateIndices(0, indices);


    float[] normals = { 
            0,  0,  1f,    // +z
            0,  0, -1f,
           1f,  0,  0,     // +x
          -1f,  0,  0,
            0, 1f,  0,     // +y
            0,-1f,  0
       };
    qa.setNormalRefFloat(normals);

    // indices for normals array
    int normalIndices[] = { 0, 0, 0, 0, 
                            1, 1, 1, 1, 
                            2, 2, 2, 2, 
                            3, 3, 3, 3,
                            4, 4, 4, 4, 
                            5, 5, 5, 5   };
    //Set the data
    qa.setNormalIndices(0, normalIndices);

    return qa;
  }  // end of createCube()



  public void updateData(Geometry geo)
  /* This method is trigged in GeometryUpdaterBeh.processStimulus()
     when it calls GeometryArray.updateData(GeometryUpdater gu).
     Java 3D will then call GeometryUpdater.updateData(Geometry geo)
     which is defined here.

     Note: only a single coordinate (corner 1) is changed.
     There is no change
     to the normals of the faces affected by that change.
  */
  { 
  //  IndexedQuadArray qa = (IndexedQuadArray)geo;   // not used
  //  double coords[] = qa.getCoordRefDouble(); 

    // is corner 1 of the cube growing or shrinking?
    if (isGrowing) {
      coords[ coordIdx(1,X_COORD) ] += CHG_DIST;  // add change
      coords[ coordIdx(1,Y_COORD) ] += CHG_DIST;
      coords[ coordIdx(1,Z_COORD) ] += CHG_DIST;
      if (currStep == MAX_STEPS)
        isGrowing = false;
      else
        currStep++;
    }
    else { // shrinking
      coords[ coordIdx(1,X_COORD) ] -= CHG_DIST;  // remove change
      coords[ coordIdx(1,Y_COORD) ] -= CHG_DIST;
      coords[ coordIdx(1,Z_COORD) ] -= CHG_DIST;
      if (currStep == 0)
        isGrowing = true;
      else
        currStep--;
    }
  }  // end of updateData()



  private int coordIdx(int num, int axisVal)
  // convert coordinate number and axis value into coords[] index
  {
    if ((num < 0) || (num >= NUM_COORDS)) {
      System.out.println("Unknown coord index");
      return 0;
    }

    if (axisVal == X_COORD)
      return num*3;
    else if (axisVal == Y_COORD)
      return num*3 + 1;
    else if (axisVal == Z_COORD)
      return num*3 + 2;
    else {
      System.out.println("Unknown axis");
      return num*3;
    }
  }  // end of coordIdx()



  // ----------------------------------------------------------


  public static void main(String[] args)
  /*
       rootBG --> pulsating cube
            |
            ----> GeometryUpdaterBeh
  */
  {
    Simple3D j3d = new Simple3D("Pulsating Cube", true, true);

    PulseCube pc = new PulseCube();
    GeometryUpdaterBeh gub = 
            new GeometryUpdaterBeh(pc.getGeom(), pc);

    //assemble scene graph
    BranchGroup bg = Simple3D.createBG();
    bg.addChild(pc.getShape());
    bg.addChild(gub);
    j3d.rootBG.addChild(bg);
  }  // end of main()


} // end of PulseCube class
