
// ExamineObj.java
// Andrew Davison, October 2021, ad@fivedots.coe.psu.ac.th

/* Usage:
     >  java ExamineObj hexSphere.obj
     >  java ExamineObj penguin/penguin.obj
     >  java ExamineObj bird/Hummingbird.obj


  Useful info on the OBJ format:

    https://jogamp.org/deployment/java3d/1.7.0-final/javadoc/
       -- look at the ObjectFile class docs

    https://en.wikipedia.org/wiki/Wavefront_.obj_file

    https://cs.wellesley.edu/~cs307/readings/obj-ojects.html

    http://paulbourke.net/dataformats/mtl/
*/

import java.util.*;
import java.io.*;
import java.nio.file.*;



public class ExamineObj
{

  public static void main(String[] args)
  {
    if (args.length != 1) {
      System.out.println("Usage: java ExamineObj <filename>");
      System.exit(1);
    }
    readObj(args[0]);
  }


  private static void readObj(String fnm)
  /* The format of the input lines are:
         - v, f, vn, vt, s, and g OBJ lines which are counted;
         - other lines which are printed with a line number;
     and blank lines and comments.
  */
  { 
    int vertCount = 0;    // count of "v" lines
    int faceCount = 0;    // count of "f" lines
    int texCount = 0;    // count of "vt" lines
    int normalCount = 0;    // count of "vn" lines
    int smoothCount = 0;    // count of "s" lines
    int groupCount = 0;    // count of "g" lines

    String mtlFnm = null;

    ArrayList<String> groupNames = new ArrayList<>();
    ArrayList<String> usedMtls = new ArrayList<>();

    int lineNum = 0;
    try {
      BufferedReader br = new BufferedReader( new FileReader(fnm));
      String line;
      String[] toks;
      while((line = br.readLine()) != null) {
        lineNum++;
        line = line.trim();
        toks = line.split("\\s+");

        if (line.length() == 0)  // blank line
          {}
        else if (line.startsWith("#"))  // comment line
          {}
        else if (toks[0].equals("v"))
          vertCount++;
        else if (toks[0].equals("f"))
          faceCount++;
        else if (toks[0].equals("vt"))
          texCount++;
        else if (toks[0].equals("vn"))
          normalCount++;
        else if (toks[0].equals("s"))
          smoothCount++;
        else if (toks[0].equals("g")) {
          groupCount++;
          if (toks.length == 1)
            addName(groupNames, "default");
          else
            addName(groupNames, toks[1]);
        }
        else {
          System.out.println("  Line " + lineNum + ": " + line);

          if (toks[0].equals("mtllib")) {
            Path parent = Paths.get(fnm).toAbsolutePath().getParent();
            Path mtlPath = Paths.get(parent+"/"+toks[1]);
            // System.out.println(parent+"/"+toks[1]);
            if (!Files.exists(mtlPath))
              System.out.println("   -- warning: cannot find the MTL file");
            else
              mtlFnm = parent+"/"+toks[1];
          }
          else if (toks[0].equals("usemtl"))
            addName(usedMtls, toks[1]);
          else if (toks[0].equals("o"))
            System.out.println("   -- warning: the Java 3D loader does not recognize this; comment it out");
        }
      }
      br.close();

      System.out.println("Total No. of lines: " + lineNum);

      if (vertCount == 0)
        System.out.println(" Error: no vertices found");
      else
        System.out.println(" No. of \"v\" lines: " + vertCount);

      if (faceCount == 0)
        System.out.println(" Error: no faces found");
      else
        System.out.println(" No. of \"f\" lines: " + faceCount);

      if (texCount > 0)
        System.out.println(" No. of \"vt\" lines: " + texCount);
      if (normalCount > 0)
        System.out.println(" No. of \"vn\" lines: " + normalCount);
      if (smoothCount > 0)
        System.out.println(" No. of \"s\" lines: " + smoothCount);

      if (groupCount > 0) {
        System.out.println(" No. of \"g\" lines: " + groupCount);
        if (groupNames.isEmpty())
          groupNames.add("default");
        else
          Collections.sort(groupNames);
        System.out.println("  Group names: " + groupNames);
      }

      if (mtlFnm != null) {
        System.out.println("--------\nExamining MTL file");
        readMtl(mtlFnm, usedMtls);
      }
    } 
    catch (IOException e) 
    {   System.out.println(e);   }
  }  // end of readObj()


  private static void addName(ArrayList<String> names, String nm)
  {
    if (!names.contains(nm))
      names.add(nm);
  }



  private static void readMtl(String fnm, ArrayList<String> usedMtls)
  /* The format of the input lines are:
         newmtl, and  ?? operations
     and blank lines and comments.
  */
  { 
    // int mtlCount = 0;    // count of "newmtl" lines

    ArrayList<String> mtlNames = new ArrayList<>();

    int lineNum = 0;
    try {
      BufferedReader br = new BufferedReader( new FileReader(fnm));
      String line;
      String[] toks;
      while((line = br.readLine()) != null) {
        lineNum++;
        line = line.trim();
        toks = line.split("\\s+");

        if (line.length() == 0)  // blank line
          {}
        else if (line.startsWith("#"))  // comment line
          {}
        else if (toks[0].equals("newmtl")) {
          // System.out.println("  Line " + lineNum + ": " + line);
          addName(mtlNames, toks[1]);
        }
        else if ((toks[0].equals("map_Ka")) || (toks[0].equals("map_Kd"))) {
          System.out.println("  Line " + lineNum + ": " + line);

          Path parent = Paths.get(fnm).toAbsolutePath().getParent();
          Path imPath = Paths.get(parent+"/"+toks[1]);
          // System.out.println(parent+"/"+toks[1]);
          if (!Files.exists(imPath))
            System.out.println("   -- warning: cannot find the image file");
          
        }
      }
      br.close();

      // System.out.println("Total No. of lines: " + lineNum);

      if (mtlNames.isEmpty())
        System.out.println("Warning: no materials found");
      else
        System.out.println("Materials defined: " + mtlNames);
        

      for(String nm : usedMtls) {
        if (!mtlNames.contains(nm))
          System.out.println(" - Obj material " + nm + " not defined");
      }
      for(String nm : mtlNames) {
        if (!usedMtls.contains(nm))
          System.out.println(" - Material " + nm + " not used in OBJ file");
      }
      
    } 
    catch (IOException e) 
    {   System.out.println(e);   }
  }  // end of readMtl()



}  // end of ExamineObj class
