
// TGViewer.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, October 2021


import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import javax.swing.border.*;

import org.jogamp.java3d.*;
import org.jogamp.vecmath.*;


public class TGViewer
{
  private static final int GRID_WIDTH = 4;
  private static final int GRID_HEIGHT = 7;
  private static final int OFFSET = 35;   // for positioning windows

  private boolean showWorld;  // show world coordinates?
  private int gridHeight;

  private JPanel grid;
  private JLabel[] cells;

  private TransformGroup tg;
  private String tgName;
  private Timer timer;

  // private int counter = 0;  // for timer testing

  // window positioning
  private static int xWin, yWin;
  static {
    GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
    GraphicsDevice defaultScreen = ge.getDefaultScreenDevice();
    Rectangle rect = defaultScreen.getDefaultConfiguration().getBounds();
    xWin = (int) rect.getMaxX() - 10;
    yWin = 10;
  }


  public TGViewer(String nm, TransformGroup tg)
  {  this(nm, tg, false);  }




  public TGViewer(String nm, TransformGroup tg, boolean showWorld)
  {
    tgName = nm;
    this.tg = tg;
    this.showWorld = showWorld;

    gridHeight = (showWorld) ? GRID_HEIGHT : GRID_HEIGHT-3;

    grid = new JPanel(new GridLayout(gridHeight, GRID_WIDTH, 3, 3));      

    cells = new JLabel[gridHeight*GRID_WIDTH];
    for (int i = 0; i < gridHeight*GRID_WIDTH; i++) {
      cells[i] = createCell();
      grid.add( cells[i]);
    }

    cells[index(0,0)].setText(tgName);

    // label rows
    if (showWorld) {
      labelCell(cells[index(1,0)], "WPos");
      labelCell(cells[index(2,0)], "Pos");
      labelCell(cells[index(3,0)], "WRots");
      labelCell(cells[index(4,0)], "Rots");
      labelCell(cells[index(5,0)], "WScale");
      labelCell(cells[index(6,0)], "Scale");
    }
    else {
      labelCell(cells[index(1,0)], "Pos");
      labelCell(cells[index(2,0)], "Rots");
      labelCell(cells[index(3,0)], "Scale");
    }

    // label columns
    labelCell(cells[index(0,1)], "X");
    labelCell(cells[index(0,2)], "Y");
    labelCell(cells[index(0,3)], "Z");

    // create window
    JFrame f = new JFrame("TGViewer for " + tgName);
    f.setContentPane(grid);
    // f.setSize(200, 200);
    f.pack();
    f.setResizable(false);

    f.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);  // do not exit appl.
    f.addWindowListener(new java.awt.event.WindowAdapter() {
       public void windowClosing(WindowEvent e) 
       {   timer.stop();  }
    });

    f.setLocation(TGViewer.xWin - f.getWidth(), TGViewer.yWin);
    TGViewer.xWin -= TGViewer.OFFSET; 
    TGViewer.yWin += TGViewer.OFFSET; 

    f.setVisible(true);

    // start timer
    timer = new Timer(500, new ActionListener() {
      public void actionPerformed(ActionEvent e)
      {  refreshTG();  }
    });
    timer.start();
  }  // end of TGViewer()



  private JLabel createCell()
  {
    JLabel jl = new JLabel("", JLabel.CENTER);
    jl.setBorder( BorderFactory.createBevelBorder(
                                 BevelBorder.RAISED));
    jl.setFont(jl.getFont().deriveFont(16f));
    jl.setPreferredSize(new Dimension(90, 30));
    return jl;
  }


  private void labelCell(JLabel jl, String s)
  {
    jl.setText(s);
    jl.setOpaque(true);
    jl.setBackground(Color.lightGray);
  }


  private int index(int row, int col)
  {
    if ((row < 0) || (row > gridHeight-1) ||
        (col < 0) || (col > GRID_WIDTH-1)) {
      System.out.println("(row, col) out of range: (" + row + 
                                            ", " + col + ")");
      return 0;
    }
    else 
      return  (row*GRID_WIDTH) + col;
  }


  private void refreshTG()
  {
    if (tg != null) {
      Vector3d pos = Simple3D.getPosition(tg);
      Vector3d rots = Simple3D.getRotations(tg);
      Vector3d scales = Simple3D.getScale(tg);

      if (showWorld) {
        Vector3d wPos = Simple3D.getWorldPosition(tg);
        Vector3d wRots = Simple3D.getWorldRotations(tg);
        Vector3d wScales = Simple3D.getWorldScale(tg);

        setCell(0,0, wPos.x);
        setCell(0,1, wPos.y);
        setCell(0,2, wPos.z);
	    
        setCell(1,0, pos.x);
        setCell(1,1, pos.y);
        setCell(1,2, pos.z);
	    
        setCell(2,0, wRots.x);
        setCell(2,1, wRots.y);
        setCell(2,2, wRots.z);
	    
        setCell(3,0, rots.x);
        setCell(3,1, rots.y);
        setCell(3,2, rots.z);
	    
        setCell(4,0, wScales.x);
        setCell(4,1, wScales.y);
        setCell(4,2, wScales.z);
	    
        setCell(5,0, scales.x);
        setCell(5,1, scales.y);
        setCell(5,2, scales.z);
      }
      else {
        setCell(0,0, pos.x);
        setCell(0,1, pos.y);
        setCell(0,2, pos.z);
	    
        setCell(1,0, rots.x);
        setCell(1,1, rots.y);
        setCell(1,2, rots.z);
	    
        setCell(2,0, scales.x);
        setCell(2,1, scales.y);
        setCell(2,2, scales.z);

      }
    }
    else {
      System.out.println("TransformGroup " + tgName + " is null");
      timer.stop();
    }
  }  // end  of refreshTG()



  public void setCell(int matRow, int matCol, double val)
  {
    int row = matRow+1;
    int col = matCol+1;

    if ((row < 1) || (row > gridHeight-1) ||
        (col < 1) || (col > GRID_WIDTH-1)) {
      System.out.println("Position out of range: (" + matRow +
                        ", " + matCol + ")");
      return;
    }

    String sVal = String.format("%.2f", val);
    cells[ index(row,col) ].setText(sVal);
    grid.repaint();
  }  // end of setCell()


  // -----------------------------


  public static void main(String[] args)
  {
    TGViewer tgv = new TGViewer("posST3d1", null);
    tgv.setCell(1,1, 45.2);
    tgv.setCell(2,3, 111);

  } // end of main()

}  // end of TGViewer class